#include <stdlib.h>
#define DMSOURCE
#include "ascdm.h"

dmKernel* CreateKernel;

/*
 * Initialize data model
 */
void dmInit( void )
{
 dmiInit();
 CreateKernel = dmKernelGet( "FITS" );
}

/*
 * dmDatasetCreate
 *
 * Create a dataset
 */
dmDataset* dmDatasetCreate( char* filename )
{
 dmKernelDataset* fptr;
 if ( cs_is_blank( filename ) ) return( NULL );
 fptr = CreateKernel->dmkDatasetCreate( filename );
 return( dmiDatasetAlloc( filename, CreateKernel, fptr, "W" ) );
}

/*
 * dmBlockOpen
 *
 * Open a block by name.  Find the block, then 
 * open in, parsing its structure.
 */
dmBlock* dmBlockOpen( dmDataset* ds, char* name )
{
 dmBlock* block;
 dmsg1( 1, "BlockOpen: ", name );
 block = dmiBlockFind( ds, name );
 if ( block ) {
  dmiBlockOpen( block );
 }
 return( block );
}

/*
 * dmBlockClose
 *
 * Releases memory for the block structure info (keys, columns, dss).
 * Still retains the basic structure so we can remember its name etc.
 */

int dmBlockClose( dmBlock* block )
{
 card buf;
 if ( block ) {
  sprintf(buf, "Close block No = %d Kernel=%ld\n", (int)block->no, (long)block->KernelBlock );
  dmsg( 1, buf );
  dmStatus( block->ds );
  block->ds->kernel->dmkBlockClose( block->KernelBlock );
  block->KernelBlock = NULL;
  dmsg( 1, "DMI Block struct close" );
  dmiBlockStructFree( block );
 }
 return( dmiErrorGet() );
}



/*
 * dmBlockGetNo
 *
 * Get block number in dataset.
 * Trap case of null block.
 * 
 */
int dmBlockGetNo( dmBlock* block )
{
 if ( block ) {
  return( block->no );
 } else {
  return( 0 );
 }
}


/*
 * dmBlockGetDataset
 *
 * Get parent dataset for block
 */
dmDataset* dmBlockGetDataset( dmBlock* block )
{
 if( block ) {
  return( block->ds );
 } else {
  return( NULL );
 }
}


/*
 * dmDatasetGetNoBlocks
 *
 * Number of blocks in dataset.
 * Return zero if null dataset.
 * Scan dataset if needed to determine number of blocks.
 */
int dmDatasetGetNoBlocks( dmDataset* ds )
{
 if ( ds ) {
  if ( !ds->scanned ) dmiDatasetScan( ds, 0 );
  return( ds->nblocks );
 } else {
  return( 0 );
 }
}

/*
 * dmDatasetGetCurrentBlockNo
 *
 */
int dmDatasetGetCurrentBlockNo( dmDataset* ds )
{
 if ( ds ) {
  return( ds->current_block );
 } else {
  return( 0 );
 }
}

/*
 * dmDatasetNextBlock
 *
 */
dmBlock* dmDatasetNextBlock( dmDataset* ds )
{
 if ( ds ) {
  return( dmDatasetMoveToBlock( ds, ds->current_block+1 ) );
 } else {
  return( NULL );
 }
}


/*
 * dmDatasetMoveToBlock
 *
 */
dmBlock* dmDatasetMoveToBlock( dmDataset* ds, int block_no )
{
 dmBlock* block;
 card buf; 
 if ( block_no <= 0 ) {
  dmiErrorSet( ERR_BAD_BLOCK_NO );
  return( NULL );
 } else if ( ds ) {
  if ( block_no > ds->nblocks ) {
   if ( !ds->scanned ) dmiDatasetScan( ds, block_no );
   if ( block_no > ds->nblocks ) return( NULL );
  }
  ds->current_block = block_no;
  block = ds->block[block_no-1];
  sprintf( buf, "Move to block %ld\n", (long)block );
  dmsg( 1, buf );
  if ( block ) {
   ds->kernel->dmkMoveToBlock( block->KernelBlock ); 
   dmiBlockOpen( block );
  }
  return( block );
 } else {
  return( NULL );
 }
}
/*
 * dmDatasetAdvanceBlocks
 *
 */
dmBlock* dmDatasetAdvanceBlocks( dmDataset* ds, int rel_no )
{
 integer block_no;
 if ( ds ) {
  block_no = ds->current_block + rel_no;
  return( dmDatasetMoveToBlock( ds, block_no ) );
 } else {
  return( NULL );
 }
}

/*
 * dmImageOpen
 *
 * Open Image block and dataset
 *
 */
dmBlock* dmImageOpen( char* name )
{
 t_filename dsname;
 card block_name;
 dmDataset* ds;
 dmiSyntaxParse( name, dsname, block_name );
 dmsg1( 1, "Dataset: ", dsname );
 dmsg1( 1, "Block:", block_name );
 ds = dmDatasetOpen( dsname );
 if ( ds ) {
  return( dmBlockOpen( ds, block_name ) );
 } else {
  return( NULL );
 }
}

/*
 * dmImageClose
 *
 * Close Image block and dataset
 */
int dmImageClose( dmBlock* block )
{
 dmDatasetClose( block->ds );
 return( dmiErrorGet() );
}

/*
 * dmGetArrDim
 */
long dmGetArrDim( dmDescriptor* dd )
{
 if ( dd ) {
  return( dd->naxes );
 } else {
  return( 0 );
 }
}

/*
 * dmDatasetTableOpen
 *
 * Currently identical to dmImageOpen
 */
dmBlock* dmDatasetTableOpen( char* name )
{
 t_filename dsname;
 card block_name;
 dmDataset* ds;
 dmiSyntaxParse( name, dsname, block_name );
 dmsg1( 1, "Dataset: ", dsname );
 dmsg1( 1, "Block:", block_name );
 ds = dmDatasetOpen( dsname );
 if ( ds ) {
  return( dmBlockOpen( ds, block_name ) );
 } else {
  return( NULL );
 }
}


/*
 * dmDatasetTableClose
 *
 * Close Table block and dataset
 * Currently identical to dmImageClose
 */
int dmDatasetTableClose( dmBlock* block )
{
 if( block ) dmDatasetClose( block->ds );
 return( dmiErrorGet() );
}


/*
 * dmTableGetNoRows
 */
long dmTableGetNoRows( dmBlock* block )
{
 if ( block ) {
  return( block->nrows );
 } else {
  return( 0 );
 }
}



int dmBlockPrintKernel( dmBlock* block )
{
 if ( block ) {
  block->ds->kernel->dmkBlockPrintKernel( block->KernelBlock );
 }
 return( 0 );
}


int dmBlockPrint( dmBlock* block, char* options )
{
 if ( cs_ss( options, "HEADER" ) ) {
  dmiBlockPrintHeader( block );
 } 
 if ( cs_ss( options, "COLNAMES" ) ) {
  dmiBlockPrintTable( block );
 }
 return( 0 );
}

dmDescriptor* dmTableOpenColumn( dmBlock* block, char* name )
{
 int no;
 dmDescriptor* col;
 if( block ) {
  for ( no = 0; no < block->ncols; no++ ) { 
   col = block->column[no];
   if ( col ) {
    if ( cs_uceq( name, col->name ) ) return( col );
   }
  }
 }
 return( NULL );
}


double dmGetScalar_d( dmDescriptor* dd )
{
 dmKernel* kernel;
 dmBlock* block;
 double result;
 result = 0.0;
 if ( dd ) {
  if ( dd->descriptor_type == ddCOLUMN ) {
   block = dd->block;
   kernel = block->ds->kernel;
   kernel->dmkColumnRead_d( block->KernelBlock, dd->KernelNo, block->current_row, 1, &result, 1 );
  } else {
   tmsg( "Unsupported" );
  }
 }
 return( result );
}

float dmGetScalar_r( dmDescriptor* dd )
{
 dmKernel* kernel;
 dmBlock* block;
 float result;
 result = 0.0;
 if ( dd ) {
  if ( dd->descriptor_type == ddCOLUMN ) {
   block = dd->block;
   kernel = block->ds->kernel;
   kernel->dmkColumnRead_r( block->KernelBlock, dd->KernelNo, block->current_row, 1, &result, 1 );
  } else {
   tmsg( "Unsupported" );
  }
 }
 return( result );
}


long dmGetScalars_d( dmDescriptor* dd, double* result, long row, long nrows )
{
 dmKernel* kernel;
 dmBlock* block;
 if ( !dd ) return( 0 );
 if ( dd->descriptor_type == ddCOLUMN ) {
  block = dd->block;
  kernel = block->ds->kernel;
  nrows = kernel->dmkColumnRead_d( block->KernelBlock, dd->KernelNo, row, nrows, result, 1 );
  block->current_row = row + nrows - 1;
 } else {
  tmsg( "Unsupported" );
  nrows = 0;
 }
 return( nrows );
}


void dmGetStaticArray_d( dmDescriptor* dd, double* result, int npixels )
{
 dmKernel* kernel;
 dmBlock* block;

 if ( !dd ) return;
 block = dd->block;
 kernel = block->ds->kernel;
 if ( dd->descriptor_type == ddCOLUMN ) { 
  if ( block->type == dmIMAGE ) {
  } else {
   kernel->dmkColumnRead_d( block->KernelBlock, dd->KernelNo, block->current_row, 1, result, npixels );
  }
 } else {
  tmsg( "Unsupported 2" );
 }
}

long dmTableNextRow( dmBlock* block ) 
{
 long row = dmNOMOREROWS;
 if ( block ) {
  if ( block->type == dmTABLE ) {
   if( block->current_row < block->nrows ) {
    row = ++(block->current_row);
   }  
  }
 }
 return( row );
}

long dmTablePutRow( dmBlock* block, void* data ) 
{
 char* ptr;
 long row = dmNOMOREROWS;
 if ( block ) {
  if ( block->type == dmTABLE ) {
   row = ++(block->current_row);
   block->ds->kernel->dmkPutRow( block->KernelBlock, row );
  }
  ptr = (char*)data; /* Just to keep compiler happy */
 }
 return( row );
}


dmDescriptor* dmTableOpenColumnNo( dmBlock* block, long no )
{
 dmDescriptor* col;
 if( block ) {
  if ( no > 0 && no <= block->ncols ) {
   col = block->column[no-1];
   return( col );
  }
 }
 return( NULL );
}


dmDescriptor* dmKeyRead_d( dmBlock* block, char* name, double* value )
{
 int no;

 dmDescriptor* key;
 if ( block && !cs_is_blank( name ) ) {
  for ( no = 1; no <= block->nkeys; no++ ) {
   key = block->key[no-1]; 
   if ( cs_uceq( name, key->name ) ) {
    if ( key->data_type == dmDOUBLE ) {
     *value = *(double*)key->value;
    } else {
     tmsg( "Bad key datatype" );
    }
    return( key );
   }
  }
 }
 return( NULL );
}

dmDescriptor* dmKeyRead_l( dmBlock* block, char* name, long* value )
{
 int no;
 dmDescriptor* key;
 if ( block && !cs_is_blank( name ) ) {
  for ( no = 1; no <= block->nkeys; no++ ) {
   key = block->key[no-1]; 
   if ( cs_uceq( name, key->name ) ) {
    if ( key->data_type == dmLONG ) {
     *value = *(long*)key->value;
    } else {
     tmsg( "Bad key datatype" );
    }
    return( key );
   }
  }
 }
 return( NULL );
}

dmDescriptor* dmDescriptorGetCoord( dmDescriptor* dd )
{
 if ( dd ) {
  return( dd->coord );
 } else {
  return( NULL );
 }
}

int dmCoordGetTransform_d( dmDescriptor* coord, double* crpix, double* crval, double* cdelt, long dim )
{
 dmTransform* t;
 if ( dim == 2 ) return(0);
 if ( coord ) {
  if ( coord->descriptor_type == ddCOORD ) {
   dmsg( 1, "Found coord" );
   t = coord->value;
   *crpix = t->crpix;
   *crval = t->crval;
   *cdelt = t->cdelt;
   return( 0 );
  }
 }
 *crpix = 0;
 *crval = 0;
 *cdelt = 0;
 return( 0 );
}



void dmStatus( dmDataset* ds )
{
 int no;
 dmBlock* block;
 card buf;
 if ( !ds ) return;
 for ( no = 1; no <= ds->nblocks; no++ ) {
  block = ds->block[no-1];
  if ( block->key ) {
   sprintf( buf, "%4d %20s %12ld Open   %12ld ", no, block->name, (long)block, (long)block->KernelBlock );
  } else {
   sprintf( buf, "%4d %20s %12ld Closed %12ld", no, block->name, (long)block, (long)block->KernelBlock ); 
  }
  dmsg( 1, buf );
 }
}


int dmSetScalar_d( dmDescriptor* dd, double value )
{
 dmBlock* block;
 dmKernel* kernel;
 if ( dd ) {
  if ( dd->descriptor_type == ddCOLUMN ) {
   block = dd->block;
   kernel = block->ds->kernel;
   kernel->dmkColumnWrite_d( block->KernelBlock, dd->KernelNo, block->current_row, 1, &value, 1 );
  } else {
   tmsg( "Unsupported" );
  }
 }
 return( 0 );
}

dmDescriptor* dmColumnCreate( dmBlock* block, char* name,  dmDataType type, int slen, char* unit, char* desc )
{
 dmDescriptor* dd;
 int colno;
 int no;
 if ( block ) {
  if ( block->nrows > 0 ) {
   tmsg( "Cannot create columns after rows have been written" );
  } else {
   colno = ++block->ncols;
   no = block->ds->kernel->dmkColumnCreate( block->KernelBlock, name, type, slen, unit, desc );
   dd = dmiDescriptorAlloc( block, name, type, ddCOLUMN, colno );
   dd->KernelNo = no;
   block->column[colno-1] = dd;
   return( dd );
  }
 }
 return( NULL );
}


void dmBlockGetNameC( dmBlock* block, char* name, int size )
{
 if ( block ) {
  cs_copy_siz( block->name, name, size );
 } else {
  cs_copy( " ", name );
 }
}


dmDescriptor* dmKeyWrite_d( dmBlock* block, char* name, double value, char* unit, char* desc )
{
 dmDescriptor* dd;
 dd = dmiDescriptorAlloc( block, name, dmDOUBLE, ddKEY, 0 );
 block->ds->kernel->dmkKeyWrite_d( block->KernelBlock, name, value, unit, desc );
 return( dd );
}
