#include <stdlib.h>
#define DMSOURCE
#include "ascdm.h"
/*
 * Global list of datasets
 */

#define dmMAXDS 100

typedef struct {
 int dmNoDatasets;
 dmDataset* dmDatasetList[dmMAXDS];  /* Datasets */
 int dmErrorStatus;
} dmContext;

dmContext* context;
dmContext  Context;

/*
 * dmiInit
 *
 * Create the dm Context.
 * All of the data model's internal state is stored in the context structure.
 * To make the program have an independent context (e.g. to make thread safe),
 * one would simply change the dmiInit routine to dynamically allocate contexts
 * and add a dmiFree routine to free them. For now, we use a static Context to
 * remove the need for users to call dmiFree.
 */
void dmiInit( void )
{
 context = &Context; 
 context->dmNoDatasets = 0;
 context->dmErrorStatus = 0;
}

void dmiFree( void )
{
 context = NULL;
}


/*
 * dmiErrorSet
 *
 * Set a DM internal error
 */
void dmiErrorSet( int status /* FITSIO error code */ )
{
 context->dmErrorStatus = status;
}

/*
 * dmkErrorSet
 *
 * Set a DM kernel error. Currently treated same as for DM internal error.
 *
 */
void dmkErrorSet( int status /* FITSIO error code */ )
{
 context->dmErrorStatus = status;
}

/*
 *
 * dmiErrorGet
 *
 * Return the DM internal error
 */
int dmiErrorGet( void )
{
 return( context->dmErrorStatus );
}

/*
 * dmiDatasetAlloc
 *
 * Internal routine to allocate a dataset and enter it
 * in the master list of datasets.
 */

dmDataset* dmiDatasetAlloc( 
 char* name, 
 dmKernel* kernel, 
 dmKernelDataset* fptr, 
 char* mode                  /* IO mode: read or write */
)
{
 dmDataset* ds;
 card buf;

 dmsg( 1, "dmiDatasetAlloc" );
 if ( fptr == NULL ) {
  return( NULL );
 } else if ( context->dmNoDatasets >= dmMAXDS ) { 
  dmiErrorSet( ERR_TOO_MANY_DS ); return( NULL ); 
 } else {
  context->dmNoDatasets++;    
  ds = (dmDataset*) malloc( sizeof( dmDataset ) );
  ds->block = (dmBlock**) malloc( dmMAXBLOCKS * sizeof( dmBlock* ) );
  sprintf( buf, "*** ALLOC %12ld dmDataset ***", (long)ds ); dmsg( 1, buf );
  context->dmDatasetList[context->dmNoDatasets-1] = ds;
  ds->no = context->dmNoDatasets;
  ds->kernel = kernel;
  ds->fptr = fptr;
  if ( cs_eq( mode, "W" ) ) {
   ds->iomode = dmWRITE;
  } else {
   ds->iomode = dmREAD;
  }

  cs_copy_siz( name, ds->name, FSIZE );
/* Mark the file as unscanned. */
  ds->nblocks = 0;
  ds->scanned = FALSE;
  ds->current_block = 0;
 }
 return( ds );
}

/*
 * dmiDatasetFree
 * 
 * Free memory allocated by dmiAlloc.
 * Free all blocks in the dataset.
 * Mark the entry in the master dataset list as null.
 */
void dmiDatasetFree( dmDataset* ds )
{
 int i;
 card buf;
 dmsg( 1, "Dataset Free" );
 context->dmDatasetList[ds->no-1] = NULL;
 for ( i = 0; i < ds->nblocks; i++ ) {
  dmiBlockFree( ds->block[i] );
 }
 sprintf( buf, "*** FREE  %12ld dmDataset ***", (long)ds ); dmsg( 1, buf );
 free( ds->block );
 free( ds );
 dmsg( 1, "Dataset Free OK" );
}


/*
 * dmiDatasetGetBlock
 *
 * Look in the list of known blocks for a named block.
 */
dmBlock* dmiDatasetGetBlock( dmDataset* ds, char* name )
{
 int no;
 dmBlock* block;
 for ( no = 0; no < ds->nblocks; no++ ) {
  block = ds->block[no];
  if ( cs_eq( name, block->name ) ) {
   return( block );
  }
 }
 return( NULL );
}

/*
 * dmiBlockCreate
 *
 * Open the block, parsing its structure.
 *
 */
int dmiBlockCreate( dmBlock* block )
{
 int status = 0;
 dmKernel* kernel;
 kernel = block->ds->kernel;
 dmsg1( 1, "Block Create ", block->name );
 dmiBlockStructAlloc( block );
 block->current_row = 0;
 if ( !block->dss ) {
  status = ERR_MALLOC_FAIL;
  dmiErrorSet( status );
 } else {
/* Open all columns in table */
  if ( block->type ==  dmTABLE  ) {
   dmsg( 1, "Open columns" );
   block->ncols = 0;
   block->nrows = 0;
  } else {
   block->ds->kernel->dmkImageOpen( block->KernelBlock );
   block->ncols = 1;
   block->nrows = 1;
  }
 }
 return( status );
}

/*
 *
 * Set up block structure
 */
int dmiBlockOpen( dmBlock* block )
{
 int status = 0;
 t_word name;
 dmDataType type;
 t_word unit; 
 dmDescriptor* col;
 card desc;
 long i;
 int dim;
 long colno;
 dmKernel* kernel;
 dmsg1( 1, "dmiBlockOpen ", block->name );
 if( block->ds->nblocks > 1 ) {
   dmStatus( block->ds );
 }
 if ( block->key ) {
  return(0);
 }
 kernel = block->ds->kernel;
 dmiBlockStructAlloc( block );
 block->current_row = 1;
 if ( !block->dss ) {
  status = ERR_MALLOC_FAIL;
  dmiErrorSet( status );
 } else {
/* Open all columns in table */
  if ( block->type ==  dmTABLE  ) {
   kernel->dmkTableOpen( block->KernelBlock, &block->ncols, &block->nrows );
   for ( i = 0; i < block->ncols; i++ ) {
    colno = i+1;
    kernel->dmkColumnOpenNo( block->KernelBlock, colno, name, &type, unit, desc, &dim );
    col = dmiDescriptorAlloc( block, name, type, ddCOLUMN, colno );
    col->KernelNo = colno;
    block->column[i] = col; 
    cs_copy( unit, col->unit );
    cs_copy( desc, col->desc );
    col->data_type = type;
    if ( dim <= 1 ) {
     col->naxes = 0;
    } else {
     col->naxes = 1;
     col->axes[0] = dim;
    }
   }
  } else {
   block->ds->kernel->dmkImageOpen( block->KernelBlock );
   block->ncols = 1;
   block->nrows = 1;
  }
  dmiHeaderOpen( block );
 }
 return( status );
}


void dmiHeaderOpen( dmBlock* block )
{
 dmKernel* kernel;
 dmDataset* ds;
 dmDataType type;
 dmDescriptor* key;
 dmDescriptor* col;
 dmDescriptor* coord;
 card unit;
 int ncols;
 card desc;
 t_word name;
 int kno;
 int no;
 int offset;
 double* crpix;
 double* crval;
 double* cdelt;
 void* value;
 ds = block->ds;
 kernel = ds->kernel;
 kno = 0;
 no = 0;
 dmsg1( 1, "Open Header ", block->name );
 dmStatus( block->ds );
 while( kno >= 0 ) {
  kno = kernel->dmkNextKey( block->KernelBlock, kno, name, &type, unit, desc, &value );
  if ( kno >= 0 ) {
   no++;
   key = dmiDescriptorAlloc( block, name, type, ddKEY, no );
   block->key[no-1] = key;
   cs_copy( unit, key->unit );
   cs_copy( desc, key->desc );
   key->KernelNo = kno;
   key->value = value;
   key->tlmin = NULL;
   key->tlmax = NULL;
  }
 }
 block->nkeys = no;
/* Second pass */
 ncols = block->ncols;
 crpix = (double*)calloc( ncols, sizeof( double ) );
 crval = (double*)calloc( ncols, sizeof( double ) );
 cdelt = (double*)calloc( ncols, sizeof( double ) );
 kernel->dmkSpecialKey_d( block->KernelBlock, "TCRPX", ncols, crpix );
 kernel->dmkSpecialKey_d( block->KernelBlock, "TCRVL", ncols, crval );
 kernel->dmkSpecialKey_d( block->KernelBlock, "TCDLT", ncols, cdelt );
 for ( no = 1; no < ncols; no++ ) {
  offset = no - 1; 
  if ( crpix[offset] || crval[offset] || cdelt[offset] ) {
   dmsg( 1, "Create coord" );
   col = block->column[offset];
   coord =  dmiDescriptorAlloc( block, "COORD", dmDOUBLE, ddCOORD, 0 );
   col->coord = coord;
   dmiStoreCoord( coord, crpix[offset], crval[offset], cdelt[offset] );
  }
 }
 free ( crpix );
 free ( crval );
 free ( cdelt );
 dmsg( 1, "Done header" );
 dmStatus( ds );
}


/*
 * dmiBlockFind
 *
 * Locate a block by name. First look in the list of known blocks,
 * then scan forward in the file to see if you can find the block.
 * As you scan, fill in the list of known blocks. 
 * If the name is blank, open the next block whatever its name is.
 */

dmBlock* dmiBlockFind( dmDataset* ds, char* name )
{
 dmBlock* block;
 dmKernelBlock* kblock;
 dmBlockType block_type;
 t_word block_name;
/* Check for null dataset */
 if ( !ds ) return( NULL );
/* Is the block in our current list? */
 block = dmiDatasetGetBlock( ds, name );
 if ( block ) {
  ds->kernel->dmkMoveToBlock( block->KernelBlock ); 
 } else if ( !ds->scanned ) {
/* Need to move to max HDU here */
  dmsg( 1, "Search for block" );
  while( (kblock = ds->kernel->dmkBlockFindNext( ds->fptr, block_name, &block_type) ) != NULL ) {
   block = dmiBlockAlloc( ds, kblock, block_name, block_type );
   if ( cs_eq( name, block_name ) || cs_is_blank( name ) ) {
    return( block );
   }
  } 
  ds->scanned = TRUE;   /* Not found, so must have finished scanning the file */
  block = NULL;
 }
 return( block );
}

/*
 * dmiDatasetScan
 *
 * Scan forward through a dataset until you get to numbered block block_no.
 * First check if we have already found this block number.
 * If block_no is zero or negative, go to end of dataset.
 * Returns zero if completed scan.
 */
int dmiDatasetScan( dmDataset* ds, int block_no )
{
 dmBlock* block;
 dmKernelBlock* kblock;
 card block_name;
 dmBlockType block_type;

 if ( block_no <= ds->nblocks ) {
  return( block_no );
 } else if ( !ds->scanned ) {
  while( ( kblock = ds->kernel->dmkBlockFindNext( ds->fptr, block_name, &block_type) ) != NULL ) {
   block = dmiBlockAlloc( ds, kblock, block_name, block_type );
   if ( block_no > 0 && block->no >= block_no ) {
    return( block_no );
   }
  } 
  ds->scanned = TRUE;   /* Not found, so must have finished scanning the file */
 }
 return( 0 );
}


/*
 * Temporary routine to parse   foo[bar]
 *
 */
void dmiSyntaxParse( char* name, char* dsname, char* block_name )
{
 int j;
 int k;
 j = cs_index( name, "[" );
  
 if ( j > 0 ) {
  if ( j == 0 ) { 
   cs_copy( " ", dsname );
  } else {
   cs_get_ss( name, dsname, 1, j-1 );
  }
  k = cs_index( name, "]" );
  if ( k > 0 ) {
   cs_get_ss( name, block_name, j+1, k-1 );
  } else {
   cs_copy( name + j, block_name );
  }
 } else {
  cs_copy( name, dsname );
  cs_copy( " ", block_name );
 }
}




/*
 * dmiImageSpec
 *
 * Make image header, create image data descriptor
 */
void dmiImageSpec( dmBlock* block, dmDataType type, long* axes, long naxes )
{
 dmDescriptor* ImageData;
 dmDataset* ds;
 ds= block->ds;
 ds->kernel->dmkImageSpec( block->KernelBlock, type, axes, naxes );
 ImageData = dmiDescriptorAlloc( block, block->name, type, ddCOLUMN, 1 );
 ImageData->naxes = naxes;
 ar_copy_i( (int*)axes, ImageData->axes, naxes );   /* int/long declaration conflict here */
 block->column[0] = ImageData;
}



void dmiBlockPrintTable( dmBlock* block )
{
 int k;
 if ( !block ) return;
 msg( " " );
 for ( k = 0; k < block->ncols; k++ ) {
  dmiBlockPrintColumn( block->column[k] );
 }
}


void dmiBlockPrintHeader( dmBlock* block )
{
 int k;
 if ( !block ) return;
 printf( "NKEYS: %d\n", (int)block->nkeys );
 for ( k = 0; k < block->nkeys; k++ ) {
  dmiBlockPrintKey( block->key[k] );
 } 
}

void dmiBlockPrintColumn( dmDescriptor* key )
{
 t_buf buf;
 t_word typename;
 int j;
 card value;
 if( !key ) return;
 dmASCDataTypeName( key->data_type, typename );
 j = cs_ends( typename );
 if ( key->naxes > 0 ) {
  sprintf( typename+j, "[%d]", key->axes[0] );
 }
 sprintf( value, "COL %d", key->no );
 sprintf( buf, "%-8s %-16s %-20s %-20s %-20s", typename, key->name, value, key->unit, key->desc );
 msg( buf );
 
}

void dmiBlockPrintKey( dmDescriptor* key )
{
 t_buf buf;
 t_word typename;
 card value;
 if( !key ) return;
 dmASCDataTypeName( key->data_type, typename );
 dmiDescriptorFormatValue( key, value, CSIZE );
 sprintf( buf, "%-8s %-16s %-20s %-20s %-20s", typename, key->name, value, key->unit, key->desc );
 msg( buf );
}

void dmiDescriptorFormatValue( dmDescriptor* key, char* value, int size )
{
 switch( key->data_type ) {
  case dmSTRING:
   cs_copy_siz( *(char**)key->value, value, size );
   break;
  case dmLOGICAL:
   cs_copy( " ", value );
   value[0] = write_q( *(logical*)key->value );
   break;
  case dmLONG:
   write_i( *(int*)key->value, value, size );
   break;
  case dmDOUBLE:
   write_d( *(double*)key->value, value, size );
   break;
  default:
   cs_copy( " ", value );
 }
}

void dmiStoreCoord( dmDescriptor* coord, double crpix, double crval, double cdelt )
{
 dmTransform* t;

 t = coord->value;
 t->crpix = crpix;
 t->crval = crval;
 t->cdelt = cdelt;
}
