#include <stdlib.h>
#include <stdio.h>
#include "jcmlib.h"
#include "dmkfits.h"



/*
 * FITS_SetKernelOption
 *
 * Set the global flag which says what flavor of FITS table
 * is to be made on each table create call.
 */
int FITS_SetKernelOption( char* option )
{
 if ( cs_eq( option, "TABLE=BINTABLE" ) ) {
  FITS_SetTableType( FITS_BINTABLE );
 } else if ( cs_eq( option, "TABLE=ASCII" ) ) {
  FITS_SetTableType( FITS_TABLE );
 }
 return( 0 );
/* All other options silently ignored */
}

/*
 * FITS_DatasetOpen
 *
 * Allocates the FITS dataset structure and the 
 * keyword memory buffer.
 */

dmKernelDataset* FITS_DatasetOpen( char* filename ) 
{
 fitsfile* fptr;
 int status = 0;
 fits_open_file( &fptr, filename, READONLY, &status ); /* CFITSIO */
 if ( status ) {
  dmkErrorSet( status );
  fptr = NULL;
  return( NULL );
 }
 return( (dmKernelDataset*)FITS_DatasetAlloc( fptr ) );
}

 

/*
 * FITS_DatasetCreate
 * 
 *  FITS kernel Dataset Create routine
 * The CFITSIO routine fails if the file already exists,
 * so we check for this and delete it if needed.
 */
dmKernelDataset* FITS_DatasetCreate( char* filename ) 
{
 fitsfile* fptr;
 int status = 0;
 if ( fio_inquire( filename ) ) sys_delete( filename );
 fits_create_file( &fptr, filename, &status ); /* CFITSIO */
 if ( status ) {
  dmkErrorSet( status );
  return( NULL );
 }
 return( (dmKernelDataset*)FITS_DatasetAlloc( fptr ) );
}

/*
 * FITS_DatasetClose
 *
 * FITS kernel dataset close routine 
 * Closes file on disk, frees kernel structure.
 */

int FITS_DatasetClose( dmKernelDataset* kds )
{
 FITS_Dataset* ds = (FITS_Dataset*)kds;
 int status = 0;
 dmsg(1, "Close FITS file" );
 fits_close_file( ds->fptr, &status );   /* CFITSIO */
 FITS_DatasetFree( ds );
 if ( status ) dmkErrorSet( status );
 return( status );
}

/*
 * FITS_BlockOpenNext
 *
 * Go to one after the max HDU found so far.
 * Allocate the block structure and find the block's name.
 * Ignore null primary HDUs
 */

dmKernelBlock* FITS_BlockOpenNext( dmKernelDataset* kds, char* block_name, dmBlockType* block_type )
{
 FITS_Dataset* ds;
 FITS_Block* block;
 card buf;
 long naxis = 0;
 int status = 0;

 ds = (FITS_Dataset*)kds;
 block = FITS_BlockGetNext( ds, block_name, block_type );
 if ( block ) {
  if ( ds->nhdu == 1 ) { /* Is it a null primary HDU? */
   fits_read_key_lng( ds->fptr, "NAXIS", &naxis, buf, &status );
   if ( naxis == 0 ) { 
    dmsg( 1, "NULL primary: get another block" );
    block = FITS_BlockGetNext( ds, block_name, block_type );
   }
  }
 }  
 sprintf( buf, "FITS Block: %ld %s\n", (long)block, block_name );
 return( (dmKernelBlock*)block );
}

FITS_Block* FITS_BlockGetNext( FITS_Dataset* ds, char* block_name, dmBlockType* block_type )
{
 int hdutype;
 int status = 0;
 FITS_Block* block;
 card buf;

 sprintf( buf, "Move to HDU %d", (int)ds->nhdu+1 ); dmsg( 1, buf );
 fits_movabs_hdu( ds->fptr, ds->nhdu + 1, &hdutype, &status );  /* CFITSIO */
 if ( status ) {
  dmsg( 1, "FITS block move error" );
  dmkErrorSet( status );
  cs_copy( " ", block_name );
  *block_type = dmUNKNOWNBLOCK;
  return( NULL );
 } else {
  ds->nhdu++;
  ds->chdu = ds->nhdu;
  block = FITS_BlockAlloc( ds, (FITS_HDU_Type)hdutype, (int)ds->chdu, "R" );
  *block_type = FITS_GetBlockType( hdutype );
  FITS_GetBlockName( block, block_name );   /* Look up naming keywords */
  dmsg1( 1, "FITS: Allocated block for ", block_name );
  return( block );
 }
}

/*
 * BlockClose
 *
 * Update the header, and free the block structures.
 */
int FITS_BlockClose( dmKernelBlock* kblock )
{
 FITS_Block* block;
 int status = 0;
 block = (FITS_Block*)kblock;
 if ( block ) { 
  if( !block->init ) {
   dmsg( 1, "Flush init" );
   FITS_BlockInit( block );            /* Flush table/image header cards to CFITSIO */
  }
  if( block->type != FITS_IMAGE ) {  /* Update number of rows */
   fits_modify_key_lng( block->ds->fptr, "NAXIS2", block->nrows, "Number of rows", &status );
  }
  FITS_BlockFree( block );
 }
 return( 0 );
}


/*
 * FITS_BlockCreate
 *
 * Create a new FITS HDU. If the HDU is a table, and it is the first HDU,
 * create a null primary header first. No FITS Block is made for the null
 * primary header, which needs to be fixed (inconsistent with reading).
 * In the new HDU, no data is written as we want to wait and define
 * some table columns or image axes first. (see BlockInit).
 * Space for the kernel block is allocated. 
 * Space for up to 100 columns is allocated.
 */
dmKernelBlock* FITS_BlockCreate( dmKernelDataset* kds, char* block_name, dmBlockType block_type )
{
 int status=0;
 int type;
 const int NCOLS = 100;
 FITS_Block* block;
 FITS_Dataset* ds;
 long naxes[2] = { 0, 0 };
 ds = (FITS_Dataset*)kds;
 type = FITS_SetBlockType( block_type );

 if ( type != FITS_IMAGE && ds->nhdu == 0 ) {
/* Null primary HDU */
  fits_create_img( ds->fptr, BITPIX_BYTE, 0, naxes, &status );  /* CFITSIO */
  ds->nhdu++;
  ds->chdu = ds->nhdu;
 }
 fits_create_hdu( ds->fptr, &status );        /* CFITSIO */
 ds->nhdu++;
 ds->chdu = ds->nhdu;
 block = FITS_BlockAlloc( ds, type, (int)ds->chdu, "W" );
 cs_copy( block_name, block->name );
 FITS_AllocCols( block, NCOLS );
 return( (dmKernelBlock*)block );
}


/*
 * Read a named special key and return its value(s) in an array.
 * The key may be indexed, in which case at most ncols values
 * will be returned.
 * This mechanism is not ideal! I would rather handle the
 * special keys at the FITS level, but that means repeated caching
 * of information.
 *
 */
int FITS_GetSpecialKey_d( dmKernelBlock* kblock, char* name, int ncols, double* value )
{
 FITS_Block* block = (FITS_Block*)kblock;
 int k;
 FITS_Key* key;
 int offset;
 int n = 0; /* Number found */
 ar_zero_d( value, ncols );
 for ( k = 1; k <= block->nkeys; k++ ) {
  key = &(block->keys[k-1]);
  if ( key->meta ) { 
   if( cs_uceq( name, key->root ) && key->index <= ncols ) {
    offset = key->index-1;
    
    if ( key->type == FITS_KEY_D ) {
     value[offset] = *(double*)key->value;
    } else if ( key->type == FITS_KEY_I ) {
     value[offset] = (double)*(long*)key->value;
    }
    n++;
   }
  }  
 }  
 return( n );
}


/*
 * Allocate DM keys
 * On input, give the FITS key number last found (0 to start off).
 * This is incremented until a key is found  which is not a special key
 * or a comment.
 * Return its data type, unit, description, and a pointer to its value.
 */

int FITS_NextKey( dmKernelBlock* kblock, int no, char* name, dmDataType* type, 
  char* unit, char* desc, void** value )
{

 FITS_Key* key;
 FITS_Block* block = (FITS_Block*)kblock;
 while( no < block->nkeys ) { 
  no++;
  key = &(block->keys[no-1]);
  if ( !key->meta && key->type != FITS_COMMENT ) {
   cs_copy( key->name, name );
   cs_copy( " ", unit );
   cs_copy( key->comment, desc );
   *type = FITS_KeyTypeToType( key->type );
   *value = key->value;
   return( no );
  }
 }
 return( -1 );
}




/*
 * FITS_ImageOpen
 *
 * Open an existing image. Read its header.
 * Need to add code to find the image dimensions and axes.
 */

int FITS_ImageOpen( dmKernelBlock* kblock )
{
 FITS_Block* block;
 block = (FITS_Block*) kblock;
 FITS_HeaderOpen( block );
 return( 0 );
}
/*
 * FITS_TableOpen
 *
 * Open an existing table. Read its header, and allocate space
 * for the columns.
 */

int FITS_TableOpen( dmKernelBlock* kblock, long* ncols, long* nrows )
{
 FITS_Block* block;
 int status = 0;
 fitsfile* fptr;
 card extname;
 int tval;
 block = (FITS_Block*) kblock; 
 *ncols = 0;
 *nrows = 0;
 FITS_HeaderOpen( block );
 dmsg( 1, "Search Columns" );
 if ( block->type == FITS_BINTABLE ) {
  fptr = block->ds->fptr;
  fits_read_key_lng( fptr, "TFIELDS", &block->tfields, NULL, &status ); /* CFITSIO */
  FITS_AllocCols( block, block->tfields );
  fits_read_btblhdr( block->ds->fptr, (int)block->tfields, &block->nrows, &tval, block->ttype, block->tform, block->tunit, extname, &block->pcount, &status ); /* CFITSIO */
  if ( status ) dmkErrorSet( status );
  *ncols = block->tfields;
  *nrows = block->nrows;
 }
 return( status );
}

/*
 *  Make CFITSIO move to a specified HDU.
 *  Update the dataset current HDU pointer.
 */
int FITS_MoveToBlock( dmKernelBlock* kblock )
{ 
 int status;
 int hdutype;
 FITS_Block* block;
 FITS_Dataset* ds;
 card buf;
 block = (FITS_Block*)kblock;
 ds = block->ds; 
 sprintf( buf, "FITS: Move to HDU %d\n", block->hdu );
 dmsg( 1, buf );
 fits_movabs_hdu( ds->fptr, block->hdu, &hdutype, &status ); /* CFITSIO */
 ds->chdu = block->hdu;
 if ( status ) dmkErrorSet( status );
 return( status );
}

/*
 * Specify the data type and dimensions of an image.
 * Write out the header keywords; we calculate BITPIX from the
 * data type and call the block init routine to write the keywords.
 * 
 */

int FITS_ImageSpec( dmKernelBlock* kblock, dmDataType type, long* axes, long naxes )
{
 int status;
 FITS_Block* block;
 FITS_Dataset* ds;
 block = (FITS_Block*)kblock;
 ds = block->ds;
 block->bitpix = FITS_TypeToBitpix( type );
 block->naxes = naxes;
 ar_copy_i( (int*)axes, (int*)block->axes, (int) naxes );
 status = FITS_BlockInit( block );
 return( status );
}

/*
 * Open a numbered column.
 * Return its name, data type, unit, description (currently blank),
 * and array dimension.
 * The FITSIO calls have already been made when the block was opened,
 * so this call just returns the cached values.
 */


int FITS_ColumnOpenNo( dmKernelBlock* kblock, long no, char* name, dmDataType* type, char* unit, char* desc, int* dim )
{
 int offset;
 long repeat;
 long width;
 FITS_Block* block = (FITS_Block*)kblock;
 if ( !block ) return( 0 );
 if ( !block->ttype ) return( 0 );
 offset = no - 1;
 if ( no > 0 && no <= block->tfields ) {
  cs_copy( block->ttype[offset], name );
  cs_copy( block->tunit[offset], unit );
  cs_copy( " ", desc );
  FITS_ColumnParseType( block->tform[offset], type, &repeat, &width );
  block->repeat[offset] = repeat;
  block->dtype[offset] = *type;
  *dim = repeat;
 }
 return( 0 );
}

/*
 * Read data from a table column.
 * Handles multiple rows and array columns.
 */
long FITS_ColumnRead_d( dmKernelBlock* kblock, int colno, long row, long nrows, double* result, long npixels )
{
 int offset;
 int null; 
 long nelements;
 card buf;
 real x;
 int i;
 int status = 0;
 dmDataType type;
 FITS_Block* block = (FITS_Block*)kblock;
 if ( colno > 0 && colno <= block->tfields ) {
  offset = colno - 1;
  type = block->dtype[offset];
  nelements = npixels * nrows;
  if ( type == dmDOUBLE ) {
   fits_read_col_dbl( block->ds->fptr, colno, row, 1, nelements, 0, result, &null, &status );  /* CFITSIO */
  } else if ( type == dmFLOAT ) {
   for( i = 1; i <= nelements; i++ ) {
    fits_read_col_flt( block->ds->fptr, colno, row, i, 1, 0, &x, &null, &status );  /* CFITSIO */
    result[i-1] = (double)x;
   }
  } else {
   tmsg( "type mismatch" );
   ar_zero_d( result, npixels );
   nrows = 0;
  }
 } else { 
  sprintf( buf, "Warning: attempt to read from bad FITS column no. %d\n", (int)colno );
  tmsg( buf );
 }
 return( nrows );
}

long FITS_ColumnRead_r( dmKernelBlock* kblock, int colno, long row, long nrows, float* result, long npixels )
{
 int offset;
 int null; 
 long nelements;
 card buf;
 double x;
 int i;
 int status = 0;
 dmDataType type;
 FITS_Block* block = (FITS_Block*)kblock;
 if ( colno > 0 && colno <= block->tfields ) {
  offset = colno - 1;
  type = block->dtype[offset];
  nelements = npixels * nrows;
  if ( type == dmFLOAT ) {
   fits_read_col_flt( block->ds->fptr, colno, row, 1, nelements, 0, result, &null, &status );  /* CFITSIO */
  } else if ( type == dmFLOAT ) {
   for( i = 1; i <= nelements; i++ ) {
    fits_read_col_dbl( block->ds->fptr, colno, row, i, 1, 0, &x, &null, &status );  /* CFITSIO */
    result[i-1] = (float)x;
   }
  } else {
   tmsg( "type mismatch" );
   ar_zero_r( result, npixels );
   nrows = 0;
  }
 } else { 
  sprintf( buf, "Warning: attempt to read from bad FITS column no. %d\n", (int)colno );
  tmsg( buf );
 }
 return( nrows );
}


int FITS_ColumnCreate( dmKernelBlock* kblock, char* name,  dmDataType type, int slen, char* unit, char* desc )
{
 FITS_Block* block = (FITS_Block*)kblock;
 int offset;
 int colno;
 int repeat = 1;
 t_word desc2;
 if ( !block ) return( 0 );
 if ( !block->ttype ) return( 0 );

 colno = ++block->tfields;
 offset = colno - 1;
 cs_copy_siz( name, block->ttype[offset], WSIZE );
 cs_copy_siz( unit, block->tunit[offset], WSIZE );
 cs_copy_siz( desc, desc2, WSIZE );
 FITS_ColumnComposeType( type, repeat, slen, block->tform[offset] );
 block->repeat[offset] = repeat;
 block->dtype[offset] =  type;

 return( colno );
}

long FITS_ColumnWrite_d( dmKernelBlock* kblock, int colno, long row, long nrows, double* value, long npixels )
{
 FITS_Block* block = (FITS_Block*)kblock;
 int offset;
 card buf;
 dmDataType type;
 long nelements;
 int status = 0;
 if ( !block ) return( 0 );
 if ( colno > 0 && colno <= block->tfields ) {
  offset = colno - 1;
  type = block->dtype[offset];
  if ( type == dmDOUBLE ) {
   nelements = npixels * nrows;
   fits_write_col_dbl( block->ds->fptr, colno, row, 1, nelements, value, &status );  /* CFITSIO */
  } else {
   tmsg( "type mismatch" );
   nrows = 0;
  }
 } else { 
  sprintf( buf, "Warning: attempt to write to bad FITS column no. %d\n", (int)colno );
  tmsg( buf );
 }

 return( nrows );
}



/*
 * FITS_BlockInit
 *
 * Write header cards to disk. This call is delayed until
 * all the columns or image axes have been defined.
 * It will be made on the first data write, or on block close
 * if there is no data. The block init flag is set to mark that
 * the call has been made.
 */
int FITS_kBlockInit( dmKernelBlock* kblock )
{
 return( FITS_BlockInit( (FITS_Block*)kblock ) );
}

int FITS_BlockInit( FITS_Block* block )
{
 FITS_Dataset* ds;
 int status = 0;

 if ( !block ) return( 0 );
 ds = block->ds; 
 /* Flush header to disk */
 if ( block->type == FITS_IMAGE ) {
  fits_write_imghdr( ds->fptr, block->bitpix, block->naxes, block->axes, &status );  /* CFITSIO */
  if ( status ) dmkErrorSet( status );
 } else if ( block->type == FITS_BINTABLE ) {
  dmsg( 1,"Write Table header" );
  fits_write_btblhdr( ds->fptr, block->nrows, block->tfields, block->ttype, /* CFITSIO */
    block->tform, block->tunit, block->name, block->pcount, &status ); /* CFITSIO */
 }
 dmsg( 1, "Init flag cleared after flush" );
 block->init = TRUE;
 return( status );
}



int FITS_PutRow(  dmKernelBlock* kblock, long row )
{
 FITS_Block* block = (FITS_Block*)kblock;
 if ( block ) {
  if ( block->nrows == 0 )  FITS_BlockInit( block );
  block->nrows = max_i( block->nrows, row );
 }
 return( block->nrows );
}


int FITS_KeyWrite_d( dmKernelBlock* kblock, char* name, double value, char* unit, char* desc )
{
 FITS_Block* block = (FITS_Block*)kblock;
 int status = 0;
 int decimals = 8;
 card cdesc;
 if ( !cs_is_blank( unit ) ) {
  sprintf( cdesc, "[%s] %s", unit, desc );
 } else {
  cs_copy_siz( desc, cdesc, CSIZE );
 }
 fits_write_key_dbl( block->ds->fptr, name, value, decimals, cdesc, &status );
 return( 0 );
}

