/*
 * dmkfits.h
 *
 * FITS Kernel for JCM data model
 *
 */


#ifndef DMKFITS_H
#define DMKFITS_H
#include "fitsio.h"
#include "ascdm.h"
#include "dmklib.h"
#include "store.h"

/* The FITS HDU TYPE enum is chosen to be compatible with the CFITSIO hdutype argument */
typedef enum { FITS_IMAGE, FITS_TABLE, FITS_BINTABLE, FITS_UNKNOWN } FITS_HDU_Type;
typedef enum { FITS_KEY_D, FITS_KEY_I, FITS_KEY_C, FITS_KEY_Q, 
               FITS_COMMENT, FITS_KEY_UNK } FITS_KeyType;
typedef enum { BITPIX_BYTE = 8, BITPIX_SHORT = 16, BITPIX_INT = 32, BITPIX_REAL = -32, BITPIX_DOUBLE = -64 } BITPIX_Type;

typedef struct {
 fitsfile* fptr;
 int nhdu;  /* Max HDUs found so far */
 int chdu;  /* Current HDU */
 Store* store;
} FITS_Dataset;

typedef struct {
 char name[10];
 char root[10];
 int index;
 FITS_KeyType type;
 void* value;
 card comment;
 logical meta;
} FITS_Key;

typedef struct {
 int hdu;  /* HDU number */
 FITS_HDU_Type type; /* FITS HDU type */
 FITS_Dataset* ds;
 FITS_Key* keys;
 card name;
 long nkeys;
 long nrows;
 long tfields;
 char** ttype;
 char** tunit;
 char** tform;
 dmDataType* dtype;
 int* repeat;
 long pcount;
 int naxes;
 long axes[20];
 int bitpix;
 logical init;  /* Has header been flushed to disk? */
} FITS_Block;


dmKernelDataset* FITS_DatasetOpen( char* filename );
dmKernelDataset* FITS_DatasetCreate( char* filename );
int FITS_DatasetClose( dmKernelDataset* ds );
dmKernelBlock* FITS_BlockOpenNext( dmKernelDataset* ds, char* block_name, dmBlockType* );
dmKernelBlock* FITS_BlockCreate( dmKernelDataset* kds, char* block_name, dmBlockType block_type );
int FITS_MoveToBlock( dmKernelBlock* kblock );
int FITS_ImageSpec( dmKernelBlock* kblock, dmDataType type, long* axes, long naxes );
int FITS_BlockClose( dmKernelBlock* kblock );
int FITS_ColumnCreate( dmKernelBlock* kblock, char* name,  dmDataType type, int slen, char* unit, char* desc );
int FITS_KeyWrite_d( dmKernelBlock* block, char* name, double value, char* unit, char* desc );

int FITS_TableOpen( dmKernelBlock* kblock, long* ncols, long* nrows );
int FITS_ImageOpen( dmKernelBlock* );
int FITS_ColumnOpenNo( dmKernelBlock*, long no, char* name, dmDataType* type, char* unit, char* desc, int* dim );
long FITS_ColumnRead_d( dmKernelBlock*, int, long, long, double*, long );
long FITS_ColumnRead_r( dmKernelBlock*, int, long, long, float*, long );
long FITS_ColumnWrite_d( dmKernelBlock* kblock, int colno, long row, long nrows, double* value, long npixels );
int FITS_kBlockInit( dmKernelBlock* kblock );
int FITS_GetSpecialKey_d( dmKernelBlock* block, char* name, int no, double* value );
int FITS_NextKey( dmKernelBlock* kblock, int, char* name, dmDataType* type,   char* unit, char* desc, void** value );
int FITS_SetKernelOption( char* option );
int FITS_PutRow(  dmKernelBlock*, long );

/*
 * FITS internal routines
 */
FITS_Block* FITS_BlockGetNext( FITS_Dataset* ds, char* block_name, dmBlockType* block_type );
FITS_Dataset* FITS_DatasetAlloc( fitsfile* fptr );
void FITS_DatasetFree( FITS_Dataset* ds );
void FITS_AllocCols( FITS_Block* block, int tfields );
void FITS_FreeCols( FITS_Block* block );
FITS_Block* FITS_BlockAlloc( FITS_Dataset* ds, FITS_HDU_Type hdutype, int hdu, char* mode );
void FITS_BlockFree( FITS_Block* block );
int FITS_HeaderOpen( FITS_Block* block );
void FITS_HeaderClose( FITS_Block* block );
int FITS_BlockInit( FITS_Block* block );
void FITS_SpecialKeys( FITS_Block* block );
void FITS_GetBlockName( FITS_Block* block, char* name );
FITS_KeyType FITS_ReadKeyNo( fitsfile* fptr, long no, char* name, char* value, char* comment );



/*
 * FITS kernel diagnostic routines
 */

int FITS_BlockPrintKernel( dmKernelBlock* block );
void FITS_PrintKey( FITS_Key* key );

/*
 * FITS auxiliary routines
 */
dmDataType FITS_BitpixToType( int bitpix );
int FITS_TypeToBitpix( dmDataType type );
void FITS_ColumnParseType( char* tform, dmDataType* type, long* repeat,
 long* width );
void FITS_SetTableType( FITS_HDU_Type type );
dmDataType FITS_KeyTypeToType( FITS_KeyType c );
FITS_KeyType FITS_PenceKeyTypeToKeyType( char c );
dmBlockType FITS_GetBlockType( int hdutype );
int FITS_SetBlockType( dmBlockType fits_block_type );
void FITS_StripKeyIndex( char* name, char* root, int* index );
void FITS_ColumnComposeType( dmDataType type, int repeat, int width, char* tform );

#endif

