#include <stdlib.h>
#include <stdio.h>
#include "jcmlib.h"
#include "dmkfits.h"

/* Routines in this file are internal to the FITS kernel */

/*
 * Allocate space for the FITS dataset info and the memory store
 * used for keyword values.
 */

FITS_Dataset* FITS_DatasetAlloc( fitsfile* fptr )
{
 card buf;
 FITS_Dataset* ds;
 if ( fptr ) {
  ds = (FITS_Dataset*)malloc( sizeof( FITS_Dataset ) ); 
  sprintf( buf, "*** ALLOC %12ld FITS DS ***", (long)ds ); dmsg(1, buf );
  ds->fptr = fptr;            /* Save CFITSIO handle */
  ds->nhdu = 0;
  ds->chdu = 0;               /* Current HDU is zero */
  ds->store = store_alloc();  /* Keyword memory buffer */
 } else {
  ds = NULL;
 }
 return( ds );
}

/*
 * Release memory for the FITS Dataset.
 */
void FITS_DatasetFree( FITS_Dataset* ds )
{
 card buf;
 dmsg(1, "Free store" );
 store_free( ds->store );		/* Free keyword memory */
 sprintf( buf, "*** FREE  %12ld FITS DS ***", (long)ds ); dmsg(1, buf );
 free( ds );				/* Free dataset */
 dmsg(1, "Free FITS DS OK" );
}


/*
 * Allocate space for table columns in the kernel block.
 * On read, specify the number of columns. On write, set tfields = 0
 * and we reserve 100 cols by default. (or you can explicitly set the tfields).
 */

void FITS_AllocCols( FITS_Block* block, int tfields )
{
 card buf;
 if ( tfields == 0 ) tfields = 100;
 block->ttype = alloc_array_c( WSIZE, tfields );
 block->tunit = alloc_array_c( WSIZE, tfields );
 block->tform = alloc_array_c( 12, tfields );
 block->repeat = (int*)malloc( tfields * sizeof( int ) );
 block->dtype = (dmDataType*)malloc( tfields * sizeof( dmDataType ));
 sprintf( buf, "*** ALLOC %12ld FITS Table ***", (long)block->ttype ); dmsg(1, buf );
}

/*
 * Release space allocated for table columns in FITS block.
 */

void FITS_FreeCols( FITS_Block* block )
{
 card buf;
 if( block->ttype ) {
  sprintf( buf, "*** FREE  %12ld FITS Table ***", (long)block->ttype ); dmsg(1, buf ); 
  free_array_c( block->ttype );
  free_array_c( block->tunit );
  free_array_c( block->tform );
  free( block->repeat );
  free( block->dtype );
  block->ttype = NULL;
  block->tunit = NULL;
  block->tform = NULL;
 }
}

/*
 *
 * Mode is "R" for an existing (on disk) block, "W" for a newly created one.
 *
 */
FITS_Block* FITS_BlockAlloc( FITS_Dataset* ds, FITS_HDU_Type hdutype, int hdu, char* mode )
{
 FITS_Block* block;
 card buf;
 block = (FITS_Block*)malloc( sizeof( FITS_Block) );    /* Alloc kernel block */
 sprintf( buf, "*** ALLOC %12ld FITS BLOCK %d ***", (long)block, (int)hdu ); dmsg(1, buf );
 block->type = hdutype;
 block->hdu =  hdu;
 block->keys = NULL;
 block->ds = ds;
 if ( cs_eq( mode, "R" ) ) {
  dmsg( 1, "Init flag cleared on open" );
  block->init = TRUE;
 } else {
  dmsg( 1, "Init flag set on create" );
  block->init = FALSE;
 } 
 /* Some default table values which may be overridden */
 block->pcount = 0;
 block->nrows = 0;
 block->tfields = 0;
 /* Some default image values which may be overridden */
 block->bitpix = BITPIX_BYTE;
 block->naxes = 2;
 block->axes[0] = 0;
 block->axes[1] = 0;
 /* Null out pointers */
  block->ttype = NULL;
  block->tunit = NULL;
  block->tform = NULL;

 return( block );
}

/*
 * Free memory for block
 */
void FITS_BlockFree( FITS_Block* block )
{
 card buf;
 if( block ) {
  FITS_HeaderClose( block );
  dmsg( 1, "Free cols" );
  FITS_FreeCols( block );
  dmsg( 1, "Free table ok" );
  sprintf( buf, "*** FREE  %12ld FITS BLOCK %d ***", (long)block, (int)block->hdu ); dmsg(1, buf );
  free( block );				/* Free kernel block */
 }
}

/*
 * Read a FITS header, caching the information in memory.
 */
int FITS_HeaderOpen( FITS_Block* block )
{
 fitsfile* fptr;
 int nkeys;
 int pos;
 int k;
 t_word name;
 card value;
 card comment;
 double dvalue;
 int lvalue;
 long ivalue;
 FITS_Key* key;
 Store* store;
 card buf;
 int hdutype;
 int status = 0;

 fptr = block->ds->fptr;    /* The FITS file */
 store = block->ds->store;  /* The memory store */
 fits_movabs_hdu( fptr, block->hdu, &hdutype, &status );
 fits_get_hdrpos( fptr, &nkeys, &pos, &status ); /* CFITSIO */
 block->keys = (FITS_Key*)malloc( nkeys * sizeof( FITS_Key ) );  /* Alloc key buffers */
 sprintf( buf, "*** ALLOC %12ld FITS HEADER TO BLOCK %12ld NKEYS %d ***", (long)block->keys, (long)block, (int)nkeys ); dmsg(1, buf );
 block->nkeys = nkeys;
 /* Loop over keys storing each appropriately depending on its type. */
 /* We read each key twice because we need to first get the key's type and
  * then read its value.
  */

 for ( k = 1; k <= nkeys; k++ ) {   
  key =&block->keys[k-1];
  key->type = FITS_ReadKeyNo( fptr, k, name, value, key->comment );
  cs_copy( name, key->name );
  switch( key->type ) {
    case FITS_KEY_C:
     fits_read_key_str( fptr, name, value, comment, &status ); /* CFITSIO */
     key->value = store_c( store, value );
     break;
    case FITS_COMMENT:
     key->value = store_c( store, value );
     break;
    case FITS_KEY_Q:
     fits_read_key_log( fptr, name, &lvalue, comment, &status ); /* CFITSIO */
     key->value = store_q( store, &lvalue, 1 );
     break;
    case FITS_KEY_I:
     fits_read_key_lng( fptr, name, &ivalue, comment, &status ); /* CFITSIO */
     key->value = store_i( store, &ivalue, 1 );
     break;
    case FITS_KEY_D:
     fits_read_key_dbl( fptr, name, &dvalue, comment, &status ); /* CFITSIO */
     key->value = store_d( store, &dvalue, 1 );
     break;
    default:
     tmsg( "Unknown key type" );
     tmsg( value );
  }  
 }
/*
 * Mark a subset of the keys as special, and parse indexed keyword names 
 */
 FITS_SpecialKeys( block );
 return(0);
}

/*
 * Release memory for kernel block keys
 */
void FITS_HeaderClose( FITS_Block* block )
{
 card buf;
 if ( block->keys ) {
  sprintf( buf, "*** FREE  %12ld FITS HEADER ***", (long)block->keys ); dmsg(1, buf );
  free( block->keys );
  block->keys = NULL;
 }
}

/*
 * Mark some FITS keywords as special (meta flag).
 * They will not be returned as data model keys.
 * Also, parse indexed keywords.
 */
void FITS_SpecialKeys( FITS_Block* block )
{
 int nspecial = 7;
 int nindex = 10;
 int index;
 t_word root;
 char* special[] = { "BITPIX", "NAXIS", "PCOUNT", "GCOUNT", "XTENSION", "TFIELDS", "EXTNAME" };
 char* special_index[]= { "TTYPE", "TFORM", "NAXIS", "TUNIT", "TCRPX", "TCRVL", "TCDLT", "CRPIX", "CRVAL", "CDELT" };
 int i;
 card name;
 FITS_Key* key;
 int k;
 for ( k = 0; k < block->nkeys; k++ ) {
  key = &(block->keys[k]);
  cs_copy( key->name, name );
  FITS_StripKeyIndex( name, root, &index );
  cs_copy( root, key->root );
  key->index = index;
  i = match_c( name, special, nspecial );
  if ( i == 0 && index > 0 ) {
   i = match_c( root, special_index, nindex );
  }
  if ( i > 0 ) {
   key->meta = TRUE;
  } else {
   key->meta = FALSE;
  }
/* Fix broken rosat files */
  if ( block->type != FITS_IMAGE ) {
   if ( cs_eq( key->root, "CRPIX" ) ) {
    cs_copy( "TCRPX", key->root );
   } else if ( cs_eq( key->root, "CRVAL" ) ) {
    cs_copy( "TCRVL", key->root );
   } else if ( cs_eq( key->root, "CDELT" ) ) {
    cs_copy( "TCDLT", key->root );
   }
  }
 }
}




/*
 * Read a numbered header key. This wrapper for CFITSIO's fits_read_keyn
 * checks for COMMENT and HISTORY records (which read_keyn does not handle)
 * and rereads them if needed. It also determines the data type of the key.
 */

FITS_KeyType FITS_ReadKeyNo( fitsfile* fptr, long no, char* name, char* value, char* comment )
{
 FITS_KeyType type;
 int status = 0;
 char dtype; 
 card cardval;
 fits_read_keyn( fptr, no, name, value, comment, &status ); /* CFITSIO */
 if ( cs_eq( name, "COMMENT" ) || cs_eq( name, "HISTORY" ) ) {
  type = FITS_COMMENT;
  fits_read_record( fptr, no, cardval, &status ); /* CFITSIO */
  cs_copy( cardval+9, value );
  cs_copy( " ", comment );
 } else {
  fits_get_keytype( value, &dtype, &status ); /* CFITSIO */
  type = FITS_PenceKeyTypeToKeyType( dtype );
 }
 return( type );
}


/*
 * FITS_GetBlockName
 *
 * Get name of block from the EXTNAME keyword. If a primary array and no EXTNAME,
 * set the name to be 'PRIMARY'.
 */
void FITS_GetBlockName( FITS_Block* block, char* name )
{
 FITS_Dataset* ds;
 int status = 0;

 ds = block->ds;
 fits_read_key_str( ds->fptr, "EXTNAME", name, NULL, &status ); /* CFITSIO */
 if ( status == KEY_NO_EXIST ) {
  if ( block->hdu == 1 ) {
   cs_copy( "PRIMARY", name );
  } else {
   cs_copy( " ", name );
  }
 } else if ( status != 0 ) {
  cs_copy( " ", name );
  dmkErrorSet( status );
 }
 cs_copy( name, block->name );  /* Cache result here too */
}


