#include <stdlib.h>
#include <stdio.h>
#include "jcmlib.h"
#include "dmkfits.h"

static FITS_HDU_Type FITS_TABLE_FLAG = FITS_BINTABLE;

void FITS_SetTableType( FITS_HDU_Type type )
{
  FITS_TABLE_FLAG = type;
}

void FITS_ColumnComposeType( dmDataType type, int repeat, int width, char* tform )
{
 char c;
 int w = 1;
 switch( type ) {
  case dmSHORT:   c = 'I';  break;
  case dmLONG:    c = 'J';  break;
  case dmLOGICAL: c = 'L';  break;
  case dmFLOAT:   c = 'E';  break;
  case dmDOUBLE:  c = 'D';  break;
  case dmSTRING:  c = 'A'; w = width;  break;
  default: c = 'A';
 }
 w *= repeat;
 sprintf( tform, "%d%c", w, c );
}

/*
 * Parse the TFORM keyword. We use Bill's routine, then convert his
 * data types to ours. BIT is not handled correctly, neither is COMPLEX.
 */
void FITS_ColumnParseType( char* tform, dmDataType* type, long* repeat,
 long* width )
{
 int typecode;
 int status = 0;
 fits_binary_tform( tform, &typecode, repeat, width, &status ); /* CFITSIO */
 switch( typecode ) {
  case 1:   *type = dmBYTE; break;
  case 11:  *type = dmBYTE; break;
  case 14:  *type = dmLOGICAL; break;
  case 16:  *type = dmSTRING; break;
  case 21:  *type = dmSHORT; break;
  case 41:  *type = dmLONG; break;
  case 42:  *type = dmFLOAT; break;
  case 82:  *type = dmDOUBLE; break;
  case 83:  *type = dmVOID; break; /* complex */
  case 163: *type = dmVOID; break; /* double complex */
  default:  *type = dmVOID;
 }
}


/*
 * Find data type of key, converting value returned from Bill's routine
 * to our enumerated type.
 */

FITS_KeyType FITS_PenceKeyTypeToKeyType( char c )
{
 FITS_KeyType type;
 switch( c ) {
    case 'C':
     type = FITS_KEY_C;
     break;
    case 'L':
     type = FITS_KEY_Q;
     break;
    case 'I':
     type = FITS_KEY_I;
     break;
    case 'F':
     type = FITS_KEY_D;
     break;
    default:
     type = FITS_KEY_UNK;
     tmsg( "Unknown key type" );
 }
 return( type );
}

/*
 * Convert data type to BITPIX value
 */
int FITS_TypeToBitpix( dmDataType type )
{
 const unsigned int nvals = 11;
 int bitpix = 0;
 int bitpix_vals[] = { 16, 32, -32, -64, 0, 0, 0, 8, 16, 32, 0 };
 if ( type < nvals ) {
  bitpix = bitpix_vals[type]; 
 } else {
  bitpix = 0;
  tmsg( "Bad data type" );
 }
 return( bitpix );
}

dmDataType FITS_BitpixToType( int bitpix )
{
 dmDataType t;
 switch( bitpix ) {
  case  16: t = dmSHORT; break; 
  case  32: t = dmLONG; break;
  case -32: t = dmFLOAT; break;
  case -64: t = dmDOUBLE; break;
  case   8: t = dmBYTE; break;
  case   0: t = dmVOID; break;
  default:  t = dmVOID; 
 }
 return( t );
}

/*
 * Convert from our FITS keyword types to our DM datatypes.
 * Comments are a special case.
 */
dmDataType FITS_KeyTypeToType( FITS_KeyType ktype )
{
 dmDataType type;
 switch( ktype ) {
  case FITS_KEY_C:
   type = dmSTRING;
   break;
  case FITS_KEY_Q:
   type = dmLOGICAL;
   break;
  case FITS_KEY_D:
   type = dmDOUBLE;
   break;
  case FITS_KEY_I:
   type = dmLONG;
   break;
  case FITS_COMMENT:
  default:
   type = dmVOID;
   break;
 }
 return( type );
}


/*
 * FITS_StripKeyIndex
 *
 * FITS indexed keywords are of the form  AAAAnnn
 * (text and numbers with trailing numbers). We look for the
 * trailing numbers and strip them off.
 */
void FITS_StripKeyIndex( char* name, char* root, int* index )
{
 int i;
 int k;
 k = cs_ends( name );
 i = k;
 *index = 0;
 while( digit_char( name[i-1] ) && i > 1 ) i--;
/* k is end of string, i is end of root */
 cs_get_ss( name, root, 1, i );
 if ( i < k )  *index = read_i( name+i );
}


/*
 * Convert from data model block type to 
 * FITS block type. The global variable FITS_TABLE_FLAG
 * is used to determine whether a binary or ASCII table
 * is made.
 */

int FITS_SetBlockType( dmBlockType type )
{
 int itype;
 switch ( type ) {
  case dmIMAGE:
   itype = FITS_IMAGE;
   break; 
  case dmTABLE:
   itype = FITS_TABLE_FLAG;
   break; 
  default:
   itype = FITS_UNKNOWN;
 }
 return( itype );
}

/*
 * FITS_GetBlockType
 *
 * Convert from our FITS block type (which is compatible with Bill's)
 * to data model block type. Both TABLE and BINTABLE map to a data model TABLE.
 */
dmBlockType FITS_GetBlockType( int hdutype )
{
 dmBlockType type;
 switch( hdutype ) {
  case FITS_IMAGE:
   type = dmIMAGE;
   break;
  case FITS_TABLE:
  case FITS_BINTABLE:
   type = dmTABLE;
   break;
  default:
   type = dmUNKNOWNBLOCK;
 }
 return( type );
}


