/*======================================================================*/
/*
 * Math routines
 *
 * <math.h>:  	log  used in logbase_d
 * 		modf used in frac_d_std
 * <float.h>:	DBL_EPSILON used in frac_d
 */
/*======================================================================*/

#include <float.h>
#include <math.h>
#include "jcmlib.h"


#define FUNC_TOL 1.0e-5

integer ceiling_d( const double x )
{
 integer i;
 i = (integer)x;
 if ( x > (double)i ) i++;
 return( i );
}

integer ceiling_r( const real x )
{
 integer i;
 i = (integer)x;
 if ( x > (real)i ) i++;
 return( i );
}

integer floor_d( const double x )
{
 integer i;
 i = (integer)(x+ 2 * DBL_EPSILON);
 if ( x < (double)i ) i--;
 return( i );
}

integer floor_r( const real x )
{
 integer i;
 i = (integer)(x + 2 * FLT_EPSILON);
 if ( x < (real)i ) i--;
 return( i );
}


/* The ANSI C modulo operator is unreliable (machine dept) for negative operands 
 and undefined for real operands */

integer modulo_i( const integer x, const integer base )
{
 integer result;
 if ( x > 0 && base > 0 ) {
  result = x % base;
 } else {
  result = x - floor_d( (double)x / (double)base ) * base;
 }
 return( result );
}

short modulo_s( const short x, const short base )
{
 short result;
 if ( x > 0 && base > 0 ) {
  result = x % base;
 } else {
  result = x - floor_d( (double)x / (double)base ) * base;
 }
 return( result );
}

real modulo_r( const real x, const real base )
{
 real result;
 result = x - floor_r( x / base ) * base;
 return( result );
}

double modulo_d( const double x, const double base )
{
 double result;
 result = x - floor_d( x / base ) * base;
 return( result );
}



/*----------------------------------------------------------------------*/
/*   divisible
 *
 *   Is a divisible by b? 
 *
 *   1996 Aug 29 JCM
 *   Args:
 *  	a		(i)	Operand
 *	b		(i)	Divisor
 */

logical divisible( const integer a, const integer b )
{
 return (  ( a/b ) * b == a  );
}

/*----------------------------------------------------------------------*/
/*
 *   min_i
 *
 *   Return smaller of i and j
 *   1996 Aug 29 JCM
 *
 */
 
integer min_i( const integer i, const integer j )
{
 integer k;
 k = i < j ? i : j;
 return( k );
}

/*----------------------------------------------------------------------*/
/*
 *   max_i
 *
 *   Return larger of i and j
 *   1996 Aug 29 JCM
 *
 */

integer max_i( const integer i, const integer j )
{
 integer k;
 k = i > j ? i : j;
 return( k );
}


/*----------------------------------------------------------------------*/
/*
 *   rmod_d
 *
 *   Calculate quotient and remainder
 *   1996 Aug 29 JCM
 *   Args:
 *	x		(i)	Input value
 *	y		(o)	Remainder
 *	r		(i)	Divisor
 *	rmod_d		(o)	Quotient
 */

integer rmod_d( const double value, double* remainder, const double divisor )
{
 integer quotient;
 quotient = floor_d( value / divisor );
 *remainder = value - divisor * quotient; 
 return( quotient );
}

/*----------------------------------------------------------------------*/
/*
 *   rmod_i
 *
 *   Calculate quotient and remainder
 *   1996 Aug 29 JCM
 */

integer rmod_i( const integer value, integer* remainder, const integer divisor )
{
 integer quotient;
 quotient = floor_d( (double)value / (double) divisor );
 *remainder = value - divisor * quotient;
 return( quotient );
}


integer rmod_r( const real value, real* remainder, const real divisor )
{
 integer quotient;
 quotient = floor_r( value / divisor );
 *remainder = value - divisor * quotient; 
 return( quotient );
}

integer rmod_s( const short value, short* remainder, const short divisor )
{
 integer quotient;
 quotient = floor_r( (real)value / (real) divisor );
 *remainder = value - divisor * quotient;
 return( quotient );
}


/*----------------------------------------------------------------------*/
/*
 *  frac_r
 *
 *  Split value into integer and fractional parts.
 *  Uses FLT_EPSILON from <float.h>
 *  1996 Aug 29 JCM
 *
 */

void frac_r( const real value, integer* int_val, real* fraction )
{
 integer integer_part;

 integer_part = floor_r( value );
 *fraction = value - integer_part;
 if ( *fraction > 1.0 - 2 * FLT_EPSILON ) { *fraction = 0.0; integer_part++; }
 *int_val = integer_part;
}

/*----------------------------------------------------------------------*/
/*
 *  frac_d
 *
 *  Split value into integer and fractional parts.
 *  Uses DBL_EPSILON from <float.h>
 *  1996 Aug 29 JCM
 *
 */

void frac_d( const double value, integer* int_val, double* fraction )
{
 integer integer_part;

 integer_part = floor_d( value );
 *fraction = value - integer_part;
 if ( *fraction > 1.0 - 2 * DBL_EPSILON ) { *fraction = 0.0; integer_part++; }
 *int_val = integer_part;
}

/*----------------------------------------------------------------------*/
/*
 *  logbase_i
 *
 *  Return log to base n of argument, rounded down to nearest integer and
 *  at least unity. Used in sort algorithm.
 *  1996 Aug 30 JCM
 *  Args:
 * 	i		(i)	Argument
 *	n		(i)	Base
 */

integer logbase_i( const integer i, const integer n )
{
 double x;
 x = logbase_d(  (double)i , n ) + FUNC_TOL;
 return( max_i( (integer)x, 1 ));
}

/*----------------------------------------------------------------------*/
/*
 *  logbase_d
 *
 *  Return log to base n of argument
 *  Uses log from <math.h>
 *  1996 Aug 30 JCM
 *  Args:
 * 	x		(i)	Argument
 *	n		(i)	Base
 */

double logbase_d( const double x, const integer n )
{
 return( log( x )/log( (double)n ) );
}

real logbase_r( const real x, const integer n )
{
 double result;
 result = ( log((double)x )/log( (double)n ) );
 return( (real) result );
}

real round_r( const real x, const integer f )
{
 integer tol;
 real y;
 real x_log;
 integer x_exp;
 integer x_rounded;
 real x_log_mantissa;

 tol = f - 1;
 if ( x == 0.0 ) {
  y = 0.0;
 } else {
  x_log = rlog10( rabs( x ) );
  x_exp = floor_d( x_log ) - tol;
  x_log_mantissa = x_log - x_exp;
  x_rounded = (integer) ( pow( 10.0, x_log_mantissa ) + 0.5 );
  y = x_rounded * tpow_r( x_exp );
  if ( x < 0 ) y *= -1;
 }
 return( y );

}

double round_d( const double x, const integer f )
{
 integer tol;
 double y;
 double x_log;
 integer x_exp;
 integer x_rounded;
 double x_log_mantissa;

 tol = f - 1;
 if ( x == 0.0 ) {
  y = 0.0;
 } else {
  x_log = log10( fabs( x ) );
  x_exp = floor_d( x_log ) - tol;
  x_log_mantissa = x_log - x_exp;
  x_rounded = (integer) ( pow( 10.0, x_log_mantissa ) + 0.5 );
  y = x_rounded * tpow_d( x_exp );
  if ( x < 0 ) y *= -1;
 }
 return( y );
}

void order_r( real* x, real* y)
{
 real tmp;
 if ( *y < *x ) { tmp = *x; *x = *y; *y = tmp; }
}

void order_d( double* x, double* y )
{
 double tmp;
 if ( *y < *x ) { tmp = *x; *x = *y; *y = tmp; }
}


real lint_r( const real x1, const real x2, const real yy1, const real yy2, const real x0 )
{
 real y;
 if ( x1 == x2 ) {
  y = ( yy1 + yy2 ) / 2.0;
 } else {
  y = ( yy1 * ( x2 - x0 ) + yy2 * ( x0 - x1 ) ) / ( x2 - x1 );
 }
 return( y );

}


double lint_d( const double x1, const double x2, const double yy1, const  double yy2, const double x0 )
{
 double y;
 if ( x1 == x2 ) {
  y = ( yy1 + yy2 ) / 2.0;
 } else {
  y = ( yy1 * ( x2 - x0 ) + yy2 * ( x0 - x1 ) ) / ( x2 - x1 );
 }
 return( y );

}

real rpow( const real x, const real n )
{
 return( (real)pow( (real)x, (real)n ) );
}

real rlog10( const real x )
{
 return( (real)log10( (double) x ) );
}

real rabs( const real x )
{
/* Not in math.h! */
 real y;
 y = x < 0.0 ? -x : x;
 return( y );
}

real rsqrt( const real x )
{
 return( (real)sqrt( (double)x ) );
}


integer iplus( const logical q )
{
 return( q ? +1 : -1 );
}

integer nint_r( const real x )
{
 integer i;
 real f;
 frac_r( x, &i, &f );
 if ( f >= 0.5 ) i++;
 return( i );
}

integer nint_d( const double x )
{
 integer i;
 double f;
 frac_d( x, &i, &f );
 if ( f >= 0.5 ) i++;
 return( i );
}

