/*======================================================================*/
/*
 *    cs
 *
 *    Null terminated string ("C string") handling routines
 *
 * cstring.h:  strlen 		called by cs_ends
 * ctype.h:    toupper          called by cs_upper
 */
/*======================================================================*/
#include <string.h>
#include <ctype.h>
#include "jcmlib.h"



/*----------------------------------------------------------------------*/
/*    cs_copy_siz  
 *    
 *    Copy at most a given number of characters from one C string to another.
 *    Resulting string may not be null terminated.
 *    1996 Aug 26 JCM
 *   Inputs:
 *     	source			Source string
 *	siz			Max number of characters
 *   Outputs:
 *     	dest			Destination string
 */

void cs_copy_siz( const char* source, char* dest, const integer maxlen )
{
 integer siz = maxlen;
 if ( source == NULL ) { *dest= '\0'; return; }
 while ( ( *dest++ = *source++ ) != '\0'  && --siz ) ;
 if( !siz ) *dest = '\0';
}

/*-----------------------------------------------------------------------*/
/*   cs_copy
 * 
 *   Copy a string until null termination reached. 
 *   WARNING: May overflow bounds of destination string. Only use this routine
 *   if you are sure the source string is null terminated with fewer characters
 *   than are available for the destination string.
 *   1996 Aug 26 JCM
 *   Inputs:
 *     	source			Source string
 *   Outputs:
 *     	dest			Destination string
 */

void cs_copy( const char* source, char* dest )
{
 if ( source == NULL ) { *dest= '\0'; return; }
 while ( ( *dest++ = *source++ ) != '\0' ) ;
}


/*----------------------------------------------------------------------*/
/*   cs_get_ss
 *   
 *   Copy a substring of the source string, place in destination string and null terminate.
 *   Return a pointer to the destination string.
 *   1996 Aug 26 JCM
 *   Inputs:
 *     	source			Source string
 *     	pos1			Start position, counting from 1
 *     	pos2		 	Stop position, counting from 1
 *   Outputs:
 *     	dest			Destination string
 */

char* cs_get_ss( const char* source, char* dest, const integer pos1, const integer pos2 )
{
 char* result;  /* Pointer to start of destination string */
 integer siz;       /* Number of characters to copy */

 result = dest;
 siz = pos2 - pos1 + 1;
 source += pos1 -1;
 if ( siz <= 0 ) { 
  *dest = '\0';
 } else {
  while ( ( *dest++ = *source++ ) != '\0'  && --siz ) ;
  if ( !siz ) *dest = '\0';
 }
 return( result );
}

/*----------------------------------------------------------------------*/
/*   cs_put_ss
 *
 *   Copy a string to a substring of the destination string, adding 
 *   trailing spaces. No null termination is added.
 *   Warning: if the substring overwrites the null termination,
 *   you may get unexpected results. The intention is that this routine be
 *   used with strings that have been set to all zero or all blank in advance.
 *   1996 Aug 26 JCM
 *   Inputs:
 *     	source			Source string
 *     	pos1			Start position, counting from 1
 *     	pos2		 	Stop position, counting from 1
 *   Outputs:
 *     	dest			Destination string
 */

void cs_put_ss( const char* source, char* dest, const integer pos1, const integer pos2 )
{
 integer siz;
 siz = pos2 - pos1 + 1;
 dest += pos1 - 1;
 while ( siz-- ) {
  if ( *source != '\0' ) {
   *dest++ = *source++;
  } else {
   *dest++ = ' ';
  }
 }
}

/*----------------------------------------------------------------------*/
/*   cs_ends
 * 
 *   Return length of string up to last non-blank character prior to first null.
 *   Can be zero.
 *   1996 Aug 26
 *
 *   Args:
 *	buf		(i)	String to check
 */

integer cs_ends( const char* buf )
{
 integer size;

 size = strlen( buf );
 buf += size - 1 ;
 while ( size > 0 && *buf-- == ' '  )  size-- ;

 return( size );
}

/*----------------------------------------------------------------------*/
/*   cs_begs
 * 
 *   Return character position of first non blank character.
 *   1996 Aug 26
 *
 *   Args:
 *	buf		(i)	String to check
 */

integer cs_begs( const char* buf )
{
 integer size;
 integer pos;

 size = strlen( buf );
 pos = 1;
 while ( pos < size && *buf++ == ' ' ) pos++;

 return( pos );
}



/*----------------------------------------------------------------------*/
/*
 *   cs_zero
 *
 *   Set a string to null
 *   1996 Aug 26
 *   Args:
 *	line		(i/o)	String
 */

void cs_zero( char * line )
{
 line[0] = '\0';
}


/*----------------------------------------------------------------------*/
/*
 *   cs_denull
 *
 *   Change nulls to blanks
 *   1996 Aug 26
 *   Args:
 *	line		(i/o)	String
 */

void cs_denull( char * line, const integer maxlen )
{
 integer siz = maxlen;
 while ( siz-- ) {
  if ( *line == '\0' ) *line = ' ';
  line++;
 }
}


/*----------------------------------------------------------------------*/
/*
 *   cs_null
 *
 *   Set a string to all nulls
 *   1996 Aug 26
 *   Args:
 *	line		(i/o)	String
 *      size		(i)     Size
 */

void cs_null( char * line, const integer maxlen )
{
 integer size = maxlen;
 while( size > 0 )  line[--size] = 0;
}

/*----------------------------------------------------------------------*/
/*
 *   cs_upper
 *  
 *   Uppercase string
 *   1996 Aug 26
 *   Args:
 *	buf		(i/o) 	String
 *
 */

void cs_upper( char* buf )
{
 while( *buf != '\0' ) *buf++ = toupper( *buf );
}
void cs_lower( char* buf )
{
 while( *buf != '\0' ) *buf++ = tolower( *buf );
}


/*--------------------------------------------------------------------------------*/
/*
 *  cs_conc
 *
 */

char* cs_conc( const char* buf1, const char* buf2, char* buf )
{
 integer end;
 cs_copy( buf1, buf );
 end = cs_ends( buf1 );
 cs_copy( buf2, buf+end );
 return( buf );
}

char* cs_conc1( const char* buf1, const char* buf2, char* buf )
{
 integer end;
 cs_copy( buf1, buf );
 end = cs_ends( buf1 );
 buf[end] = ' ';
 cs_copy( buf2, buf+end+1 );
 return( buf );
}

void cs_mixcase( char* buf )
{
 *buf++ = toupper( *buf );
 while( *buf != '\0' ) *buf++ = tolower( *buf ); 
}


logical cs_uceq( const char* buf1, const char* buf2 )
{
 t_buf ubuf1;
 t_buf ubuf2;
 cs_copy( buf1, ubuf1 );
 cs_copy( buf2, ubuf2 );
 cs_upper( ubuf1 );
 cs_upper( ubuf2 );
 return( cs_eq( ubuf1, ubuf2 ) );
}
