/*======================================================================*/
/*
 *   cs lib 2
 *
 *   C string routines to manipulate strings
 *
 *   string.h:    strncmp   		used by cs_eq_siz
 *		  strchr		used by cs_ss_char
 *    		  strstr		used by cs_index and cs_ss
 */
/*======================================================================*/
#include <stdio.h>
#include <string.h>
#include "jcmlib.h"


/*----------------------------------------------------------------------*/
/*   cs_logic
 *
 *   Return a character Y for true and N for false.
 *
 */

char cs_logic( const logical q )
{
 return( q ? 'Y' : 'N' );
}

/*----------------------------------------------------------------------*/
/*
 *   cs_advptr
 *
 *   Advance pointer to next non blank character.
 *
 *   Note:   buf       is a pointer to pointer to char.
 *           *buf      is a pointer to char.
 *           **buf     is the value of the char.
 *           (*buf)++  increments the pointer to char, making it pointeger to the next character in the string.
 *           (**buf)++ increments the ASCII value of the character.
 *
 *   1996 Aug 26 JCM
 *   Args:
 *	buf		Pointer to C string
 */

void cs_advptr( char** buf )
{
 while( **buf == ' ' ) (*buf)++;
}


/*----------------------------------------------------------------------*/
/*
 *   cs_is_blank
 *
 *   Test whether null terminated string consists only of blank characters.
 *   1996 Aug 26 JCM
 *   Args:
 *	buf		(i) String
 */

logical cs_is_blank( const char* buf )
{
 while( *buf == ' ' ) buf++;
 return( *buf  == '\0' );
}



/*----------------------------------------------------------------------*/
/*
 *   cs_eq
 *
 *   Test for equality for null/blank terminated strings.
 *   Strings are null terminated; return equal if strings are the same
 *   with the exception of trailing blanks.
 *   1996 Aug 26 JCM
 *   Args:
 * 	buf1		(i)	String
 *	buf2		(i)	String
 */

logical cs_eq( const char* buf1, const char* buf2 )
{
 
 while( *buf1 == *buf2 && *buf1 != '\0' )  {
  buf1++;
  buf2++;
 }
 if ( *buf1 == '\0' ) return( cs_is_blank( buf2 ) );
 if ( *buf2 == '\0' ) return( cs_is_blank( buf1 ) );
 return( FALSE );
}


/*----------------------------------------------------------------------*/
/*
 *   cs_eq_siz
 *
 *   Test for equality for null terminated strings.
 *   Return true of the first /siz/ characters of buf1 and buf2 are equal.
 *   Not the same as the strncmp routine as trailing blanks are not
 *   significant.
 *   1996 Aug 26 JCM
 *   Args:
 * 	buf1		(i)	String
 *	buf2		(i)	String
 *      siz		(i)	Size
 */

logical cs_eq_siz( const char* buf1, const char* buf2, const integer rsiz )
{
 integer siz = rsiz;
 while( siz-- ) {
  if ( *buf1 == '\0' ) return( cs_is_blank( buf2 ) );
  if ( *buf2 == '\0' ) return( cs_is_blank( buf1 ) );
  if ( *buf1 != *buf2 ) return( FALSE );
  buf1++;
  buf2++;
 }
 return( TRUE );
}

logical cs_eq_ss( const char* buf, const integer pos1, const integer pos2, const char* comp )
{
 integer siz;
 const char* ptr;

 ptr = buf + pos1 - 1;
 siz = pos2 - pos1 + 1;
 return( cs_eq_siz( ptr, comp, siz ) );
}



/*----------------------------------------------------------------------*/
/*
 *   cs_ss
 *
 *  Test whether substring is present.
 *  Wraps the standard routine strstr.
 *
 */

logical cs_ss( const char* buf, const char* sub )
{
   return( strstr( buf, sub ) != NULL ); 
}

/*----------------------------------------------------------------------*/
/*
 *   cs_ss_char
 *
 *  Test whether character is present.
 *  Wraps the standard routine strchr.
 *
 */

logical cs_ss_char( const char* buf, const char sub )
{
   return( strchr( buf, sub ) != NULL ); 
}

/*----------------------------------------------------------------------*/
/*
 *   cs_index
 *
 *   Return the character position of the first occurence of the given substring.
 *   Return zero if the substring is not present.
 *   1996 Aug 26 JCM
 *   Args:
 *	buf		(i)	String to search
 *	sub		(i)	String to search for
 */

integer cs_index( const char* buf, const char* sub )
{
 char* ptr;

 ptr = strstr( buf, sub );
 if ( ptr == NULL )
  return(0);
 else 
  return( ptr - buf + 1 );
}

/*
 *   We look for a substring in buf starting at position pos.
 *   If pos is zero or negative, return zero. If the substring
 *   is not found, return zero. Otherwise return the
 *   position of the start of the substring.
 */

integer cs_eindex( const char* buf, const integer pos, const char* sub )
{
 char* ptr;
 if ( pos <= 0 ) {
  return( 0 );
 } else {
  ptr = strstr( buf+pos-1, sub );
  if ( ptr == NULL ) {
   return( 0 );
  } else { 
   return( ptr - buf + 1 );
  }
 }
}
/*----------------------------------------------------------------------*/
/*
 *   cmatch
 *
 *  Return array index of first equality match of string in string array
 *  Return 0 if no match
 *  1996 Aug 26 JCM
 *  Args:
 *	buf		(i) Search string
 *	array		(i) Pointer array to strings
 *	n		(i) Size of array
 *
 */

integer match_c( const char* buf, char** array, const integer n )
{
 integer no;

 for ( no = 1; no <= n; no++ )
  {
  if ( cs_eq( buf, *array++ ) ) return( no );
   }
 return( 0 );
}

integer matchr_c( const char* buf, char** array, const integer n1, const integer n2 )
{
 integer j;
 j = match_c( buf, array + n1 -1, n2 - n1 + 1 );
 return( j > 0 ? j + n1 -1 : 0 );
}

integer cmatch_c( const char* buf, char** array, const integer n )
{
 integer no;

 for ( no = 0; no < n; no++ )
  {
  if ( cs_eq( buf, *array++ ) ) return( no );
   }
 return( -1 );
}


integer mmatch_c( const char* targ, const char** list, const integer n )
{ 
 integer entry = 0;
 integer match_entry = 0;
 integer best_match_len = 0;
 integer n_match = 0;
 integer partial = 0;
 integer n_partial = 0;
 integer targ_len;
 integer match_len;
 const char* ptr;
 integer result;
/*
 * Look for the shortest string in a list that matches the target.
 *
* We loop through each list entry, seeing if it is exactly equal to the
* target. If it is, we remember its (1-based) entry number and increment the
* number of matches found so far. If not, we see if it is equal to
* the target up to the length of the target (e.g. target is "FOO" and
* entry is "FOOBAR"); this counts as a partial match. We see if it
* is the best partial match so far (shortest string that matches targ).
* If so, we remember it and its length.
* At the end, if we found an exact match we return its entry number;
* otherwise if we found exactly one partial match we return its
* entry number; else we return zero.
* At the moment, all of the info about number of matches and lengths
* is thrown away; a future version of this routine will generate return
* information about ambiguous matches.
*
*/

 targ_len = cs_ends( targ );
 for ( entry = 1; entry <= n; entry ++ ) {
  ptr = list[entry-1];
  if( cs_eq( targ, ptr ) ) {
   match_entry = entry;
   n_match++;
  } else if ( cs_eq_siz( targ, ptr, targ_len ) ) {
   match_len = cs_ends( ptr );
   if( partial  == 0 || match_len <= best_match_len ) {
    if ( match_len == best_match_len ) n_partial++;
    partial = entry;
    best_match_len = match_len;
   }
  }
 }

 if( n_match >= 1 ) {
  result = match_entry;
 } else if ( n_partial == 1 ) {
  result = partial;
 } else {
  result = 0;
 }
 return( result );
}


/*----------------------------------------------------------------------*/
/*
 *   cs_match
 *
 *  Return array index of first equality match of string in string array
 *  The size argument does not include the terminating null.
 *  Return 0 if no match
 *  1996 Aug 26 JCM
 *  Args:
 *	buf		(i) Search string
 *	array		(i) Pointer to first char in array
 *	n		(i) Size of array
 *      size		(i) Size of each array string; Number of bytes to advance pointer less one
 */

integer cs_match( const char* buf, const char* array, const integer n, const integer size1 )
{
 integer no;
 integer size = size1;
 size++;  /* Allow for terminating null */

 for ( no = 1; no <= n; no++ )
  {
  if ( cs_eq( buf, array ) ) return( no );
  array += size;
   }
 return( 0 );
}


integer cs_index_char( const char* buf, const char c )
{
 char* ptr;
 ptr = strchr( buf, c );
 if ( ptr == NULL ) {
  return( 0 );
 } else {
  return(  ptr - buf + 1 );
 }
}

void subcmd_split( const char* opt, char* cmd, char* sub, const char sep )
{
 integer i;

 i = cs_index_char( opt, sep );
 if ( i == 0 ) {
  cs_copy( opt, cmd );
  cs_copy( " ", sub );
 } else if ( i == 1 ) {
  cs_copy( " ", cmd );
  cs_copy( opt+1, sub );
 } else {
  cs_copy_siz( opt, cmd, i-1 );
  cs_copy( opt+i, sub );
 }
}



