#include "jcmlib.h"
#include <math.h>


#define MAX_EXP_ARG 700.0
#define MIN_EXP_ARG 1.0E-10
#define MIN_SINH_ARG 1.0E-10
#define MAX_SINH_ARG 1.0E+40
#define TINY_D 1.0E-100

double eexp( const double x )
{
 double result;
 if ( x > MAX_EXP_ARG ) {
  result = infinity_d( 1 );
 } else if ( x < -MAX_EXP_ARG ) {
  result = 0.0;
 } else {
  result = exp( x );
 }
 return( result );
}

double ecosh( const double x )
{
 double result;
 result = 0.5 * ( eexp( x ) + eexp( -x ) );
 return( result );
}

double esinh( const double x )
{
 double result;
 if ( fabs( x ) < MIN_SINH_ARG ) {
  result = x;
 } else {
  result = 0.5 * ( eexp( x ) - eexp( -x ) ); 
 }
 return( result );

}

double asinh( const double x )
{
 double result;
 double ax;

 ax = fabs( x );
 if ( ax < MIN_SINH_ARG ) {
  result = ax;
 } else if ( ax > MAX_SINH_ARG ) { 
  result = log( 2.0 * ax );
 } else {
  result = log( ax + sqrt( 1.0 + ax * ax ) );
 }
 if ( x < 0.0 ) result *= -1;
 return( result );
}

double acosh( double x )
{
 double eps;
 double result;
 eps = x - 1.0;
 
 if ( eps < 0.0 ) {
  result = 0.0;
 } else if ( eps < MIN_SINH_ARG ) {
  result = sqrt( 2.0 * eps );
 } else if ( eps > MAX_SINH_ARG ) {
  result = log( 2.0 * x );
 } else {
  result = log( x + sqrt ( x * x - 1.0 ) );
 }
 return( result );
}


double bose_einstein( const double x )
{
 double result;
 if ( x > MAX_EXP_ARG ) {
  result = 0.0;
 } else if ( x < -MAX_EXP_ARG ) {
  result = -1.0;
 } else if ( fabs(x) < MIN_EXP_ARG ) {
  if ( fabs(x) < TINY_D ) {
   result = infinity_d( 1 );
  } else {
   result = 1.0 / x;
  }
 } else {
  result = 1.0 / ( eexp(x) - 1.0 );
 }
 return( result );
}

double dplanckn( const double x, const integer n )
{
 double result;
 if ( n == 0 ) {
  result = bose_einstein( x );
 } else if ( x > MAX_EXP_ARG ) {
  result = 0.0;
 } else if ( x < -MAX_EXP_ARG ) {
  result = -1.0 * ipow_d( x, n );
 } else if ( fabs(x) < MIN_EXP_ARG ) {

  if ( fabs( x ) < TINY_D ) {
   if ( n == 1 ) {
    result = 1.0;
   } else if ( n < 0 ) {
    result = infinity_d( 1 );
   } else {
    result = 0.0;
   }
  } else {
   result = ipow_d( x, (n-1) );
  }
 } else {
  result = ipow_d( x, n ) / ( eexp( x ) - 1.0 );
 }
 return( result );
}

/*
 *  Raise to integer power
 *   ipow_d( x, 0 ) = 1
 *   ipow_d( x, 1 ) = x
 *   ipow_d( x, -n ) = 1 / x^n
 *   ipow_d( x, n ) = x^n
 */
double ipow_d( const double x, const integer m )
{
 integer n = m;
 double result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n == 1 ) {
  return ( x );
 } else if ( n < 0 ) {
  if ( x == 0.0 ) return( infinity_d( 1 ) );
  while( n++ ) result *= x;
  return( 1.0/result );
 } else {
  if ( x == 0.0 ) return ( x );
  while( n-- ) result *= x;
  return( result );
 }
}

real ipow_r( const real x, const integer m )
{
 integer n = m;
 real result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n == 1 ) {
  return ( x );
 } else if ( n < 0 ) {
  if ( x == 0.0 ) return( infinity_d( 1 ) );
  while( n++ ) result *= x;
  return( 1.0/result );
 } else {
  if ( x == 0.0 ) return ( x );
  while( n-- ) result *= x;
  return( result );
 }
}

/* Ten to the power */

double tpow_d( const integer m )
{
 integer n = m;
 double result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n < 0 ) {
  while( n++ ) result *= 10.0;
  return( 1.0/result );
 } else {
  while( n-- ) result *= 10.0;
  return( result );
 }
}

real tpow_r( const integer m )
{
 integer n = m;
 real result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n < 0 ) {
  while( n++ ) result *= 10.0;
  return( 1.0/result );
 } else {
  while( n-- ) result *= 10.0;
  return( result );
 }
}

double dgauss( const double x, const double x0, const double sigma )
{
 double arg;
 double result;
 if ( sigma == 0.0 ) {
  result = 0.0;
 } else {
  arg = ( x - x0 ) / sigma;
  result = eexp( -0.5 *  ( arg * arg ) );
 }
 return( result );
}

