#ifndef JCMLIB_
#define JCMLIB_

#define JCMLIB "JCMLIB UTILS"
#define JCMLIBVER "CV1.5"

/* V1.1 1997 Feb 19 Made consistent use of long int */
/* V1.2 1997 Mar  9 Moved all prototypes into jcmproto.h */
/* V1.3 1997 Mar 10 Fixed SIO and TFR bugs, unitlib bug */
/* V1.4 1997 May  2 Added new allocate routines, vsub_d */
/* V1.5 1997 Nov 27 Cleaned up, added const declarations, moved stuff back to jcmlib.h */

#define JCM_ERR_START 1000
#include "jcmerr.h"
/* 
 * JCMLIB global symbols
 */

#define TRUE  1
#define FALSE 0

/*
 *  String sizes
 */
#undef CSIZE
#define FSIZE 255	/* Size of t_filename path name */
#define CSIZE 80	/* Size of card */
#define BSIZE 255	/* Size of t_buf  */
#define WSIZE 31	/* Size of t_word */

/*
 *  File symbols
 */
#define FIO_NOPERM   -1
#define FIO_NOFILE    0
#define FIO_READONLY  1
#define FIO_READWRITE 2
#define FIO_CLEAR     0
#define FIO_CHECK     1
#define IOFS_POP_CLOSE  1
#define IOFS_POP_ONLY  -1
/* Global zero */
extern const double VZERO_D[3];

/* JCM special types */

/*
 *
 *  JCMLIB types
 *  In general I have avoided the use of special types, which tend to
 *  make a general purpose library less reusable.
 *  For example, 3-vector arguments are just double* and you must read the
 *  documentation to know that you need a double[3].
 *  The 'integer' type is just a #define for long; the 'real' type is a 
 *   #define for float.
 *  Thus, you may currently use just long and float in your calls.
 *  The 'logical' type is a typedef for int and is used for 
 *  boolean values.
 *  The 'complex' and 'dcomplex' types support complex numbers.
 *
 *  There are also some special string types: t_word for short strings,
 *  card for 80 byte ASCII cards, t_buf for long strings, and t_filename
 *  for pathnames. These are never used in the external prototypes,
 *  but you may want to use FSIZE to check if your pathnames will exceed
 *  JCMLIB's capacity.
 *
 *  Note that all string tests in the library ignore trailing blanks;
 *  and a string of blanks is considered the same as a null string.
 *  This improves compatibility with Fortran programs.
 */

#define integer  long
#define real     float
typedef int logical;    /* C standard specifies result of logical expr is int */

typedef char card[CSIZE+1];
typedef char t_filename[FSIZE+1];
typedef char t_word[WSIZE+1];
typedef char t_buf[BSIZE+1];

typedef struct { real x;   real y; }   complex;
typedef struct { double x; double y; } dcomplex;


/*
 *   JCMLIB routine prototypes
 *  --  GENERAL: --
 *   1. Initialization
 *   2. System related routines
 *   3. Bit related functions
 *  --  STRINGS: --
 *   4. Converting between strings and numbers.
 *   5. Basic string manipulation
 *   6. Substrings
 *   7. String cleanup routines
 *   8. F90 style string functions
 *   9. Miscellaneous string functions
 *  10. Token parsing
 *  --  NUMBERS: --
 *  11. Sorting
 *  12. Arithmetic and Math
 *  13. Sexagesimal functions
 *  14. Special functions
 *  15. Complex number routines
 *  16. Extreme value handling
 *  17. Random numbers.
 *  18. Circular and trig functions
 *  19. 3-vector routines
 *  --   ARRAYS:  ---
 *  20. Array memory management
 *  21. Basic array operations
 *  22. F90 style array operations
 *  23. Logical arrays
 *  --   INPUT AND OUTPUT --
 *  24. General file I/O (FIO)
 *  25. Byte streams (SIO)
 *  26. Raw character streams (CIO)
 *  27. Text files (FIO)
 *  28. Message streams (IOFS)
 *  29. Command line interface (TPIO)
 *  30. Escape commands (ESC)
 *  31. Debug routines (DEBUG)
 *  32. Error handling (ERR)
 *  --   MISCELLANEOUS --
 *  33. Units
 *  34. Another string parser
 *  35. Miscellaneous and obsolete routines
 */

/*
 *  \section{Initialization}
 */

extern void jcmlib_init( void );
/*
 *   This routine should be called before any other jcmlib
 *   routine. It initializes various things, 
 *   including the NaN handling, I/O buffers, and unit manipulation.
 */

/*
 *  \section{System type routines}
 */

extern void sys_filepath( const char* dir, const char* name, const char* ext, char* buf, integer siz );
/*
 *   Construct a full pathname from name. The only version currently supported in C
 *   is the Unix one; the Fortran library supports other operating systems too.
 *   If name is an absolute path (begins with / in Unix) then ignore dir,
 *   otherwise prepend dir with a / if dir is not blank. Then if ext is not
 *   blank, append a dot followed by ext. Put the result in buf, of max size siz.
 *   Example:
 *      sys_filepath( "/usr/local/data", "total", "dat", buf, maxlen );
 *  puts  "/usr/local/data/total.dat" in buf, while
 *      sys_filepath( "/usr/local/data", "/total", " ", buf, maxlen );
 *  puts "/total"  in buf. 
 */

extern void sys_pathstrip( const char* path, char* dir, char* name, const integer maxlen );
/*
 *   Take a pathname and split it into the directory name and the file name.
 */

extern void sys_os( char*  );
/*
 *   Return a string (< 32 bytes) indicating the type of operating system.
 */

extern void jcmlib_ver( char* ver, const integer maxlen );
/*
 *  Return the current version of JCMLIB in a string.
 */

extern logical ver_init( const char* prog, const char* ver );
/*
 *   Load a package and its version into the VER system; a call to ver_msg
 *   will then display the package and version number.
 */

extern void ver_msg( void );
/*
 *   Issue a message describing the version of various packages currently loaded into
 *   the VER system (at least JCMLIB).
 *   TPIO command: \VER
 */

extern void date_and_time( char* date, char* ctime, char* zone );
/*
 *  Return strings containing the current date, time, and time zone.
 */

extern void sys_date( char* buf, const integer size );
/*
 *  Returns the current date.
 */

extern void sys_time( char* buf, const integer size );
/*
 *  Return the current time.
 */

extern void sys_today( char* buf, const integer size );
/*
 *  Returns the current date and time.
 */

extern void sys_wait( const integer t );
/*
 *  Pause for t seconds
 */

extern void sys_getenv( const char* name, char* value );
/*
 *  Return the value of an environment variable, or the equivalent.
 */

extern void sys_delete( const char* name );
/*
 *   Delete a file from the file system.
 */

extern void sys( const char* cmd );
/*
 *  Issue a system command.
 */

extern void sys_type( const char* filename );
/*
 *  Type a file to the screen (using 'cat' on a Unix system).
 */

extern double sys_stamp( void );
/*
 *   Produce a double valued number related to the date and time,
 *   suitable for labelling metadata etc.
 */

extern void jcmlib_id( char* label, const char* ver );
/*
 *  Construct the string 'label' by appending the string 'ver' to
 * the date and time. 
 */


/*
 *   \section{Bit related functions}
 */  

extern void ieee_get_i( const char* buf, integer* data, const integer n );
extern void ieee_get_s( const char* buf, short* data, const integer n );
extern void ieee_get_d( const char* buf, double* data, const integer n );
extern void ieee_get_r( const char* buf, real* data, const integer n );
/*
 *  Get n items of data from an IEEE byte buffer, byte swapping if necessary
 */

extern void ieee_put_r( const real* data, char* buf, const integer n );
extern void ieee_put_d( const double* data, char* buf, const integer n );
extern void ieee_put_s( const short* data, char* buf, const integer n );
extern void ieee_put_i( const integer* data, char* buf, const integer n );
/*
 *  Put n items of data to an IEEE byte buffer, byte swapping if necessary
 */

extern double read_binary_d( const char* buf, const char mode );
extern real read_binary_r( const char* buf, const char mode );
extern integer read_binary_i( const char* buf, const char mode );
extern short read_binary_s( const char* buf, const char mode );
/*
 *  Parse a string of binary, octal or hex digits. Mode is
 *  'B', 'O' or 'Z'
 */

extern void write_binary_d( const double value, char* buf, const char mode );
extern void write_binary_r( const real value, char* buf, const char mode );
extern void write_binary_s( const short value, char* buf, const char mode );
extern void write_binary_i( const integer value, char* buf, const char mode );
/*
 *  Convert value to a string of binary, octal or hex digits. Mode is
 *  'B', 'O' or 'Z'
 */

extern void bytes_split_i( const integer value, integer* bytes );
extern void bytes_split_r( const real value, integer* bytes );
extern void bytes_split_d( const double value, integer* bytes );
extern void bytes_split_s( const short value, integer* bytes );
extern void bytes_split_c( const char* value, integer* bytes, const integer n_bytes );
/*
 *  Split an input value into individual bytes. Each byte value is stored
 *  in an integer.
 */

extern integer bytes_combine_i( const integer* bytes );
extern short bytes_combine_s( const integer* bytes );
extern real bytes_combine_r( const integer* bytes );
extern double bytes_combine_d( const integer* bytes );
extern void bytes_combine_c( const integer* bytes, char* value, const integer n_bytes );
/*
 *   Combine an array of byte values into a single value.
 *   For instance, bytes_combine_d combines 8 integers into a single
 *   double precision value.
 */

extern void byte_set_s( short* number, const integer value, const integer byte );
extern void byte_set_i( integer* number, const integer value, const integer byte );
extern void byte_set_r( real* number, const integer value, const integer byte );
extern void byte_set_d( double* number, const integer value, const integer byte );
extern void byte_set_c( char* data, const integer value, const integer byte );
/*
 *  Set a single byte in a number. 
 *   byte_set_d( &v, 120, 2 )
 *  would set the second byte in v to 120. Bytes are numbered starting at 1.
 */

extern integer byte_get_d( const double value, const integer byte );
extern integer byte_get_s( const short value, const integer byte );
extern integer byte_get_r( const real value, const integer byte );
extern integer byte_get_i( const integer value, const integer byte );
extern integer byte_get_c( const char* value, integer byte );
/*
 *  Get a single byte from a number. Bytes are numbered starting at 1.
 *  byte_get_d( v, 2 ) gets the second byte of v.
 */

unsigned short single_bit_s( const integer bit_no );
unsigned long  single_bit_i( const integer bit_no );
/*
 *   Return an integer with a single bit set.
 */

extern void bit_set_i( integer* value, const integer bit_no );
extern void bit_set_s( short* value, const integer bit_no );
/*
 *  Set a specified bit in value
 */

extern void bit_clr_i(  integer* value, const integer bit_no );
extern void bit_clr_s(  short* value, const integer bit_no );
/*
 *  Clear a bit in value
 */

extern short bit_shft_s( const short value, const integer bits );
extern integer bit_shft_i( const integer value, const integer bits );
/*
 *  Shift right by a given number of bits.
 */

extern logical bit_test_i( const integer value, const integer bit );
extern logical bit_test_s( const short value, const integer bit );
/*
 *  Test if given bit is set
 */

extern integer bit_unpack_i( const integer value, const integer start_bit, const integer n_bits );
extern integer bit_unpack_s( const short value, const integer start_bit, const integer n_bits );
/*
 * Extract bits start_bit through start_bit + n_bits - 1 from value.
 */


extern short   bit_ar_set_s( const integer* bits );
extern integer bit_ar_set_i( const integer* bits );
extern real    bit_ar_set_r( const integer* bits );
extern double  bit_ar_set_d( const integer* bits );
/*
 *  Combine a set of bits into a value
 */

extern integer bit_ar_get_r( const real value, integer* bits );
extern integer bit_ar_get_d( const double value, integer* bits );
extern integer bit_ar_get_s( const short value, integer* bits );
extern integer bit_ar_get_i( const integer value, integer* bits );
/*
 *  Split a value into a set of bits
 */


extern integer bits_unsigned( const short value );
/*
 *  Shift a signed value into the unsigned range (but as a signed value), wrapping
 *  negative values to values above 32767.
 */

extern integer bits_val( const integer* bits, const integer n );
/*
 *  Make an integer from an array of bits by repeated multiplication by 2
 */

/*
 *  \section{Conversion between string and numeric}
 */

extern void Fortran_Format_Parse( const char*  Fortran_Format, char* C_Format, char* DataType, int* Field_Length );
/*
 *   Parse a single Fortran format edit descriptor, and construct the equivalent C format,
 *   as well as returning a string indicating the data type,  and the field length.
 */

extern integer read_i( const char* buf );
extern double  read_d( const char* buf );
extern real    read_r( const char* buf );
extern short   read_s( const char* buf );
extern logical read_q( const char* buf );
/*
 *   The read routine is similar to ANSI C's atoi, atod functions.
 *   It converts a string to a numeric value. If the string is blank,
 *   it returns zero and sets an internal error. If the string is "-", or "NaN"
 *   it recognizes the NaN condition and returns the appropriate NaN value.
 */

extern integer read_digit( const char c );
extern integer read_hex_digit( const char c );
extern integer read_i1( const char* buf );
/*
 *  Convert char '0' to '9' into an integer; or '0' to 'F' for hex version.
 *   The read_i1 variant reads only the first non-blank character and returns
 *   a digit from 0 to 9.
 */


extern void read_val_d( char** buf, integer* sign_val, integer* int_val, double* frac_val );
extern void read_val_de( char** buf, integer* sign_val, integer* int_val, double* frac_val );
extern void read_val_ee( char** buf, integer* sign_val, integer* int_val, double* frac_val, integer* exp_val );
/*
 *   The read_val routines parse a string, attempting to read a number from it.
 *  When they run out of number, they leave the pointer pointing at the next
 *  position of the string. 
 *  read_val_d reads a simple floating point number and returns separately
 *   its sign, integral part and fractional part.
 *  read_val_de supports exponential notation (E+00 etc) and returns the
 *   same info as read_val_d.
 *  read_val_ee is the most sophisticated. It returns the exponent
 *   separately, and supports the following notations: (^ is 'to the power of')
 *     14
 *     14.0
 *     14.0E-03
 *     14.0D-03
 *     14.0x10^-3
 *     14.0x2^32
 *     14.0x10^2.718
 *     10^5
 *     2.48^5
 */


extern char* write_i( const integer val, char* buf, const integer siz );		/* Write integer to string */
extern char* write_d( const double val, char* buf, const integer siz );
extern char* write_r( const  real val, char* buf, const integer siz );
extern char* write_s( const short val, char* buf, const integer siz );
extern char  write_q( const logical q );
/*
 * Write a numeric value to a string, the inverse of the read routines.
 * The write_q routine returns 'T' or 'F'. (see also cs_logic which returns Y or N)
 */

extern char* writef_i( const integer val, const char* fmt, char* buf, const integer siz );
extern char* writef_r( const real val, const char* fmt, char* buf, const integer siz );
extern char* writef_s( const short val, const char* fmt, char* buf, const integer siz );
extern char* writef_d( const double val, const char* fmt, char* buf, const integer siz );
/*
 *   Write a value to a string using the C format string provided.
 */

extern char* vwritef_d( const double* p, const char* fmt, char* buf, const integer siz );
extern char* vwritef_r( const real* p, const char* fmt, char* buf, const integer siz );
/*
 *   Write a 3-vector to a string, using the format string for each element.
 */


extern char* write_opt_s( const char* prompt, const short val, char* buf, const integer siz );
extern char* write_opt_i( const char* prompt, const integer val, char* buf, const integer siz );
extern char* write_opt_r( const char* prompt, const real val, char* buf, const integer siz );
extern char* write_opt_d( const char* prompt, const double val, char* buf, const integer siz );
extern char* write_opt_c( const char* prompt, const char* val, char* buf, const integer siz );
/*
 *  Concatenate the string 'prompt' with the value 'val', first enclosing
 *  'val' in square parentheses. Used to prompt the user with a default.
 */


extern char* write_e_r( const real x, const real off, char* buf, const integer siz );
extern char* write_e_d( const double x, const double off, char* buf, const  integer siz );
/*
 *   Write an extended value to a string. An extended value (x,off)
 *   is interpreted as x * ( 10 ** off ). 
 */

extern char* write_tp_i( const integer i, char* buf, const integer siz );
/*
 *   Write an integer to a string enclosed in 'tex parens', curly braces.
 *   e.g. "{32}".
 */

extern char* write_rj_i( const integer i, char* buf, const integer siz );
/*
 *   Write an integer to a string, right justified in a field of size 'siz'.
 */


/*
 *  \section{Basic string manipulation}
 *
 *  The <string.h> routines are somewhat perverse in their calling
 *  sequences, at least I can never remember them. Also, I like to
 *  use the Fortran convention of considering strings to be equal if
 *  they differ only by trailing blanks (ASCII 32). 
 */

extern void cs_copy_siz( const char* source, char* dest, integer siz );	
/*
 *   Copy at most siz characters from source to dest, or until null found
 *   Result is null terminated.
 */

extern void cs_copy( const char* source, char* dest );
/*
 *  Copy null terminated string from source to dest; use only if
 *  you know dest is big enough.
 */

extern integer cs_ends( const char* buf );		
/*
 *  Return the length of a string (not including trailing blanks)
 *  Result is always at least 1.
 */

extern integer cs_begs( const char* buf );	
/*
 *  Return the position of the first non-blank character in the string.
 *  Result is always at least 1.
 */

extern void cs_upper( char* buf );	
extern void cs_lower( char* buf );	
/*
 *   Change case of string, in place. 
 */

extern void cs_mixcase( char* buf );
/*
 *   The mixcase variant 
 *   uppercases the first letter of the string and lowercases the
 *   remainder. As the owner of a surname which does not follow this
 *   orthographic paradigm, I caution you not to use this routine
 *   too enthusiastically - I've seen plenty of commercial software
 *   which assumes all proper nouns can be mixcased like this.
 */

extern logical cs_is_blank( const char* buf );					
/*
 *  True if string is blank or null.
 */

extern logical cs_eq( const char* buf1, const char* buf2 );				
/*
 *  True if strings are equal up to null termination, ignoring
 *  trailing blanks.
 */

extern logical cs_eq_siz( const char* buf1, const char* buf2, const integer siz );			/* String equal first n chars */
/*
 *   True if strings are equal in their first siz bytes, ignoring
 *   trailing blanks.
 */

extern logical cs_uceq( const char* buf1, const char* buf2 );
/*
 *  Case-insensitive string equality.
 *  True if buf1 and buf2 are equal, ignoring case.
 */

extern logical cs_mmatch( const char* template, const char* buf, const integer n ); 
/*
 *   True if the string 'template' matches 'buf' on the first n characters.
 */

extern integer mmatch_c( const char* buf, const char** array, const integer n );
/*
 *  Minimum match on list. Find the shortest string in the array which matches
 *  buf for at least m characters, where m is the number of characters in
 *  buf ignoring trailing blanks.
 */

extern void cs_zero( char * line );	
/*
 *  Make a string blank; currently implemented by setting line[0] to null.
 */

extern void cs_null( char * line, const integer siz );	
/*
 *  Fill line with (siz) nulls.
 */

extern logical cs_nullstr( const char* source );
/*
 *  Test if string is null.
 */

extern logical alpha_char( const char c );
extern logical digit_char( const char c );
extern logical numeric_char( const char c );
/*
 *   Return true if c is an alphabetic character, a digit, or a
 *   numeric character (digit, period, sign).
 */

extern void subcmd_split( const char* opt, char* cmd, char* sub, const char sep );
/*
 *  Parse the opt to look for the sep character; return anything
 *  before the sep in cmd, anything after in sub. If sep is not found,
 *  cmd = opt and sub is set to blank.
 */

extern char* cs_conc( const char* buf1, const char* buf2, char* buf );			/* Concatenate */
extern char* cs_conc1( const char* buf1, const char* buf2, char* buf );
/*
 *   Concatenate two strings, stripping trailing blanks from the first,
 *   putting result in the third. The conc1 variant
 *   adds a single space between the strings.
 */

extern logical cs_gt( const char* x, const char* y );					/* ASCII comparison */
extern logical cs_lt( const char* x, const char* y );
/*
 *   ASCII order inequality tests on strings.
 */

extern integer cs_list_add( char** list, integer* list_size, const integer max_size, const char* item );
/*
 *  Add an entry 'item' to a list of strings and return its number in the list.
 *  If item is already in the list, return its number, i.
 *  If it is not already in the list, add it to the end of the list,
 *  as long as list_size does not exceed max_size, 
 *  increment list_size, and return its number i = list_size.
 *  The number in the list is counted from 1, so item = list[i-1].
 */

/*
 *  \section{Substrings}
 */

extern char* cs_get_ss( const char* source, char* dest, const integer pos1, const integer pos2 );	
/*
 * Get source(pos1:pos2) and put it in dest. The character positions are numbered
 * starting at 1. Equivalent to cs_copy_siz( source[pos1-1], dest, pos2-pos1+1 )
 */

extern void cs_put_ss( const char* source, char* dest, const integer pos1, const integer pos2 );		
/*
 * Put source into substring of dest(pos1:pos2). Do not null terminate;
 * pad with blanks if source is shorter than pos2-pos1+1.
 */

extern logical cs_eq_ss( const char* buf, const integer pos1, const integer pos2, const char* comp );
/*
 *  True if buf(pos1:pos2) is equal to comp.
 */

extern logical cs_ss( const char* buf, const char* sub );				
/*
 *  True if substring is present in buf
 */

extern integer cs_index( const char* buf, const char* sub );
/*
 *  Fortran INDEX routine.
 *  Returns the byte position of substring sub in text buf; thus
 *    cs_index( "ABCDEFGCDE", "CDE" ) = 3
 *  since CDE starts at the 3rd position in ABCDEFGCDE.
 *  Returns zero if sub is not a substring of buf.
 */

extern integer cs_eindex( const char* buf, const integer pos, const char* sub );
/*
 *  Extended INDEX routine. Returns byte position of first appearance of
 *  substring sub in buf that is greater than or equal to pos.
 *  Thus 
 *    cs_eindex( "ABCDEFGCDE", 4, "CDE" ) = 8
 *  since the next appearance of CDE after position 4 is position 8.
 */

extern integer cs_index_char( const char* buf, const char c );
/*
 *   Return the byte position of the first occurrence of character c in
 *   string buf. Return zero if c is not in buf.
 */

extern logical cs_ss_char( const char* buf, const char sub );				
/* 
 *  True if character is present in string
 */

extern logical cs_split_string( const char* buf, const char sep, char* buf1, char* buf2 );  
/*
 * Split string at separator char 'sep'.
 * Look for first case of 'sep', everything before sep is put in buf1, everything
 * after sep is put in buf2.
 */

extern char* cs_upto( const char* source, const char targ, char* dest );
/*
 *   Copy source to dest up to but not including first occurrence of
 *   character targ.
 */


/*
 *  \section{String cleanup routines}
 */
extern void cs_denull( char* buf, const integer siz );
/*
 *   Change nulls to spaces. Warning: string may no longer be null terminated.
 */

extern void cs_detab( char* buf );		
/*
 *   Change tabs to spaces (1 tab goes to 1 space).
 */

extern void cs_decmt( char* buf, const char esc, const char cmt );
/*
 *   Remove everything after the first occurrence of cmt, except
 *   if cmt is preceded by esc, in which case just the esc is dropped.
 *   For example, if esc is \ and cmt is #, 
 *      "read  rh\#42.dat  # Read my data file which is called rh#42.dat"
 *   gets converted to 
 *      "read  rh#42.dat   "
 */

extern void cs_remove_nl( char* line );				
/*
 *   Remove trailing \r and \n characters from a line.
 */

extern void cs_ubar( char* source );
/*
 *  Change all spaces to underbars.
 */

extern void cs_deubar( char* source );
/*
 *  Change all underbars to spaces.
 */

extern void cs_compress_spaces( const char* source, char* dest );	
/*
 *  Copy source to dest, but compress multiple spaces to single ones.
 */

extern void cs_strpak( const char* source, char* dest );            /*F strpak */    /* Pack string, remove spaces and newlines */
/*
 *   Remove spaces and newlines from a string.
 */

extern void string_replace( char* cbuf, const char* old, const char* new );
/*
 *  Edit a string in place, replacing occurrences of 'old' with 'new'.
 */

extern char* cs_sparen( const char* source, char* dest );		/*F gparen */
/*
 *   Copy source to dest, enclosing it in square parentheses [].
 */
extern char* cs_paren( const char* source, char* dest );	
/*
 *    Copy source to dest, enclosing it in round parentheses ().
 */

extern void cs_gparen( const char* source, const char lpar, const char rpar, char* dest ); 
/*
 *  Copy source to dest, enclosing it in characters lpar and rpar.
 *  Trailing spaces on source are removed.
 */

extern void cs_deparen( char** buf, char* dest );
/*
 *  Look for a string in parentheses; strip the parentheses
 *  and put the contents in dest; and advance the buf pointer
 *  to after the closing parenthesis.
 */


/*
 * \section{F90 style string functions}
 */


extern char* adjustl( const char* a, char* b, const integer width );
extern char* adjustr( const char* a, char* b, const integer width );
/*
 *   Left or right adjust string a, returning string b of max length
 *   width. adjustr blank pads the leading part of the string.
 */

extern integer len_trim( const char* a );
/*
 *  Return the length of the string excluding trailing blanks
 */

extern integer scan( const char* a, const char* set );
/*
 * F90 SCAN function. Consider a set of characters 'set' (null-terminated).
 * Which is the first character in set to appear in the string 'a'?
 * Returns the (1-based) index of that character within the string.
 * e.g.  scan( "ABCDE", "XFBG" ) = 2 since the first good character is
 * B and that is the second character in ABCDE.
 */

extern integer verify( const char* a, const char* set );
/*
 *  F90 VERIFY function. Consider a set of characters 'set' (null-terminated).
 *  Return the 1-based index of the first character in 'a' that does NOT appear in 'set', or
 *  0 if all characters in 'a' are members of set.
 *  e.g.  scan( "ABCDE", "XFAB" ) = 3 since C is not in set and is the third 
 * character in ABCDE.
 */

extern char* repeat( const char* a, const integer ncopies, char* result, const integer width );
/*
 * Repeat string 'a' ncopies times, into buffer 'result'.
 * Each copy is of size 'width'.
 * 
 */



/*
 * \section{Miscellaneous string functions}
 */

extern char cs_logic( const logical q );						
/* 
 * Return 'Y' or 'N' from logical 
 */

extern logical nancode( const char* buf );
/*
 *  True if the string is recognized by JCMLIB as a NaN code
 *  Specifically, '-' is recognized as a notation for NaN.
 */

extern void cgreek( const integer n, char* buf );
/*
 *   Return the nth Greek letter name.
 */

extern integer igreek( const char* buf );
/*
 *  If buf is a Greek letter name, return its number, else zero.
 */

extern integer iroman( const char* buf );
/*
 *  Convert a Roman numeral to an integer.
 */

extern void croman( const integer n, char* buf );
/*
 *  Compose a Roman numeral given an integer.
 */

/*
 * \section{Token parsing}
 */

extern void cs_advptr( char** buf );						
/* 
 * Advance to next non space character. Usage:
 *    ptr = buf;
 *    cs_advptr( &ptr );  // skipped over leading blanks
 */


extern void cs_get_c( char** buf, char* token, const integer siz );				
extern void cs_get_cs( char** buf, char* token, const integer siz );			
extern void cs_get_cl( char** buf, char* token, const integer siz );			
/*
 *  Peel off the next token from buf, advancing the buf pointer. 
 *  (pointer is set to null if the end of the buffer is reached).
 *  Place result in 'token' with max length 'siz'.
 *  In the cs_get_c routine, the token is terminated by space, comma, 
 *  null, or STRING SEP (default: semicolon). 
 *  In the cs_get_cs routine, the token is terminated by null or 
 *  STRING SEP.
 *  In the cs_get_cl routine, the token is terminated by null (ie the
 *  rest of the string is returned).
 *  
 */



extern double cs_get_d( char** buf );
extern real cs_get_r( char** buf );
extern integer cs_get_i( char** buf );
extern short cs_get_s( char** buf );
extern logical cs_get_q( char** buf );
/*
 *  Read a numeric value from the string buf. The numeric value must
 *  be terminated with a space, comma, or the end of the string.
 *  The pointer buf is advanced to the next non-blank space; at the end of
 *  string, the pointer is set to NULL.
 *  e.g.   while( ptr ) {  xval[n++] = cs_get_d( &ptr ); }
 *  will read a series of numbers from a string like
 *   ptr = "1.03 14.82E4 -0.03,4.82"
 */

extern double cs_getd_d( char** buf, const double xdefault );
extern real cs_getd_r( char** buf, const real xdefault );
extern short cs_getd_s( char** buf, const short xdefault );
extern integer cs_getd_i( char** buf, const integer xdefault );
/*
 *  The cs_getd routines are the same as cs_get, but return the provided
 *  default value if the string is blank. 
 */


extern void cs_put_c( const char* token, char** buf, integer* siz );	
extern void cs_put_cs( const char* token, char** buf, integer* siz );	
extern void cs_put_cn( const char* token, char** buf, integer* siz );
/*
 *  Place a token in the buffer, advancing the buffer pointer and
 *  decreasing the size remaining (siz). 
 *  The cs_put_c routine puts a space after the token; cs_put_cs
 *  puts STRING_SEP (semicolon); cs_put_cn puts a null after the token.
 */

extern void cs_put_d( const double token, char** buf, integer* siz );
extern void cs_put_r( const real token, char** buf, integer* siz );
extern void cs_put_s( const short token, char** buf, integer* siz );
extern void cs_put_i( const integer token, char** buf, integer* siz );
/*
 *  Put a numeric value in the buffer, formatting it with the
 *  write_d family of routines and then using cs_put_c.
 */

extern integer list_parse( char* buf, char** list, const integer maxitems );
/*
 *   Repeatedly apply cs_get_c to buf, generating a list of string tokens.
 *   You need to provied maxitems, the maximum size of the list.
 */

extern char cs_copy_token( char** buf, const char* seps, const char* pars, char* dest, const integer siz );
extern void cs_put_token( const char* buf, char** dest, const char sep, integer* siz );		/* Put buf to dest with sep, reduce siz */
extern void cs_get_token( char** buf, char* token, const integer mode, const integer siz );		/* Get token from buf by mode */
/*
 *  Low level routines used by the cs_get and cs_put family;
 *  allow specification of the token separator and parenthesis characters.
 *  Here for the record because I sometimes use them, but not yet intended
 *  for public use.
 */


/*
 * \section{Sorting}
 */


extern void csort_i( const integer* x, const integer n, integer* jsort );
extern void csort_d( const double* x, const integer n, integer* jsort );		/* Sort doubles */
extern void csort_r( const real* x, const integer n, integer* jsort );		/* Sort reals */
extern void csort_s( const short* x, const integer n, integer* jsort );
extern void csort_c( char** x, const integer n, integer* jsort );
/*
 *   Sort an array, returning an array jsort of sorted (0-based) array
 *   offsets. Thus x[ jsort[0] ] is the lowest value entry.
 *   Uses shell or heap sort algorithms depending on array size,
 *   as suggested in Press et al.
 */

extern void creindex_i( const integer* jsort, const integer n, integer* isort );		
/*
 *  Reindex an array. Given jsort, the output of one of the csort routines,
 *  make isort, the inverse index. The difference is:
 *     x[jsort[3]] is the fourth smallest entry in x
 *   jsort says: the sorted order is: first one is entry no. 23, second one is entry no. 14, etc.
 *  It allows us to traverse the original list in sorted order.
 *     isort[3] is the position of x[3] in the sorted list.
 *   isort says: entry no. 1 is mth in order, entry no 2 is nth in order, etc.
 *  It allows us to traverse the original list in original order and find
 *   out the rank of each entry.
 */

extern void cresort_s( const short* x, const integer n, short* y, integer* jsort );			/* Resort index to inverse index */
extern void cresort_d( const double* x, const integer n, double* y, integer* jsort );			/* Resort index to inverse index */
extern void cresort_r( const real* x, const integer n, real* y, integer* jsort );			/* Resort index to inverse index */
extern void cresort_i( const integer* x, const integer n, integer* y, integer* jsort );
/*
 *   Resort an array given the sort index. Given x and its sort index jsort,
 *   make y which is a copy of x but in sorted order.
 */

extern void sort_r( const real* x, const integer n, integer* jsort );		/* Sort reals */
extern void sort_s( const short* x, const integer n, integer* jsort );
extern void sort_d( const double* x, const integer n, integer* jsort );		/* Sort doubles */
extern void sort_i( const integer* x, const integer n, integer* jsort );
extern void sort_c( char** x, const integer n, integer* jsort );
extern void reindex_i( const integer* x, const integer n, integer* y );			/* Resort index to inverse index */
extern void resort_i( const integer* x, integer n, integer* y, integer* jsort );
extern void resort_s( const short* x, const integer n, short* y, integer* jsort );			/* Resort index to inverse index */
extern void resort_d( const double* x, const integer n, double* y, integer* jsort );			/* Resort index to inverse index */
extern void resort_r( const real* x, const integer n, real* y, integer* jsort );			/* Resort index to inverse index */
/*
 *   The sort, reindex, and resort functions are like the csort, creindex
 *   and cresort functions, but use (1-based) array indices instead of
 *   (0-based) array offsets, for compatibility with Fortran usage.
 */

/*
 * \section{Arithmetic and Math}
 */

extern integer rmod_d( const double x, double* y, const double r );		
extern integer rmod_i( const integer x, integer* y, const integer r );			
extern integer rmod_s( const short value, short* remainder, const short divisor );
extern integer rmod_r( const real value, real* remainder, const real divisor );
/*
 *  Return the quotient (as the function value) and the remainder (in y)
 *  of x modulo r.
 */


extern integer logbase_i( const integer x, const integer n );				/* Nearest integer to log base n */
extern double logbase_d( const double x, const integer n );			/* Log to base n */
extern real logbase_r( const real x, const integer n );
/*
 *    Return  log to base n of x. The integer version is used in
 *    one of the sort algorithms; the double version may be of more general use.
 */

extern void frac_r( const real value, integer* int_val, real* fraction );
extern void frac_d( const double value, integer* int_val, double* fraction );
/*
 *   Split value into integer and fractional part.
 */

extern integer min_i( const integer i, const integer j );				/* Min of two integers */
extern integer max_i( const integer i, const integer j );				/* Max of two integers */
/*
 *   Find the min or max of two integers. Usually done in C with a macro for
 *   efficiency, but I prefer the robustness of a real function.
 */

extern logical divisible( const integer a, const integer b );			/* A is divisible by B */
/*
 *   Return true if a is divisible by b.
 */

extern integer ceiling_d( const double x );
extern integer ceiling_r( const real x );
/*
 *  F90 CEILING function: return integer equal to or greater than x.
 */

extern integer floor_d( const double x );
extern integer floor_r( const real x );
/*
 *  F90 FLOOR function: return integer less than or equal to x.
 */

extern integer nint_d( const double x );
extern integer nint_r( const real x );
/*
 *   Nearest integer to x.
 */

extern integer modulo_i( const integer x, const integer base );
extern short   modulo_s( const short x, const short base );
extern real    modulo_r( const real x, const real base );
extern double  modulo_d( const double x, const double base );
/*
 *   Return remainder of x when divided by base.
 */

extern double round_d( const double x, const integer f );
extern real round_r( const real x, const integer f );
/*
 *   Round x to f significant digits (done by multiplying by a power, taking logs, 
 *   then using integer truncation).
 */

extern void order_r( real* x, real* y);
extern void order_d( double* x, double* y );
/*
 *   Swap the order of x and y if needed to put them in ascending order.
 *   Useful to enforce that a min and max range, e.g. for a plot, are in
 *   the correct order.
 */

extern logical clip_r( const real x, const real y, const real x1, const real x2, const real yy1, const real yy2 );
extern logical clip_d( const double x, const double y, const double x1, const double x2, const double yy1, const double yy2 );
extern logical clip_z( const complex z, const complex z1, const complex z2 );
extern logical clip_dz( const dcomplex z, const dcomplex z1, const dcomplex z2 );
/*
 *  Return true if  (x,y) is in the rectangle (x1,x2)*(y1,y2)
 */

extern double rlap_check_d( const double x, const double x1, const double x2, logical* left, logical* right );
extern real rlap_check_r( const real x, const real x1, const real x2, logical* left, logical* right );
/*
 *  Return the argument x if it is in the range [x1,x2]; otherwise
 *  return x1 or x2 if it is off the left or right end, and set the
 *  left or right flags accordingly.
 */

extern logical overlap_r( const real xx1, const real xx2, const real yy1, const real yy2,    real* u1, real* u2 );
extern logical overlap_d( const double xx1, const double xx2, const double yy1, const double yy2,    double* u1, double* u2 );
/*
 *  Intersect the ranges [xx1,xx2] and [yy1,yy2], result in [u1,u2].
 *  If there is no overlap, u2 = u1 and the function value is false.
 */


extern double lint_d( const double x1, const double x2, const double yy1, const double yy2, const double x0 );
extern real lint_r( const real x1, const real x2, const real yy1, const real yy2, const real x0 );
/*
 *   Linear interpolation: 
 *   return y =  yy1 + ( yy2 - yy1 ) * ( x - x1 ) / ( x2 - x1 )
 */

extern real rabs( const real x );
/*
 *   Return absolute value. This and following routines are single precision
 *   equivalents of the <math.h> standard routines.
 */

extern real rlog10( const real x );
/*
 *   Log to base 10
 */
extern real rsqrt( const real x );
/*
 *   Square root
 */

extern real rpow( const real x, const real n );
/*
 *   Raise to power
 */

extern real   ipow_r( const real x, const integer n );
extern double ipow_d( const double x, const integer n );
/*
 *   Raise x to integer power.
 */

extern real tpow_r( const integer n );
extern double tpow_d( const integer n );
/*
 *   Raise 10.0 to an integer power.
 */

extern double linterp_d( const double* x, const double* y, const integer n, const double x0 );
extern real linterp_r( const real* x, const real* y, const integer n, const real x0 );
/*
 *   Linearly interpolate in array. Consider arrays x[n], y[n] as a function
 *  y(x), and estimate y(x0), returning it as the function value.
 */

extern void movag_d( const double* yy0, double* y, const integer n, const integer k );
extern void movag_r( const real* yy0, real* y, const integer n, const integer k );
/*
 *  Calculate the moving average of yy0[n], returning the smoothed
 *  array in y[n]. The smoothing is a boxcar of width 2k+1.
 */

extern real poly_r( const real x, const real* coeffs, const integer degree );
extern double poly_d( const double x, const double* coeffs, const integer degree );
/*
 *  Evaluate a polynomial at argument x:
 *    value = coeffs[0] + x * coeffs[1] + x^2 * coeffs[2] + ... x^degree * coeffs[degree]
 */

extern void poleval_r( const real* x, real* y, const integer n, const real* coeffs, const integer degree );
extern void poleval_d( const double* x, double* y, const integer n, const double* coeffs, const integer degree );
/*
 *  Evaluate a polynomial at an array of arguments x[n], returning the
 *  evaluated array y[n]
 */

extern void lsfit_d( const double* x, const double* y, const integer n, double* a, double* b );
extern void lsfit_r( const real* x, const real* y, const integer n, real* a, real* b );
/*
 *   Least squares fit of y = ax + b to arrays x[n], y[n]
 */

extern real trap_r( const real* x, const real* y, const integer n );
extern double trap_d( const double* x, const double* y, const integer n );
/*
 *   Simple trapezium integral of y(x)
 */

extern real trap_log_r( const real* x, const real* y, const integer n );
extern double trap_log_d( const double* x, const double* y, const integer n );
/*
 *   Result is log of integral of 10^y dx.
 *   Useful for integrating flux spectra given in log units.
 */

extern real trap_dx_r( const real* y, const integer n, const real dx );
extern double trap_dx_d( const double* y, const integer n, const double dx );
/*
 *   Integral of array given at evenly spaced points dx apart.
 */

/*
 *  \section{Sexagesimal functions}
 */

extern void ssexin_r( const real v, integer* sign, integer* deg, integer* min, integer* sec, real* frac );
extern void ssexin_d( const double v, integer* sign, integer* deg, integer* min, integer* sec, double* frac );
/*
 *   Split a value into sign, degrees, minutes, seconds, fraction of sec
 */

extern real ssexout_r( const integer sign, const integer deg, const integer min, const integer sec, const real frac );
extern double ssexout_d( const integer sign, const integer deg, const integer min, const integer sec, const  double frac );
/*
 *  Combine sign, deg, min, sec, fraction into a single value.
 *  Sign should be +1 or -1.
 */

extern void wsexin_r( const real r, const integer w, logical* sign, integer* d, integer* m, integer* s, real* f );
extern void wsexin_d( const double r, const integer w, logical* sign, integer* d, integer* m, integer* s, double* f );

/*  
 *  Wrapped sexagesimal routine.
 *  The WSEXIN routine is like SSEXIN but scales the result by 360/w;
 *  it also uses logical instead of integer sign.
 *  thus if r = 270 and w = 24, you get  (TRUE, 18, 0, 0, 0.0).
 */

extern real wsexout_r( const logical sign, const integer d, const integer m, const integer s, const real f, const integer w );
extern double wsexout_d( const logical sign, const integer d, const integer m, const integer s, const double f, const integer w );
/*
 *  WSEXOUT is like SEXOUT but scales the result by 360/w; thus
 *  wsexout( TRUE, 18, 0, 0, 0.0, 24 ) = 270.0. 
 */

extern void psexin_d( const double r, logical* sg, integer* d, integer* m, integer* s, double* f );
extern void psexin_r( const real r, logical* sg, integer* d, integer* m, integer* s, real* f );
/*
 *  Packed sexagesimal routine. 
 *  PSEXIN takes a packed sexagesimal fraction and 
 *  parses it. Inverse of PSEXOUT.
 */

extern double psexout_d( const logical sg, const integer d, const integer m, const integer s, const double f );
extern real psexout_r( const logical sg, const integer d, const integer m, const integer s, const real f );
/*
 *  PSEXOUT packs a sexagesimal set of values into a single real as follows:
 *  psexout( FALSE, 18, 30, 00, 0.5 ) = -183000.5
 */

extern void sex_fmt( const logical sg, const integer d, const integer m, const integer s, const double fs, char* buf, const integer siz );
/*
 *  Low level routine to format a sexagesimal value into a string.
 */

extern char* sexwrite_hr_d( const double x, char* buf, const integer siz );
extern char* sexwrite_d( const double x, char* buf, const integer siz );
/*
 * The sexwrite routine writes a value in degrees into a sexagesimal format
 * string; sexwrite_hr_d converts it to hours and then writes it.
 */

/*
 * \section{Special functions}
 */

extern double eexp( const double x );
/*
 *   exp function, modified to fail 'softly' on large arguments.
 */

extern double esinh( const double x );
extern double ecosh( const double x );
/*
 *   cosh and sinh functions, modified to make sure they never
 *   fail on large or small arguments.
 */

extern double acosh( const double x );
extern double asinh( const double x );
/*
 *   Arccosh and arcsinh functions
 */

extern double dgauss( const double x, const double x0, const double sigma );
/*
 *   The gauss function:   exp ( - ( x-x0 / sigma ) ^2 )
 */

extern double bose_einstein( const double x );
/*
 *   The Bose-Einstein function:  1 / exp(x) - 1
 */

extern double dplanckn( const double x, const integer n );
/*
 *   The generalized Planck function:  x^n / ( exp x - 1 )
 */

extern double hexagam( const double x );
extern double acych( const double x );
extern double acyc( const double x );
extern double E1_func( const double x );
/*
 *   Obscure special functions I've occasionally used.
 */



/*
 * \section{Complex number routines}
 */


extern complex zset( const real x, const real y );
extern dcomplex dzset( const double x, const double y );
/*
 *  Construct complex number from real and imag parts
 */
extern real rreal( const complex z );
extern double dreal( const dcomplex z );
extern real rimag( const complex z );
extern double dimag( const dcomplex z );
/*
 *   Return real and imaginary parts of complex numbers
 */

extern complex zconj( const complex z1 );
extern dcomplex dzconj( const dcomplex z1 );
/*
 *   Return complex conjugate
 */

extern complex zmult( const complex z1, const complex z2 );
extern dcomplex dzmult( const dcomplex z1, const dcomplex z2 );
/*
 *  Multiply two complex numbers
 */
extern complex zadd( const complex z1, const complex z2 );
extern dcomplex dzadd( const dcomplex z1, const dcomplex z2 );
/*
 *  Add two complex numbers
 */
extern void zinc( const complex z1, complex* z2 );
extern void dzinc( const dcomplex z1, dcomplex* z2 );
/*
 *  Add z1 to z2, result in z2. 
 */

extern dcomplex dzprod( const dcomplex z1, const dcomplex z2 );
extern complex zprod( const complex z1, const complex z2 );
/*
 *   Multiply two complex numbers in conjugate form (z1 * z2-conj)
 */

extern logical gt_z( const complex z1, const complex z2 );
extern logical ge_z( const complex z1, const complex z2 );
extern logical lt_z( const complex z1, const complex z2 );
extern logical le_z( const complex z1, const complex z2 );
extern logical gt_dz( const dcomplex z1, const dcomplex z2 );
extern logical ge_dz( const dcomplex z1, const dcomplex z2 );
extern logical lt_dz( const dcomplex z1, const dcomplex z2 );
extern logical le_dz( const dcomplex z1, const dcomplex z2 );
/*
 *   Inequality operators for complex numbers
 */



/*
 *   \section{Extreme value handling}
 *
 *  The current package does not rely on IEEE NaN handling, but reserves
 *  other special values instead. As IEEE becomes more common, I hope to
 *  replace these routines with IEEE compliant ones.
 *
 */

extern void nan_init( void );
/*
 *  Reinitialize NaN buffers, usually just rely on jcmlib_init to call this.
 */

extern logical qnan_i( const integer i );
extern logical qnan_s( const short i );
extern logical qnan_r( const real x );
extern logical qnan_d( const double x );
/*
 *   True if argument is a NaN
 */

extern logical qinf_i( const integer i );
extern logical qinf_s( const short i );
extern logical qinf_r( const real x );
extern logical qinf_d( const double x );
/*
 *   True if argument is +Inf or -Inf
 */

extern double nan_d( void );
extern real nan_r( void );
extern short nan_s( void );
extern integer nan_i( void );
/*
 *   Return NaN value of appropriate type
 */

extern double infinity_d( const integer i );
extern real  infinity_r( const integer i );
extern short infinity_s( const integer i );
extern integer infinity_i( const integer i );
/*
 *    Returns +Inf or -Inf depending on sign of argument
 */

extern void nan_set( char* cnan, char* cinf, char* cninf, const char mode );
/*
 *   Reset bit patters for NaN, +Inf, -Inf. The bit pattern strings
 *   are in hex; the mode argument is 'S', 'I', 'R', or 'D' for
 *   the particular data type of NaN to be set.
 */

/*
 *  \section{Random numbers}
 */

extern void sys_rand_init( void );
/*
 *  Reinitializes the random number routine by getting
 *  a new seed from the POSIX time call. This is called automatically
 *  in jcmlib_init.
 */

extern double sys_rand( void );
/*
 *  Return a random number uniformly distributed between 0.0 and 1.0
 */

/*
 *  \section{Circular and trig functions}
 */

extern real   radin_r( const real x ) ;
extern double radin_d( const double x ) ;

/* 
 *  Convert from degrees to radians
 */

extern real   radout_r( const real x );
extern double radout_d( const double x );
/*
 *   Convert from radians to degrees
 */

extern real   circ_r( const real x );
extern double circ_d( const double x );
/*
 *  Return argument forced to range 0,360
 */

extern real circ_ew_r( const real x );
extern double circ_ew_d( const double x );
/*
 *  Return argument forced to range -180, 180
 */

extern real circ_rad_r( const real x );
extern double circ_rad_d( const double x );
/*
 *   Return argument forced to range  0, 2pi
 */

extern real circ_ew_rad_r( const real x );
extern double circ_ew_rad_d( const double x );
/*
 *   Return argument forced to range -pi, pi.
 */

extern double argd_d( const double x, const double y );
extern real argd_r( const real x, const real y );
extern real argd_z( const complex z );
extern double argd_dz( const dcomplex dz );
/*
 *  Return the argument of a complex number given its real and imaginary parts.
 *  Alternatively, return the polar coordinate azimuth given the x and y values.
 *  The routine  argd( x, y ) is the same as atan2( y, x ) and it should
 *  be used in preference to arc tangent routines atan( y/x ) which lose
 *  information about which quadrant is meant.
 */

extern real rsind( const real x );
extern double dsind( const double x );
extern real rcosd( const real x );
extern double dcosd( const double x );

/*
 *  sin and cos functions with arguments in degrees.
 */

extern real rasind( const real x );
extern double dasind( const double x );
extern real racosd( const real x );
extern double dacosd( const double x );
/*
 *  Arcsin and arccos functions with results in degrees, in range 0 to 360.
 */

extern void rotate2_d( const double x, const double y, const double roll, double* xr1, double* yr1 );
/*
 *   Rotate x,y through angle roll, giving xr1, yr1.
 */

extern integer iplus( const logical q );
/*
 *  Return 1 if true, -1 if false. Used for North/South angular
 *  encoding.
 */

extern real spharc_r( const real a1, const real b1, const real a2, const real b2 );
extern double spharc_d( const double a1, const double b1, const double a2, const double b2 );
/*
 *   Calculate the distance (arc on the sphere)
 *   between two pairs of latitude, longitude values
 *   (arguments and result in degrees).
 */

/*
 *  \section{3-Vector routines}
 */

extern double** vec_alloc_d( const integer n );
/*
 *    Allocate an array of n 3-vectors.
 */

extern void vec_free_d( double** v );
/*
 *   Free the 3-vector array.
 */

extern real vdot_r( const real* a, const real* b );
extern double vdot_d( const double* a, const double* b );
/*
 *  Scalar product of 2 3-vectors 
 */

extern real vnorm_r( const real* a );
extern double vnorm_d( const double* a );
/*
 *  Euclidean norm (modulus) of a 3-vector.
 */

extern void vcross_r( const real* a, const real* b, real* c );
extern void vcross_d( const double* a, const double* b, double* c );
/*
 *  Vector (cross) product of two 3-vectors, returned in a third.
 */

extern real vmaxmod_r( const real* r );
extern double vmaxmod_d( const double* r );
/*
 *  Maxmod modulus of a 3-vector: the largest absolute value of
 *  any of its components. 
 */

extern real vunit_r( real* r );
extern double vunit_d( double* r );
/*
 *  Convert a 3-vector to a unit vector, and return its modulus as
 *  the function value.
 */


extern void vcmult_r( real* p, const real c );
extern void vcmult_d( double* p, const double c );
/*
 *  Multiply a vector by a constant, in place.
 */


extern void vadd_r( const real* p, real* r );
extern void vadd_d( const double* p, double* r );
/*
 *   Add vector p to vector r, result in r.
 */

extern void vsub_r( const real* p, real* r );
extern void vsub_d( const double* p, double* r );
/*
 *   Subract vector p from vector r, result in r.
 */

extern void vzero_r( real* p );
extern void vzero_d( double* p );
/*
 *   Set a vector to zero. See also VZERO_D symbol.
 */

extern logical vnull_r( const real* v );
extern logical vnull_d( const double* v );
/*
 *   Test if a 3-vector is zero.
 */

extern void vcopy_r( const real* p, real* r );
extern void vcopy_d( const double* p, double* r );
/*
 *   Copy vector p to vector r.
 */

extern void mat_prex_r( real** aa, const real* y, real* x );
extern void mat_prex_d( double** aa, const double* y, double* x );
/*
 *   Premultiply matrix aa by vector y, result in x.
 */ 

extern void mat_postx_r( real** aa, const real* y, real* x );
extern void mat_postx_d( double** aa, const double* y, double* x );
/*
 *    Postmultiply matrix aa by vector y, result in x.
 */

extern void mat_mult_r( real** aa, real** bb, real** cc );
extern void mat_mult_d( double** aa, double** bb, double** cc );
/*
 *  Matrix multiply aa and bb, result in cc
 */

extern void vec_r( const real r, const real theta, const real phi, real* p );
extern void vec_d( const double r, const double theta, const double phi, double* p );
/*
 *   Convert from spherical polars r, theta, phi to cartesian 3-vector.
 *   Angles in degrees.
 */

extern void polar_r( const real* p, real* r, real* theta, real* phi );
extern void polar_d( const double* p, double* r, double* theta, double* phi );
/*
 *   Convert from cartesian to spherical polars
 */

extern real polarll_r( const real* p, real* llong, real* lat );
extern double polarll_d( const double* p, double* llong, double* lat );
/*
 *   Convert from (unit) 3-vector to latitude and longitude.
 */

extern void unitll_d( const double llong, const double lat, double* p );
extern void unitll_r( const real llong, const real lat, real* p );
/*
 *   Convert from latitude and longitude to unit 3-vector.
 */

extern void vec6_r( const real r, const real theta, const real phi, const real rrdot, const real tdot, const real phidot, real* p );
extern void vec6_d( const double r, const double theta, const double phi, const double rrdot, const double tdot, const double phidot, double* p );
/*
 *   Convert from spherical polar coords and their derivatives to
 *   position-velocity 6-vector.
 */

extern void polar6_r( const real* p, real* r, real* theta, real* phi, real* rrdot, real* tdot, real* phidot );
extern void polar6_d( const double* p, double* r, double* theta, double* phi, double* rrdot, double* tdot, double* phidot );
/*
 *   Convert from position-velocity 6-vector to spherical polar coords and
 *   their derivatives.
 */

/*
 *
 * \section{Array memory management}
 */

extern void mfree( void* p );
/*
 *   Safe memory free:  if ( p ) free( p );
 *   Sometimes you have a substructure which may or may not have
 *   been allocated; this routine says 'if it's there, then free it.'
 *   POSIX says nothing happens if you free( NULL ), but on some systems
 *   you seem to get a segv.
 */

extern char** alloc_array_c( const integer size, const integer n );
/*
 * Allocate an array of n character strings each of size+1 bytes,
 * and initialize each of them to null.
 *   char** p = alloc_array_c( 32, 40 );
 *   strcpy( p[38], "<= 32 bytes here" );
 */

extern void free_array_c( char** word );
/*
 * Free space allocated by alloc_array_c.
 */



extern short** ar_alloc_s( const integer dim, const integer n );
extern integer** ar_alloc_i( const integer dim, const integer n );
extern real** ar_alloc_r( const integer dim, const integer n );
extern double** ar_alloc_d( const integer dim, const integer n );

/* 
 * Simple alloc of arrays of n vectors of size dim.
 * p = ar_alloc_d( 20, 4 )  gives p[i] as an array
 * of four doubles, so that q = p[i] has elements
 * q[0] to q[3].
 * See also vec_alloc_d in section on 3-vectors.
 */

extern void ar_free_s( short** p );
extern void ar_free_i( integer** p );
extern void ar_free_r( real** p );
extern void ar_free_d( double** p );
/*
 *  Free the space reserved by ar_alloc routines.
 */

extern real*    alloc_array_r(     const integer start,    const integer end);
extern integer* alloc_array_i(     const integer start,    const integer end);
extern short*   alloc_array_s(     const  integer start,    const integer end);
extern double*  alloc_array_d(     const integer start,    const integer end);

/*
 *   These routines allocate arrays more in the Fortran spirit.
 *   Allocate an array whose index goes from start to end. 
 *   Thus:
 *     integer* p;
 *     p = alloc_array_i( start, end );
 *   gives an array of end-start+1 elements called p[start] to p[end].
 */

extern void free_array_r(  real* v);
extern void free_array_d(  double* v);
extern void free_array_i(  integer* v);
extern void free_array_s(  short* v);

/* 
 *  Free the arrays allocated by alloc_array.
 */

extern real**    alloc_matrix_r(   const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern double**  alloc_matrix_d(   const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern integer** alloc_matrix_i(   const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern short**   alloc_matrix_s(   const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern complex** alloc_matrix_z(   const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern dcomplex** alloc_matrix_dz(   const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);

/*
 *  Allocate 2D array whose bounds are from xmin to xmax, ymin to ymax 
 */

extern void free_matrix_s( short** matrix, const integer xmin, const integer ymin);
extern void free_matrix_i( integer** matrix, const integer xmin, const integer ymin);
extern void free_matrix_r( real** matrix, const integer xmin, const integer ymin);
extern void free_matrix_d( double** matrix, const integer xmin, const integer ymin);
extern void free_matrix_z( complex** matrix, const integer xmin, const integer ymin);
extern void free_matrix_dz( dcomplex** matrix, const integer xmin, const integer ymin);

/* Free arrays allocated by the alloc_matrix routines; you need to tell
 * them the minimum bounds, because they don't remember! 
 */

extern void ar_transpose_s(   const short** x,  short** y,  const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern void ar_transpose_i(   const integer** x,  integer** y,  const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern void ar_transpose_r(   const real** x,  real** y,  const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern void ar_transpose_d(   const double** x,  double** y,  const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern void ar_transpose_q(   const logical** x,  logical** y,  const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
extern void ar_transpose_dz(   const dcomplex** x,  dcomplex** y,  const integer xmin,  const integer xmax, const integer ymin,  const integer ymax);
extern void ar_transpose_z(   const complex** x,  complex** y,  const integer xmin,  const integer xmax,  const integer ymin,  const integer ymax);
/*
 *  Make the transpose of an array whose bounds are given by (xmin,xmax) by (ymin,ymax).
 *  See the alloc_array functions.
 */

/*
 *
 * \section{Basic array operations}
 */

extern void ar_zero_i( integer* a, const integer siz );
extern void ar_zero_r( real* a, const integer siz );
extern void ar_zero_s( short* a, const integer siz );
extern void ar_zero_d( double* a, const integer siz );
extern void ar_zero_c( char** x, const integer n );
extern void ar_zero_dz( dcomplex* z, const integer n );
extern void ar_zero_z( complex* z, const integer n );
/*
 *    Zero an array of the given size.
 */

extern void ar_copy_d( const double* a, double* b, const integer siz );
extern void ar_copy_i( const integer* a, integer* b, const integer siz );
extern void ar_copy_s( const short* a, short* b, const integer siz );
extern void ar_copy_r( const float* a, float* b, const integer siz );
extern void ar_copy_q( const logical* a, logical* b, const integer siz );
extern void ar_copy_c( const char** x, char** y, const integer n );
extern void ar_copy_z( const complex* z1, complex* z2, const integer size );
extern void ar_copy_dz( const dcomplex* z1, dcomplex* z2, const integer size );
/*
 *   Copy one array to another.
 */

/* Array.f */

extern double ar_dot_d( const double* x, const double* y, const integer siz );
extern real ar_dot_r( const real* x, const real* y, const integer n );
extern short ar_dot_s( const short* x, const short* y, const integer n );
extern integer ar_dot_i( const integer* x, const integer* y, const integer siz );
extern dcomplex ar_dot_dz( const dcomplex* x, const dcomplex* y, const integer size );
extern complex ar_dot_z( const complex* x, const complex* y, const integer size );
/*
 *  Evaluate the scalar Euclidean product of two arrays.
 */

extern void ar_bounds_r( const real* x, const integer n, real* xmin, real* xmax );
extern void ar_bounds_s( const short* x, const integer n, short* xmin, short* xmax );
extern void ar_bounds_i( const integer* x, const integer n, integer* xmin, integer* xmax );
extern void ar_bounds_d( const double* x, const integer n, double* xmin, double* xmax );
/*
 *   Evaluate the minimum and maximum values present in an array
 */

extern void ar_stats_s( const short* x, const integer n, real* mean, real* var );
extern void ar_stats_r( const real* x, const integer n, real* mean, real* var );
extern void ar_stats_i( const integer* x, const integer n , double* mean, double* var );
extern void ar_stats_d( const double* x, const integer n , double* mean, double* var );
/*
 *   Evaluate the mean and variance of an array. Note that the results
 *   are always of floating point type.
 */

extern void ar_fill_c( char** x, const integer n, const char* x0 );
extern void ar_fill_i( integer* x, const integer n, const integer x0 );
extern void ar_fill_s( short* x, const integer n, const short x0 );
extern void ar_fill_d( double* x, const integer n, const double x0 );
extern void ar_fill_r( real* x, const integer n, const real x0 );
extern void ar_fill_z( complex* z, const integer n, const complex z0 );
extern void ar_fill_dz( dcomplex* z, const integer n, const dcomplex z0 );
/*
 *   Fill an array with the same constant value for each element.
 */

extern void ar_cmult_s( short* x, const integer n, const short x0 );
extern void ar_cmult_i( integer* x, const integer n, const integer x0 );
extern void ar_cmult_r( real* x, const integer n, const real x0 );
extern void ar_cmult_d( double* x, const integer n, const double x0 );
extern void ar_cmult_z( complex* z, const integer n, const complex z0 );
extern void ar_cmult_dz( dcomplex* z, const integer n, const dcomplex z0 );
/*
 *  Multiply an array by a constant, in place.
 */

extern void ar_cadd_s( short* x, const integer n, const short x0 );
extern void ar_cadd_i( integer* x, const integer n, const integer x0 );
extern void ar_cadd_r( real* x, const integer n, const real x0 );
extern void ar_cadd_d( double* x, const integer n, const double x0 );
extern void ar_cadd_z( complex* z, const integer n, const complex z0 );
extern void ar_cadd_dz( dcomplex* z, const integer n, const dcomplex z0 );
/*
 *  Add a constant to an array, in place.
 */

extern void ar_add_d( const double* x, double* y, const integer n );
extern void ar_add_s( const short* x, short* y, const integer n );
extern void ar_add_r( const real* x, real* y, const integer n );
extern void ar_add_i( const integer* x, integer* y, const integer n );
extern void ar_add_z( const complex* z1, complex* z2, const integer n );
extern void ar_add_dz( const dcomplex* z1, dcomplex* z2, const integer n );
/*
 *   Add two arrays, with the result in y.
 */

extern void ar_sub_s( const short* x, short* y, const integer n );
extern void ar_sub_i( const integer* x, integer* y, const integer n );
extern void ar_sub_r( const real* x, real* y, const integer n );
extern void ar_sub_d( const double* x, double* y, const integer n );
extern void ar_sub_dz( const  dcomplex* z1, dcomplex* z2, const integer n );
extern void ar_sub_z( const complex* z1, complex* z2, const integer n );
/*
 *   Subtract array x from y, result in y.
 */

extern void ar_wadd_s( const short* x1, const short* x2, const short w1, const short w2,short* y, const integer n );
extern void ar_wadd_i( const integer* x1, const integer* x2, const integer w1, const integer w2,integer* y, const integer n );
extern void ar_wadd_r( const real* x1, const real* x2, const real w1, const real w2,real* y, const integer n );
extern void ar_wadd_d( const double* x1, const double* x2, const double w1, const double w2,double* y, const integer n );
extern void ar_wadd_z( const complex* z1, const complex* z2, const complex w1, const complex w2,  complex* z, const integer n );
extern void ar_wadd_dz( const dcomplex* z1, const dcomplex* z2, const dcomplex w1, const dcomplex w2,  dcomplex* z, const integer n );
/*
 *  Weighted array addition.
 *  Evaluate  y = w1 * x1 + w2 * x2.
 */

extern void ar_bin_s( short* x, const integer n, const short x0, const short dx );
extern void ar_bin_r( real* x, const integer n, const real x0, const real dx );
extern void ar_bin_i( integer* x, const integer n, const integer x0, const integer dx );
extern void ar_bin_d( double* x, const integer n, const double x0, const double dx );
extern void ar_bin_dz( dcomplex* z, const integer n, const dcomplex z0, const dcomplex dz );
extern void ar_bin_z( complex* z, const integer n, const complex z0, const complex dz );
/*
 *   Fill an array with equally spaced values starting at x0 and spaced by dx.
 */

extern integer ar_cbisect_d( const double* x, const integer n, const double x0 );
extern integer ar_cbisect_r( const real* x, const integer n, const real x0 );
extern integer ar_cbisect_s( const short* x, const integer n, const short x0 );
extern integer ar_cbisect_i( const integer* x, const integer n, const integer x0 );
/*
 *  Find the array offset i, such that x[i] <= x0 < x[i+1].
 *  i is in the range 0 to n-1.
 */

extern integer ar_bisect_d( const double* x, const integer n, const double x0 );
extern integer ar_bisect_r( const real* x, const integer n, const real x0 );
extern integer ar_bisect_s( const short* x, const integer n, const short x0 );
extern integer ar_bisect_i( const integer* x, const integer n, const integer x0 );
/* 
 * Find the (1-based, Fortran-style) array index i, preceding x0, such that
 *   x[i-1] <= x0 < x[i]. i is between 1 and n.
 */

extern void ar_copy_si( const   short* x,  integer* y,  const integer n);
extern void ar_copy_is( const  integer* x, short* y, const integer n);
extern void ar_copy_rd( const   real* x,  double* y,  const integer n);
extern void ar_copy_dr( const double* x, real* y, const integer n);
extern void ar_copy_dzz( const dcomplex* x, complex* y, const integer n);
extern void ar_copy_zdz( const complex* x, dcomplex* y, const integer n);
/*
 *  Change precision of an array: copy from short to integer, double to
 *  real, etc.
 */


extern integer match_r(   const real value,  const real* list,  const integer n );
extern integer match_i(   const integer value,  const integer* list,  const integer n );
extern integer match_d(   const double value,  const double* list,  const integer n );
extern integer match_s(   const short value,  const short* list,  const integer n );
extern integer match_c( const char* buf, char** array, const integer n );				/* Match string in array of strings */
/*
 *   Look for value in list, testing for equality of value with
 *   list[i] for each i. Return the value i+1  (i.e. the Fortran 1-based
 *   index) for the first such match.
 */

extern integer cmatch_r(   const real value,  const real* list,  const integer n );
extern integer cmatch_i(   const integer value,  const integer* list,  const integer n );
extern integer cmatch_d(   const double value,  const double* list,  const integer n );
extern integer cmatch_s(   const short value,  const short* list,  const integer n );
extern integer cmatch_c( const char* buf, char** array, const integer n );				/* Match string in array of strings */
/*
 *   Look for value in list, testing for equality of value with
 *   list[i] for each i. Return the value i+1  (i.e. the C 0-based
 *   offset) for the first such match.
 */



/*
 *
 * \section{F90 style array operations}
 *  
 *  The following routines are provided as analogs to the Fortran 90
 *  intrinsic functions.
 */

extern integer maxval_i(  const integer* x,	  const integer  n	 );
extern short maxval_s(  const short* x, const integer  n );
extern real maxval_r(  const real* x,	 const integer  n	 );
extern double maxval_d( const double* x,	const  integer  n	 );

/*  Return the maximum value in an array of size n */

extern integer minval_i(  const integer* x,	const 	 integer  n		 );
extern short minval_s(  const short* x, const integer  n );
extern real minval_r(  const real* x,	 const integer  n	 );
extern double minval_d( const double* x,const 	 integer  n	 );

/* 
 * Return the minimum value in an array of size n 
 */

extern integer minloc_i(  const integer* x,	const  integer  n	 );
extern integer minloc_s(  const short* x,	const  integer  n	 );
extern integer minloc_r(  const real* x,	const  integer  n	 );
extern integer minloc_d(  const double* x,	const 	 integer  n	);
/* 
 *  Return the Fortran (1-based) index of the minimum value in the array.
 *  Thus,  i = minloc_d( x, n ) gives x[i-1] as the minimum. 
 */

extern integer maxloc_i( const integer* x,	 const integer  n	 );
extern integer maxloc_s( const  short* x,	 const integer  n	 );
extern integer maxloc_r( const  real* x,	 const integer  n	 );
extern integer maxloc_d( const  double* x,	 const integer  n	 );
/* 
 *  Return the Fortran (1-based) index of the maximum value in the array.
 *  Thus,  i = maxloc_d( x, n ) gives x[i-1] as the maximum. 
 */

extern void ar_minmax_d(  const double* x,	const  integer  n,	 double* xmin,	 double* xmax,	 integer*  imin, integer*  imax	);
extern void ar_minmax_r(  const real* x,	const  integer  n,	 real* xmin,	 real* xmax,	 integer*  imin, integer*  imax	);
extern void ar_minmax_i(  const integer* x,	const  integer  n,	 integer* xmin,	 integer* xmax,	 integer*  imin, integer*  imax	);
extern void ar_minmax_s(  const short* x,	const  integer  n,	 short* xmin,	 short* xmax,	 integer*  imin, integer*  imax	);
/*
 *   The minloc, maxloc, minval and maxval routines are wrappers
 *   for this routine which returns the minimum and maximum values
 *   in the array and the (1-based) indices of those values.
 */

extern integer ar_product_i(  const integer* a,	 const integer n 	 );
extern short   ar_product_s(  const short* a,	const  integer n 	 );
extern real    ar_product_r(  const real* a,	 const integer n 	 );
extern double  ar_product_d(  const double* a,const 	 integer n 	 );
extern dcomplex ar_product_dz(  const dcomplex* a,const 	 integer n );
extern complex ar_product_z(  const complex* a,	 const integer n 	 );

/*
 *  Like the F90 PRODUCT intrinsic, this returns the product of all the
 *  elements in the array.
 */

extern integer ar_sum_i(  const integer* a,	 const integer n 	 );
extern short   ar_sum_s(  const short* a, const integer n  );
extern real    ar_sum_r(  const real* a,	const  integer n 	 );
extern double  ar_sum_d( const double* a,	const  integer n 	 );
extern dcomplex ar_sum_dz(  const dcomplex* a,	const  integer n 	 );
extern complex ar_sum_z(  const complex* a,	const  integer n 	 );
/*
 *   Like the F90 SUM intrinsic, this returns the sum of the array elements.
 */

extern integer merge_i(  const integer i, const integer j, const logical mask);
extern short   merge_s(  const short i, const short j, const logical mask);
extern real    merge_r(  const real i, const real j, const logical mask);
extern double  merge_d(  const double i, const double j, const logical mask);
extern char *  merge_c(  char * i, char * j, const logical mask);
extern complex merge_z( const complex z1, const complex z2, const logical mask );
extern dcomplex merge_dz( const dcomplex z1, const dcomplex z2, const logical mask );
/*
 *  These routines, a trivial wrapper for the C ? operator, provide
 *  some of the function of F90 MERGE; they return j if mask is true, else i.
 */

extern void ar_merge_s(  const short* true, const short* false, short* merge, const logical* mask, const integer size);
extern void ar_merge_i(  const integer* true, const integer* false, integer* merge, const logical* mask, const integer size);
extern void ar_merge_d(  const double* true, const double* false, double* merge, const logical* mask, const integer size);
extern void ar_merge_r(  const real* true, const real* false, real* merge, const logical* mask, const integer size);
extern void ar_merge_c(  const char** true, const char** false, char** merge, const logical* mask, const integer size);
extern void ar_merge_cp(  char** true, char** false, char** merge, const logical* mask, const integer size);
extern void ar_merge_z( const complex* z1, const complex* z2, complex* z, const logical* mask, const integer size );
extern void ar_merge_dz( const dcomplex* z1, const dcomplex* z2, dcomplex* z, const logical* mask, const integer size );


/*
 *  Merge two arrays based on a third logical-valued array.
 *  For each element in mask which is true, the corresponding element in merge
 *  is set to the corresponding value in the array 'true', otherwise it
 *  is set to the entry in 'false'.
 *  The merge_c version actually makes copies of the strings, while
 *  the merge_cp version just copies the pointers.
 */

extern integer ar_pack_s( const short* x,	 const logical* mask,	 short* y,	 const integer n	);
extern integer ar_pack_i( const integer* x,	const 	 logical* mask, integer* y, const integer n);
extern integer ar_pack_r( const real* x,	 const logical* mask,	 real* y,	 const integer n	);
extern integer ar_pack_d( const double* x,	 const logical* mask,	 double* y,	 const integer n	);
extern integer ar_pack_c( const char** x,	 const logical* mask,	 char** y,	 const integer n	);
extern integer ar_pack_cp( char** x,	 const logical* mask,	 char** y,	 const integer n	);
extern integer ar_pack_dz( const dcomplex* z1, const logical* mask, dcomplex* z2, const integer n );
extern integer ar_pack_z( const complex* z1, const logical* mask, complex* z2, const integer n );
/*
 *  F90 PACK function. Based on the entries in the mask[n] array, take
 *  the x[n] array to make a new array y[m] containing only the 'true' entries.
 *  Return the value of m as the function value.
 *  The cp version works on char pointers, while the c version copies the
 *  actual strings.
 */


extern integer ar_unpack_s( const short* x,	 const logical* mask,	 short* y,	 const short  x0, 	 const integer n	);
extern integer ar_unpack_i( const integer* x,	 const logical* mask,	 integer* y,	 const integer  x0, 	 const integer n	);
extern integer ar_unpack_r( const real* x,	 const logical* mask,	 real* y,	 const real  x0, 	 const integer n	);
extern integer ar_unpack_d( const double* x,	 const logical* mask,	 double* y,	 const double  x0, 	 const integer n	);
extern integer ar_unpack_cp( char** x,	 const logical* mask,	 char** y,	 char*  x0, 	 const integer n	);
extern integer ar_unpack_c( const char** x,	 const logical* mask,	 char** y,	 const char*  x0, 	 const integer n	 );
extern integer ar_unpack_dz( const dcomplex* z1, const logical* mask, dcomplex* z2, const dcomplex z0, const integer n );
extern integer ar_unpack_z( const complex* z1, const logical* mask, complex* z2,  const complex z0, const integer n );
/*
 *  F90 UNPACK function. Based on the entries in the mask[n] array, make the y[n]
 *  array by placing the next unused value of x[m] if mask is true, or the value
 *  of x0 if mask is false. Returns the number of entries from x used (the number
 *  of true elements of mask) as the function value.
 */



extern void ar_cshift_s(  const short* x, short* y, const integer n, const integer shift);
extern void ar_cshift_i(  const integer* x, integer* y, const integer n, const integer shift);
extern void ar_cshift_r(  const real* x, real* y, const integer n, const integer shift);
extern void ar_cshift_d(  const double* x, double* y, const integer n, const integer shift);
extern void ar_cshift_cp(  char** x, char** y, const integer n, const integer shift);
extern void ar_cshift_c(  const char** x, char** y, const integer n, const integer shift);
extern void ar_cshift_dz(  const dcomplex* x, dcomplex* y, const integer n, const integer shift);
extern void ar_cshift_z(  const complex* x, complex* y, const integer n, const integer shift);
/*
 * F90 CSHIFT function. Circular shift right by shift elements,
 */

extern void ar_eoshift_s(  const short* x, short* y, const integer n, const integer shift, const short fill);
extern void ar_eoshift_i(  const integer* x, integer* y, const integer n, const integer shift, const integer fill);
extern void ar_eoshift_r(  const real* x, real* y, const integer n, const integer shift, const real fill);
extern void ar_eoshift_d(  const double* x, double* y, const integer n, const integer shift, const double fill);
extern void ar_eoshift_cp( char** x, char** y, const integer n, const integer shift, char* fill);
extern void ar_eoshift_c(  const char** x, char** y, const integer n, const integer shift, const char* fill);
extern void ar_eoshift_dz(  const dcomplex* x, dcomplex* y, const integer n, const integer shift, const dcomplex fill);
extern void ar_eoshift_z(  const complex* x, complex* y, const integer n, const integer shift, const complex fill);
/*
 *  F90 EOSHIFT function. End-off shift right by shift elements, fill end values
 *  with 'fill' value.
 */





/*
 *  \section{Logical arrays}
 */

extern void ar_not_q( logical* mask, const integer n );
/*
 *  Complement all entries in array.
 */

extern void ar_true_q( logical* mask, const integer n );
extern void ar_false_q( logical* mask, const integer n );
/*
 *   Set all entries to true or false.
 */

extern logical ar_all_q( const logical* mask, const integer n );
/*
 *  True if all entries are true.
 */

extern logical ar_any_q( const logical* mask, const integer n );
/*
 *   True if any entries are true.
 */

extern integer ar_count_q( const logical* mask, const integer n );
/*
 * Count the number of true entries in the array
 */

extern void ar_and_q( const logical* x, logical* y, const integer n );
extern void ar_or_q( const logical* x, logical* y, const integer n );
/*
 *  Perform logical and or or of x and y (how'd you like those
 *  conjunctions?) and put result in y.
 */


/*
 * \section{FIO: general file I/O}
 *
 * The FIO system maintains internal information on open files.
 * It supports text files, byte stream files, and raw mode files;
 * all files are given a unique integer file ID (fid) which you use
 * to refer to them. FIO looks for files in the current directory
 * and in the list of directories given in the JCMPATH environment
 * variable.
 * The simplest use of FIO is for text files (see the next section).
 * An example to copy a text file might be:
 *
 *   card line;
 *   integer rfid = fio_open_r( "my.dat" );
 *   integer wfid = fio_open_w( "out.dat" );
 *   if ( rfid > 0 && wfid > 0 ) {
 *    while ( fio_read( rfid, line, CSIZE ) ) {
 *     fio_write( wfid, line );
 *    }
 *   }
 *   fio_close( rfid );
 *   fio_close( wfid );
 */

extern void fio_init( void );
/*
 *  You don't need to call this, it reinitializes the file I/O system
 *  but that's done in jcmlib_init().
 */

extern logical fio_inquire( char* name );
/*
 *  Test whether file exists.
 */

extern integer fio_access( const char* name );
/*
 *  Test permissions of file. Uses POSIX stat; 
 *  returns one of the following four special values:
 *   FIO_NOFILE     File not found
 *   FIO_NOPERM     File exists but you have no permissions on it
 *   FIO_READONLY   File is readonly to you
 *   FIO_READWRITE  It's all yours!
 */

extern void fio_get_name( integer file, char* name, const integer maxlen );	
/*
 *  Return the name of the file given its file ID (see open calls below)
 */

extern void fio_get_type( integer file, char* type, const integer maxlen );	
/* 
 *   Report the type of file
 */

extern integer fio_status( integer file, integer mode );
/*
 *   Return the status of a file. If mode = FIO_CHECK (=1), just report; if mode = 
 *   FIO_CLEAR (=0),  clear the status to zero.
 */


extern void fio_close( integer file_no );					
/*
 *   Close any FIO file, checks the type of file to determine the appropriate
 *   form of close to do.
 */

extern void fio_clear_eof( integer file );			/*  Clear file EOF condition */
extern logical fio_test_eof( integer file );		/*  Test if file is at EOF */
extern logical fio_eof( integer fid, integer mode );
/*
 *  Test for and clear the EOF status on a file. The fio_eof tests EOF
 *  and also, if mode is FIO_CLEAR (=0), clears it.
 */

extern logical fio_any_eof( integer mode );
/*
 *   Test if there is any EOF condition set on any file. If mode is FIO_CLEAR (=0),
 *   clear all such conditions.
 */

extern void fio_set_status ( integer fid, logical ios ) ;
/*
 *  Set the status for a given file to a specific value; normally done internally
 */

extern integer fio_get_cfd( integer fid );
/*
 *   Return the C <fcntl.h> file descriptor for the file.
 */

extern integer fio_file_n( void );				/*  Number of files */
/*
 * Return number of allocated FIO files
 */

extern void fio_list( void );							/* List all files */
/*
 *  Display to message stream a list of all open files and their
 *  characteristics.
 *  Implements TPIO command \?FIO
 */



/*
 * \section{Byte stream (C stdio) file I/O }
 */

extern integer sio_open_r( char* filename );
extern integer sio_open_w( char* filename );
extern integer sio_open_rw( char* filename );
/*
 *  Open a stream I/O file for read, write or read/write.
 *  Return a JCMLIB file id (0 for failure).
 */

extern logical sio_seek( integer fid, integer pos );
/*
 *   Move to the specified byte in an SIO file.
 */

extern logical sio_write_q( integer fid, logical* buf, integer n, integer* nw );
extern logical sio_write_c( integer fid, char* buf, integer n, integer* nw );
extern logical sio_write_d( integer fid, double* buf, integer n, integer* nw );
extern logical sio_write_r( integer fid, real* buf, integer n, integer* nw );
extern logical sio_write_i( integer fid, integer* buf, integer n, integer* nw );
extern logical sio_write_s( integer fid, short* buf, integer n, integer* nw );
/*
 *  Write n items from the buffer to the SIO file; return true on success
 *  and nw as the number of items written.
 */

extern logical sio_read_c( integer fid, char* buf, integer n, integer* nw );
extern logical sio_read_d( integer fid, double* buf, integer n, integer* nw );
extern logical sio_read_r( integer fid, real* buf, integer n, integer* nw );
extern logical sio_read_q( integer fid, logical* buf, integer n, integer* nw );
extern logical sio_read_i( integer fid, integer* buf, integer n, integer* nw );
extern logical sio_read_s( integer fid, short* buf, integer n, integer* nw );
/*
 *  Read n items from the SIO file into the buffer; nw is the number actually read.
 */

extern void sio_flush( integer fid );
/*
 *   Flush a STDIO file.
 */

extern void sio_flush_stdout( void );
/*
 *  Flush stdout.
 */

/*
 *   \section{C file I/O and raw I/O}
 */

extern integer cio_open_r( char* filename );
extern integer cio_open_w( char* filename );
extern integer cio_open_rw( char* filename );
/*
 *   Open a file using C <fcntl.h> I/O, for read, write, or read-write.
 *   Return a JCMLIB file ID.
 */


extern void cio_close( integer fid );
/*
 *   Close a JCMLIB file of type CIO.  Recommend you use fio_close instead.
 *   For a raw mode file, must do clear_raw first; fio_close does this for you.
 */

extern logical cio_write( integer fid, char* buf, integer n, integer* m );
/*
 *   Write n bytes to a CIO file. Returns true if OK, and m is the number
 *   of bytes actually written.
 */

extern logical cio_read( integer fid, char* buf, integer n, integer* m );
/*
 *  Read n bytes from a CIO file. Returns true if OK, and m is the number of
 *  bytes actually read.
 */

extern logical cio_tty( integer fid );
/*
 *  Returns the result of POSIX  isatty() on the file.
 */

extern integer cio_read_byte( integer fid );
/*
 *  Read and return a byte from the file.
 */

extern integer cio_open_raw( void );
/*
 *   Open /dev/tty in raw mode.
 *   Return a JCMLIB file ID.
 *   Close the file with fio_close( fid ).
 */

extern integer cio_read_raw( integer fid );
/*
 *  Read and return a byte from a raw mode file.
 */


/*
 *  \section{FIO routines for ASCII sequential record files (text files)}
 */

extern void process_file( char* filename, void (*process_line)( char* ) );
extern void process_file1( char* filename, void (*process_line)( void*, char* ), void* ptr );
/*
 *   Routines for processing a text file. The user must supply a function f,
 *       void f( char* line );
 *   which will handle a single line of input. The process_file routine
 *       process_file( filename, f );
 *   opens the file, and passes valid lines of text to the function.
 *   However, blank lines and lines of text whose first character is the comment character are
 *   ignored. Also, tabs are converted to spaces.
 *   On end of file, the file is closed.
 *
 *   The second routine, process_file1, lets you pass extra data
 *   down to the line processing routine.
 *       process_file1( filename, f1, extra_info );
 *       void f1( void* extra_info, char* line )
 *   For example, you might be reading a file whose lines themselves contain 
 *   filenames. Then the extra_info might be a directory for the files:
 *        process_file1( filename, f1, data_dir );
 *   with
 *        void f1( void* extra, char* line ) {
 *         t_filename path;
 *         char* dir = (char*) extra;
 *         sys_filepath( dir, line, " ", path, FSIZE );
 *         do_some_more( path );
 *        }
 * That lets you do_some_more on every file listed in filename.
 */


extern integer fio_open_r( const char* filename );
/*
 *  Open an ASCII sequential record file (ASR file) for read. Return a JCMLIB file ID (fid)
 *  (0 on failure).
 */

extern integer fio_open_w( const char* filename );
/*
 *  Open an ASCII sequential file for write, return a file ID (0 on failure).
 *  If the file already exists, it will be destroyed and overwritten.
 */

extern integer fio_open_a( const char* filename );
/*
 *  Open an ASCII sequential file for write, return a file ID (0 on failure).
 *  This version opens to append, i.e. the existing file is not overwritten
 *  but is appended to.
 */

extern integer fio_open_d( const char* filename, integer rec_bytes );
/*
 *   Open a filename with fixed record length, return a file ID.
 *   Record length is in bytes.
 */


extern logical fio_read( integer fid, char* line, integer lmax );
extern logical tty_read( char* buf, integer siz );
/*
 *  Read a line from FIO/ASR file; return FALSE on error or end of file.
 *  The tty_read version reads from stdin.
 */

extern logical fio_write( integer fid, const char* buf );
extern logical tty_write( const char* buf );
/*
 *  Write a line to an FIO/ASR file; return FALSE on error.
 *  The tty_write version writes to stdout.
 */

extern logical fio_write_prompt( integer fid, char* buf );
extern logical tty_writel( char* buf );
/*
 *  fio_write_prompt is like fio_write but does not add a newline, so further text
 *  may be written to the same line.
 *  tty_writel is the same but writes to stdout.
 */

extern logical fio_write_eol( integer fid );
/*
 *  Write a newline to the file.
 */

extern void fio_rewind( integer fid );
/*
 *   Rewind an FIO/ASR file.
 */

extern integer fio_get_lineno( integer fid );
/*
 *  Get the current line number of the file
 */

extern integer fio_get_pos( integer fid );
/*
 *  Return byte position in current line.
 */

extern void fio_set_retry( logical q );
/*
 *  Enable or disable the RETRY feature which, on failing to open
 *  a file, prompts the user for an alternate filename.
 */

extern logical fio_get_retry( void );
/*
 *  Return the state of the RETRY feature.
 */

extern logical fio_interactive( integer fid );
/*
 *  True if fid is zero or equal to TTY_OUT or TTY_IN.
 */

extern integer fio_tty_in( void );
extern integer fio_tty_out( void );
extern integer fio_tty_err( void );
/*
 *  Return the FID values being used for TTY_IN, TTY_OUT, and TTY_ERR
 *  which are the JCMLIB equivalents for stdin, stdout and stderr.
 */

extern integer fio_get_recl( integer fid );
/*
 *   Return record length of file
 */

extern void fio_set_recl( integer fid, integer recl );
/*
 *   Reset record length of file (not recommended).
 */

extern void pager_set( logical q );
/*
 *  Enable or disable the pager, which buffers tty_write calls
 *  to make the pause after each 'page' in a way similar to Unix 'more'.
 */

extern void pager_set_length( integer page_length );
/*
 *  Set the length of the pager 'page', default 40 lines.
 */

extern void pager_print_state( void );
/*
 *  Print the current status of the pager.
 *  TPIO command \?PAGER
 */


extern void fio_path_init( void );
/*
 *  Reinitialize the JCMPATH path by reading the JCMPATH environment
 *  variable. This variable may contain a space or comma-separated list of
 *  directories which will be used for searches by the FIO open routines.
 *  Environment variables preceded by '$' will be expanded.
 *  The routine is called by jcmlib_init() and is not normally needed.
 */

extern void fio_path_set( char* pathlist );
/*
 *  Overwrite the current set of paths with a new set parsed from the
 *  space or comma-separated list of directories in the argument.
 *  Not normally required.
 */

extern void fio_path_list( void );
/*
 *   Display the current path. TPIO command \?PATH.
 */

/*
 *  \section{Standard I/O message streams}
 *
 *  The IO basic I/O system is layered on top of FIO.
 *  It provides functionality similar to the Unix stdin, stdout and stderr
 *  streams, here called TTY_IN, TTY_OUT and TTY_ERR.
 *  The IO layer has in addition MSG_IN and MSG_OUT.
 *  The connection of these streams to individual files may
 *  be controlled within the program. At startup, all three are linked to
 *  their Unix equivalents, so MSG_IN = TTY_IN and MSG_OUT = TTY_OUT.
 *  The io_out( fid ) routine may be used to reallocte MSG_OUT to a file.
 */

extern void io_init( void );			
/*
 *  Initialize the IO layer; done by jcmlib_init().
 */
extern void io_in( integer file_no );				/* Set IO input */
/*
 *   Set input to come from the given file.
 */
extern void io_out( integer file_no );				/* Set IO output */
/*
 *   Send output to the given file.
 */
extern void io_err( integer file_no );
/*
 *  Send error messages to the given file.
 */

extern logical msg( char* buf );					/* Write text to output */
extern logical tmsg( char* buf );					/* Write text to stdout */
/*
 * Issue message to message out stream (msg) or to TTY OUT stream (tmsg).
 */

extern logical msgl( char* buf );					/* Write text to output, prompt */
extern logical tmsgl( char* buf );					/* Write text to stdout, prompt */
/*
 *  msgl is like msg but doesn't add a newline. tmsgl is like tmsg
 *  but doesn't add a newline.
 */

extern logical msg_endline( void );
/*
 *  Issue a newline to the message out stream.
 */
extern void dashline( integer n );
/*
 *  Issue a line composed of n dashes to the message out stream;
 *  a convienent way of visually breaking up the output.
 */


extern void jcm_err_msg( const char* buf );
extern void jcm_err_msg1( const char* buf, const char* buf2 );
/*
 *   Issue error message to error stream. Second form concatenates
 *   two strings and writes them out.
 */

extern logical io_input( char* buf, integer siz );			/* Read text from input, directly */
extern logical tty_input( char* buf, integer siz );			/* Read text from stdin */
/*
 *   Read one line of message from message in stream (io input) or
 *  the TTY IN stream (tty_input).
 */
extern void io_rewind( void );
/*
 *   Rewind the message in stream
 */
extern logical io_int( integer );					/* Test IO input is interactive */
/*
 *  Test if the message in stream is interactive (or equal to TTY IN)
 */

extern integer io_chan( integer mode );
/*
 *   Return the current FID for MSG IN (mode=0) or MSG OUT (mode=1).
 */

extern logical input_test_eof( void );				/* Test IO input is at EOF */
/*
 *  Test if MSG IN is at EOF.
 */


extern void io_print_state( void );
/*
 *   Print state of basic IO streams.
 */

extern void io_opush( integer fid );
extern void io_ipush( integer fid );
/*
 *   Push a file ID onto the input (ipush) or output (opush) file stacks.
 *  This has the effect of making the given file MSG IN or MSG OUT,
 *  but unlike with io_in and io_out, the previous incumbent is remembered
 *  and may be recovered with io_ipop or io_opop.
 */

extern void io_ipop( integer mode );
extern void io_opop( integer mode );
/*
 *  Pop the input or output file stacks.
 *  The argument can be IOFS_POP_CLOSE, which closes the file it pops,
 *  or IOFS_POP_ONLY, which moves the stack pointer but keeps the
 *  file open. Usually you want the first of these.
 */

extern void iofs_print_state( void );
/*
 *   Display the state of the IO file stack.
 */

extern logical iofs_eof( integer mode );
extern logical iofs_ok( integer mode );
extern integer iofs_status( integer mode );
/*
 *  Check for EOF, status OK and status values on the
 *  MSG IN steam.
 */

extern void iofs_detab( logical q );
/*
 *  Enable or disable the removal of tabs from input in the iofs_read routine.
 *  Default is enabled.
 */

extern logical iofs_read( char* buf, char* prompt, integer siz );
/*
 *   Wrapper for io_input, handling continuation lines, removing tabs,
 *   and supporting system escapes.
 *   This routine is further wrapped by the TPIO routines.
 */


extern integer iofs_get_level( void );
/*
 *   Find out the IOFS stack level.
 */

extern integer iofs_istack( integer level );
/*
 *  Find out the stream at a given level of the IOFS stack.
 */


/*
 *  \section{TPIO: a command line interface}
 *
 * TPIO lets your programs prompt the user for input while letting
 * them enter multiple commands on a line, and allows you to use the
 * same program to read from files without prompting. It buffers input
 * and returns tokens as the calling program requests them.
 * It also supports a variety of special escape commands which may be
 * entered at the user prompt (or in a script). Escape commands are
 * processed and then another attempt is made to get a token, with the
 * same prompt. For instance, the calling program may contain:
 *
 *    tpio_cs( "Enter object name", name, FSIZE );
 *    n = tpio_d_i( "Enter number of iterations", 10 );
 *
 * In interactive mode, the user will see the prompt
 *  Enter object name:
 * They may then type
 *  Enter object name:  3C 273
 * and will get the prompt
 *  Enter number of iterations [10]:
 * They can then type a number of iterations to use, or just hit return
 * and get the default value of 10 suggested. They might have instead entered
 *  Enter object name:  3C 273; 22
 * if they already knew what the next prompt was going to be. The semicolon
 * indicates the end of the previous string; tpio_cs returns text up to the
 * semicolon and saves the 22 for the next call, so the user is not prompted
 * for it, tpio_d_i just returns 22 silently.
 * Maybe the user wanted to issue a couple of escape commands, to set the
 * debug level and send output to a file; they could do:
 *  Enter object name: \[*3]
 *  [*3] Debug level set to 3
 *  Enter object name: 3C 273
 *  Enter number of iterations [10]: \>save.log
 *  Enter number of iterations [10]: 22
 *
 * What happened here is that TPIO spotted the escape commands (which
 * begin with backslash), executed them, and then issued the prompt again
 * without the calling program having to worry. Any program using the TPIO
 * interface has all the TPIO escape commands available at each prompt
 * or token-read.
 *
 * The supported escape commands are:
 * \>filename       Create file filename and make it the MSG OUT message stream.
 * \<filename       Open file filename and make it the MSG IN message stream.
 * \>*              Restore MSG OUT to be stdout
 * \<*              Restore MSG IN to be stdin
 * \[*n]            Set the debug level to n (for integer n)
 * \VER             Display the version of JCMLIB
 * \SET name value  Reassign an escape character: e.g.
 * \SET WORD !      changes the word separator from comma to exclamation mark
 * \SET ESC :       changes the escape code from backslash to colon (so to
 *                   change it back you'd have to do :SET ESC \)
 * \?PATH           Display the JCMLIB FIO directory search path
 * \?ESC            Display the settings of all escape characters
 * \?FIO            Display the list of open files
 * \?IOFS           Display the file stack
 * \?TPIO           Display the TPIO buffers
 * \?BIO            Display the MSG IN and OUT assignments
 * There are also commands \, \DO, \END, \EXIT, \ECHO which 
 * are experimental and not yet documented.
 */


extern void tpio_init( void );
/* 
 *   Initialization routine for the TPIO subsystem, called by jcmlib_init().
 *   You won't need to call it.
 */

extern void tpio_c( char* prompt, char* token, integer siz );
extern void tpio_cs( char* prompt, char* token, integer siz );
extern void tpio_cl( char* prompt, char* token, integer siz );
/*
 *  Read a token from the MSG IN stream. If the MSG IN current line is
 *  empty, read a new line. In this case, if MSG IN is TTY IN (interactive),
 *  prompt the user with the text in 'prompt' before reading the new line.
 *
 *  tpio_c reads a token up to the first space, WORD SEP (comma), or 
 *  STRING SEP (semicolon) or null; although if the text includes parentheses
 *  LPAR and RPAR, separators within those parentheses are not counted.
 *  Thus repeated calls to tpio_c on
 *  alpha beta;gamma del(inserted text)ta,,epsilon
 *  will return the six tokens
 *  {alpha}, {beta}, {gamma}, {del(inserted text)ta}, {}, {epsilon}
 *  (the curly braces are for illustration only).
 *  The tpio_cs function uses only STRING SEP and null, so it would return
 *  the two tokens
 *  {alpha beta}, {gamma del(inserted text)ta,,epsilon}
 *  on the same data.
 *  tpio_cl returns the remainder of the line. 
 */

extern integer tpio_i( char* prompt );
extern short tpio_s( char* prompt );
extern real tpio_r( char* prompt );
extern double tpio_d( char* prompt );
/*
 *   These routines perform a tpio_c and then read a numeric value of
 *   the appropriate type. Typical usage:
 *       z = tpio_d( "Enter redshift" );
 *
 */


extern integer tpio_d_i( char* prompt, integer xdefault );
extern double tpio_d_d( char* prompt, double xdefault );
extern short tpio_d_s( char* prompt, short xdefault );
extern real tpio_d_r( char* prompt, real xdefault );
/*
 *  Like tpio_i, etc, but if the returned string token is blank,
 *  sets the value to be xdefault. In interactive use the value of
 *  xdefault is wrapped in square parens and appended to the prompt.
 */

extern void tpio_in( char* filename );					/* Set input file */
/*
 *  Push a file onto the TPIO input stack. Opens the file and pushes
 *  it on the stack with io_ipush. The special filename '*' indicates
 *  the interactive stream.
 */

extern void tpio_cc( char* prompt, char* buf, integer siz );
/*
 *   Special command mode for TPIO. Like tpio_c, but 
 *   enables some reserved escape commands. To be documented later.
 */

extern void tpio_dbg( void );					
extern void tpio_print_state( void );
extern void tpio_disp( integer i );
/* 
 *   Print TPIO internals; various specialized debug information.
 */


extern void tpio_setbuf( char** arr, integer n );
/*
 *  Adds some tokens to the tpio buffer directly.
 */

extern void tpio_cmdargs( char** args, integer nargs, logical* interact, logical* opts );
/*
 *  Parses command line arguments provided as input. Returns the interact and opts
 *  flags, and adds remaining arguments to the tpio buffer.
 *  This routine is used in the precess family of utilities, which
 *  have the command line syntax:
 *    command_name initial_commands : infile : outfile
 *
 */

/*
 * \section{The JCMLIB escape characters}
 *
 * These routines will not normally be called directly;
 * they are accessed through the TPIO interface and
 * used for string parsing etc.
 * The special characters are:
 *  Name      Default value       Use
 *   ESC       \                  Escape commands in TPIO
 *   LPAR      (                  Paren in token routines                  
 *   RPAR      )                  Paren in token routines                  
 *   WORD      ,                  Word separator in token routines
 *   STRING    ;                  String separator in token routines
 *   SYS       $                  Shell escape command (at beginning of line) 
 *                                in TPIO
 *   CONT      $                  Continuation line (at end of line) in TPIO
 *   CMT       #                  Comment character in TPIO
 *   MACRO     $                  Macro character used in TPIO loops
 *   ARG       blank              Used in tpio_cmdargs to separate command
 *                                line arguments. Note that some JCM programs
 *                                (eg precess) reset this character to ':'.
 */


extern void esc_init( void );
/*
 *   Reinitialize all escape characters to their default values.
 *   Called by jcmlib_init().
 */

extern void esc_status( void );		
/*
 *  Display values of escape characters
 *  TPIO command: \?ESC
 */

extern char esc_get( const char* name );
/*
 *  Get current value of character given symbolic name
 */

extern void esc_reset( const char* name, const char val );
/*
 * Explicitly set value of character.
 * TPIO command: \SET name val  
 */


/*
 * \section{Debug routines}
 */

extern void dbg_set( const integer level );
/*
 *  Set the current global debug level
 */

extern logical dbg_level( const integer i );	
/*
 *   True if the current global debug level is 
 *    >= the argument.
 */

extern integer dbg_get( void );
/*
 *  Return the current global debug level
 */

extern void dmsg( const integer level, const char* buf );
extern void dmsg1( const integer level, const char* buf, const char* buf2 );
/*
 *  Issue message to error stream if debug level is >= level
 *  dmsg1 is a convenience routine to concatenate two messages
 */

extern void dmsg_tag( const integer level, char* buf );
/*
 *   Writes a string [n] containing the current debug level. 
 */

/*
 * \section{Error handling}
 *
 * The error codes have numbers starting at JCM_ERR_START;
 * they are listed in jcmerr.h and the descriptive strings are in
 * errdata.h
 */

extern void err_set( const integer code );
/*
 *  Set the current error status
 */

extern integer err_get( void );
/*
 *   Get the current error status
 */

void err_string( const integer code, char* buf, const integer maxlen );
/*
 *   Convert a JCMLIB error code to a descriptive string.
 */

extern void err_clear( void );
/*
 *   Clear error status
 */

extern logical err_test( const  integer mode );
/*
 *  Test error status. If argument is FIO_CHECK, just look;
 *  if argument is FIO_CLEAR, then set error status to zero.
 */

/*
 *  \section{Units}
 */

extern void unit_init( void );
/*
 *  Initialize the unit library
 */

extern void unit_def_read( const char* filename );
/*
 *   Read a file of composite unit definitions
 */

extern void unit_texmode( const logical q );
/*
 *   Enable or disable the composition of unit exponents in TeX-compatible
 *   format (default enabled)
 */

extern void unit_mult( const char* u1, const char* u2, char* u, const integer siz );
/*
 *  Multiply two units, returning the result
 */

extern void unit_div( const char* u1, const char* u2, char* u, const integer siz );
/*
 *   Divide two units, returning the result
 */

extern void unit_exp( const char* unit1, const double power, char* unit2, const integer siz );
/*
 *   Raise a unit to a power
 */

extern void unit_resolve( const char* u1, char* u2, const integer siz );
/*
 *    Resolve a unit by expanding composite unit definitions
 */

extern void unit_dparse( const char* u1, char* u2, const integer siz );
/*
 *   Debugging routine.
 *   Parse a unit u1, display its parsed structure, and compose it
 *   again as u2.
 */

extern void si_write_r( real x, char* cy );
extern void si_write_d( double x, char* cy );
/*
 *   Write a number as a normalized value followed by an SI prefix.
 *   The result may be concatenated later to a unit name.
 */

extern real si_read_r( char* cy );
extern double si_read_d( char* cy );
/*
 *    Read a string consisting of a number and an SI prefix.
 *    Convert the SI prefix to rescale the number and return the value.
 */


extern integer siexp( char* pfx, logical* ok );
/*
 *  Low level routine to convert an SI prefix to an exponent.
 *  siexp( "M", &ok ) = 6;  siexp( "n", &ok ) = -9; with ok = TRUE in both cases.
 *  siexp( "foo", &ok ) = 0, with ok = FALSE.
 */

extern void si_getpfx( integer exponent, char* pfx, integer* remainder );
/*
 *   Low level routine to convert an exponent to an SI prefix.
 *  Returns prefix and remainder.
 */



/*
 * \section{The JCMLIB string parser}
 *
 * The parser splits text up into tokens of different types.
 * The types are of several levels.
 */
extern void char_name( char c, char* name );
/*
 *  Return the name of a given ASCII character.
 */

extern void parser( char* buf, integer* itemcode, integer* itempos, integer* nitems, char* cmode );
/*
 *  Parse a buffer, generating the parse codes, using the cmode argument
 *  to determine the method of analysis.
 */

extern void parse_disp( char* buf, integer* itemcode, integer* itempos, integer nitems );
extern void parse_tdisp( char* buf, integer* itemcode, integer* itempos, integer nitems );
/*
 *  Given a buffer and its parse codes, display the analysis of the buffer.
 */

extern void parse_read( integer i, char* buf, integer* itempos, char* token, integer* siz );
/*
 *  Extract the i'th token from buf.
 */


/*
 *  \section{Fortran string routines}
 */
void cs_copy_cf( char* cname, char* fname,  long fsiz );
void cs_copy_fc( char* fname, char* cname, long fsiz, long maxlen );

/* Copy a C null terminated string to a Fortran blank terminated string.
 * maxlen is the length of the c string, fsiz the length of the Fortran one.
 */


char* fc_AllocString( char* fstring, long maxlen );
/*
 *  Copy a Fortran string to a freshly allocated C string.
 *  The string may be blank or null.
 */
void fc_FreeString( char* cstring, char* fstring, long fsiz );
/*
 *  Free a string created by fc_AllocString, copying it first to
 *  a Fortran string if fstring is not the null pointer.
 */

char** fc_AllocStringArray( char* array, long fsiz, long n );
void fc_FreeStringArray( char** fc_StringArray, char* value, long fsiz, long n );
/*
 *  Like fc_AllocString/FreeString, but for an array of equal length strings.
 *  Note that the Fortran array is a C char*, it is just a long array of char
 *  notionally divided up into segments of length fsiz.
 */

void fc_StringCopyItem( char* cstring, char* fstring, long maxlen, long i );
void fc_StringPutItem( char* fstring, long maxlen, long i, char* cstring );
/*
 *
 */

/*
 *  \section{Miscellaneous and obsolete routines}
 */

extern integer prio_select( integer* uprio, logical* there, integer n );
extern void gquot( integer val, integer* ival, integer nval, integer* k, integer* irem );

/* Obscure routines used in other JCM code */


extern void transfer_i( void*, integer*, const integer );
extern void transfer_r( void*, float*, const integer );
extern void transfer_d( void*, double*, const integer );
extern void transfer_s( void*, short*, const integer );
extern void transfer_c( void*, char*, const integer );
/*
 *    The transfer routine copies from an unknown type array
 *    to a specified type array. These routines are not really
 *    needed in C and are provided for compatibility with my
 *    Fortran library.
 */

extern integer matchr_d( const double value, const double* list, const integer n1, const integer n2 );
extern integer matchr_s( const short value, const short* list, const integer n1, const integer n2 );
extern integer matchr_i( const integer value, const integer* list, const integer n1, const integer n2 );
extern integer matchr_r( const real value, const real* list, const integer n1, const integer n2 );
extern integer matchr_c( const char* buf, char** array, const integer n1, const integer n2 );

/*
 *   Like match, but only test a subarray from n1 to n2.
 */

extern integer cs_match( const char* buf, const char* array, const integer n, const integer item_size );			/* Match string in array of strings given array element size */
/*
 *  Like match_c, but instead of working on a char** array, works on a char*
 * array in which strings are stored in fixed areas of size item_size separated
 * by nulls. In other words, {array} is an array of n strings which may
 * be referenced as 
 *   array[0],  array[item_size+1], array[2*(item_size+1)], etc..
 * I was using arrays like this in some programs, but have now settled
 * on using char** arrays allocated by alloc_array_c for most purposes.
 */

extern char* vec_print_d( double* p, char* fmt, char* buf );
/*
 *   Write a 3-vector to a string. Superseded by vwritef_d.
 */

extern void eflag_parse( char** args, integer nargs, logical local_flag_set( char*, char*, logical ) );
extern void eflag_list( char** efln, logical* eflags, char** evals, integer neflags, char* cmt, integer tab );
extern logical eflag_set( char* opt, char* oval, char** efln,  char** evals, integer neflags, logical* eflags, logical eswitch );
extern logical eflag_syntax( char* opt, char* word, char* value );

/*
 * These routines handle internal flags in other JCMLIB programs.
 */

/*
 *   The following are a complicated debug support system which
 *   I don't recommend.
 */
extern void dbg_push( const char* package, const char* routine );
extern void dbg_pop( void );
/*
 *   Push and pop a package and routine name onto the trace stack
 *   This and the following routines may be used as a crude 
 *   debugger.
 */

extern void dbg_trace( const logical q );
/*
 *  Enable debug tracing; used in combination with the dbg_push and
 *  dbg_pop routines.
 */

extern void dbg_listpkg( void );
/*
 *   List packages on the debug stack
 */

extern integer dbg_addpkg( const char* pkg );
extern void dbg_setpkg( const char* pkg, const integer level );
/*
 *   Set package debug at a given level
 */

extern void dbg_disptree( void );
extern void dbg_disppkg( void );
/*
 *   Display the trace tree
 */
extern void dbg_fmttree( char* buf, const integer siz );

/*
 *   The rest are miscellaneous obsolete routines which I am phasing out.
 */
extern void sbf_put_r( real* data, integer n );
extern void sbf_put_i( integer* data, integer n );
extern void sbf_put_s( short* data, integer n );
extern void sbf_put_d( double* data, integer n );

extern integer bit_unpack_i_slow( const integer value, const integer start_bit, const integer n_bits );
extern integer bit_unpack_s_slow( const short value, const integer start_bit, const integer n_bits );

/* fs */

extern integer fs_ends( char* buf, integer size );					/* Length of strng to last non blank, nulls allowed */
extern void fs_zero( char* line, integer size );					/* Make string blank, no trailing null appended */

extern void jcm_cmdlineargs( integer argc, char **argv, card *args, integer* nargs, integer asiz );

extern void vprint( char* buf );
/* Undocumented debugging routine for TPIO */

/* Obsolete error handling routines */

extern void err_save( const char* code, const char* text );
extern void err_disp( const integer mode );
extern void err_save_io( const char* code, const char* text, const integer status );



#endif
