#include <math.h>
#include "jcmlib.h"

static const double polar_lat = 90.0;
static const real rpolar_lat = 90.0;
static double polar_ang_d( const double*, const double );

double dacosd( const double x )
{
 return( radout_d( acos( x ) ) );
}

static double polar_ang_d( const double* p, const double r )
{
 double x;
 double re;
/* Make sure you get an accurate off axis angle. If the off axis
   angle is small, safer to take the size of the normal component.
 */
 x = acos( p[2] / r );
 if ( fabs( x ) < 0.001 ) {
  re = sqrt( p[0] * p[0] + p[1] * p[1] );
  x  = asin( re / r );
 }
 return( radout_d( x ) );
}

double dasind( const double x )
{
 return( radout_d( asin( x ) ) );
}

real racosd( const real x )
{
 return( radout_r( (real)acos( (double)x ) ) );
}

real rasind( const real x )
{
 return( radout_r( (real) asin( (double)x ) ) );
}


double dcosd( const double x )
{
 return( cos( radin_d( x ) ));
}

double dsind( const double x )
{
 return( sin( radin_d( x ) ));
}

real rcosd( const real x )
{
 return( (real)cos( (double)radin_r( x ) ));
}

real rsind( const real x )
{
 return( (real)sin( (double)radin_r( x ) ) ) ;
}

void vec_d( const double r, const double theta, const double phi, double* p )
{
 if ( theta == 0.0 || theta == 180.0 ) {
  p[0] = 0.0;
  p[1] = 0.0;
 } else {
  p[0] = r * dcosd( phi ) * dsind( theta );
  p[1] = r * dsind( phi ) * dsind( theta );
 }
 p[2] = r * dcosd( theta );
}

void vec_r( const real r, const real theta, const real phi, real* p )
{
 if ( theta == 0.0 || theta == 180.0 ) {
  p[0] = 0.0;
  p[1] = 0.0;
 } else {
  p[0] = r * rcosd( phi ) * rsind( theta );
  p[1] = r * rsind( phi ) * rsind( theta );
 }
 p[2] = r * rcosd( theta );
}

void polar_d( const double* p, double* r, double* theta, double* phi )
{
 *r = vnorm_d( p );
 if ( *r > 0 ) {
  *phi = argd_d( p[0], p[1] );
  *theta = polar_ang_d( p, *r );
 } else { 
  *phi =0.0;
  *theta = 0.0;
 }
}

void polar_r( const real* p, real* r, real* theta, real* phi )
{
 *r = vnorm_r( p );
 if ( *r > 0.0 ) {
  *phi = argd_r( p[0], p[1] );
  *theta = racosd( p[2]/(*r) );
 } else { 
  *phi =0.0;
  *theta = 0.0;
 }
}


double polarll_d( const double* p, double* llong, double* lat )
{
 double r;
 r = vnorm_d( p );
 if ( r > 0.0 ) {
  *llong = argd_d( p[0], p[1] );
  *lat  = polar_lat - polar_ang_d( p, r );
 } else {
  *llong = 0.0;
  *lat = polar_lat;
 }
 return( r );
}

real polarll_r( const real* p, real* llong, real* lat )
{
 real r;
 r = vnorm_r( p );
 if ( r > 0.0 ) {
  *llong = argd_r( p[0], p[1] );
  *lat  = rpolar_lat - racosd( p[2]/r );
 } else {
  *llong = 0.0;
  *lat = rpolar_lat;
 }
 return( r );
}


void unitll_d( const double llong, const double lat, double* p )
{
 if ( fabs( lat ) == polar_lat ) {
  p[0] = 0.0;
  p[1] = 0.0;
 } else {
  p[0] = dcosd( llong ) * dcosd( lat );
  p[1] = dsind( llong ) * dcosd( lat );
 }
 p[2] = dsind( lat );
}

void unitll_r( const real llong, const real lat, real* p )
{
 if ( rabs( lat ) == rpolar_lat ) {
  p[0] = 0.0;
  p[1] = 0.0;
 } else {
  p[0] = rcosd( llong ) * rcosd( lat );
  p[1] = rsind( llong ) * rcosd( lat );
 }
 p[2] = rsind( lat );
}



void polar6_r( const real* p, real* r, real* theta, real* phi, real* rrdot, real* tdot, real* phidot )
{
 double dr, dtheta, dphi, drdot, dtdot, dphidot;
 double dp[6];
 ar_copy_rd( p, dp, 6 );
 polar6_d( dp, &dr, &dtheta, &dphi, &drdot, &dtdot, &dphidot );
 *r = (real)dr;
 *theta = (real)dtheta;
 *phi = (real)dphi;
 *rrdot = (real) drdot;
 *tdot = (real)  dtdot;
 *phidot= (real) dphidot;
}

void polar6_d( const double* p, double* r, double* theta, double* phi, double* rrdot, double* tdot, double* phidot )
{
 double e[3];
 double de[3];
 double enorm;
 double ec[3];
 double r0;

 polar_d( p, &r0, theta, phi );
 *r = r0;
 if ( r0 > 0.0 ) {
  vcopy_d( p, e );
  vcopy_d( p+3, de );
  *rrdot = vdot_d( e, de ) / r0;
  e[2] = 0.0;
  de[2] = 0.0;
  enorm = vnorm_d( e );
  vcross_d( e, de, ec );
  if ( enorm > 0.0 ) {
   *tdot = radout_d( ( p[2] * vdot_d( e, de ) - p[5] * ( enorm * enorm ) ) / ( enorm * r0 * r0 ) );
   *phidot = radout_d(  ec[2] / ( enorm * enorm ) );
  } else {
   *tdot = radout_d( vnorm_d( de ) / r0 );
   if ( *theta > 90.0 ) *tdot *= -1.0;
   *phidot = 0;
  }
 } else {
  *rrdot = 0;
  *tdot = 0;
  *phidot = 0;
 }
}

void vec6_r( const real r, const real theta, const real phi, const real rrdot, const real tdot, const real phidot, real* p )
{
 double dp[6];
 vec6_d( (double)r, (double)theta, (double)phi, (double)rrdot, (double)tdot, (double)phidot, dp );
 ar_copy_dr( dp, p, 6 );
}


void vec6_d( const double r, const double theta, const double phi, const double rrdot, const double tdot, const double phidot, double* p )
{
 double tdotr, phidotr;
 double ca, sa, cb, sb;
 double w;

 vec_d( r, theta, phi, p );
 tdotr = radin_d( tdot );
 phidotr = radin_d( phidot );
 ca = dcosd( phi );
 sa = dsind( phi );
 cb = dcosd( theta );
 sb = dsind( theta );
 w = r * tdotr * cb + rrdot * sb;
 p[3] = -p[1] * phidotr      + w * ca;
 p[4] =  p[0] * phidotr      + w * sa;
 p[5] = -r    * tdotr   * sb + rrdot * cb;
}

real spharc_r( const real a1, const real b1, const real a2, const real b2 )
{
 return( (real)spharc_d( (double)a1, (double)b1, (double)a2, (double)b2 ) );
}

double spharc_d( const double a1, const double b1, const double a2, const double b2 )
{
 double v1[3];
 double v2[3];
 double w;
 double dw;
 double x, y;
 integer i;

 unitll_d( a1, b1, v1 );
 unitll_d( a2, b2, v2 );
 w = 0.0;
 for ( i = 0; i < 3; i++ ) {
  dw = ( v1[i] - v2[i] ) / 2.0;
  w += dw * dw;
 }
 x = w < 1.0 ? sqrt( 1.0 - w ) : 0;
 y = sqrt( w );
 return( 2.0 * argd_d( x, y ) );
}
