#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "jcmlib.h"

static void utl_str_rfclean( char* tbuf );

static void numberclean( char* tbuf, char* buf, integer siz );
static char* nanstring = "  -  ";
static const double efmt_max = 1.0E5;
static const double efmt_min = 1.0E-3;
static char* efmt = "%12.5E";
static char* ffmt = "%12.5f";

/*----------------------------------------------------------------------*/
/*
 *  read_d
 *
 * Read a float from a string
 * Uses atof from <stdlib.h>
 */

double read_d( const char* buf )
{
 char* err;
 double x;
 if ( cs_is_blank( buf ) ) {
  return( 0.0 );
 } else if ( nancode( buf ) ) {
  return( nan_d() );
 } else {
  x = strtod( buf, &err );
  if ( !cs_is_blank( err ) ) err_set( JCM_ERR_READ_STRING );
  return ( x );
 }
}

real read_r( const char* buf )
{
 double x;
 char* err;
 if ( cs_is_blank( buf ) ) {
  return( 0.0 );
 } else if ( nancode( buf ) ) {
  return( nan_r() );
 } else {
  x = strtod( buf, &err );
  if ( !cs_is_blank( err ) ) err_set( JCM_ERR_READ_STRING );
  return ( (float)x );
 }
}

/*----------------------------------------------------------------------*/
/*
 *  read_i
 *
 * Read an integer from a string
 * Uses atoi from <stdlib.h>
 *
 */

integer read_i( const char* buf )
{
 integer x;
 char* err; 
 if ( cs_is_blank( buf ) ) {
  return( 0 );
 } else if ( nancode( buf ) ) {
  return( nan_i() );
 } else {
  x = strtol( buf, &err, (int)10 );
  if ( err != NULL ) {
   if ( !cs_is_blank( err ) ) { err_set( JCM_ERR_READ_STRING ); }
  }
  return ( x  );
 }
}

short read_s( const char* buf )
{
 integer x;
 char* err;
 if ( cs_is_blank( buf ) ) {
  return( 0 );
 } else if ( nancode( buf ) ) {
  return( nan_s() );
 } else {
  x = strtol( buf, &err, 10 );
  if ( !cs_is_blank( err ) ) err_set( JCM_ERR_READ_STRING );
  return ( (short)x  );
 }
}




logical read_q( const char* buf )
{
 buf += cs_begs( buf ) - 1;
 return ( cs_ss_char( "YyTt", *buf ) );
}


/*----------------------------------------------------------------------*/
/*
 *  write_i
 *
 * Write an integer to a string, left justified. Check to not overflow buffer.
 * Uses sprintf from <stdio.h>
 * Uses cs_copy_siz from csproto.h
 */

char* write_i( const integer val, char* buf, const integer siz )
{
 card tbuf;
 if ( qnan_i( val )) {
  cs_copy_siz( nanstring, buf, siz );
 } else {
  sprintf( tbuf, "%ld", (long)val );
  cs_copy_siz( tbuf, buf, siz );
 }
 return( buf );
}

char* write_s( const short val, char* buf, const integer siz )
{
 card tbuf;
 if ( qnan_s( val )) {
  cs_copy_siz( nanstring, buf, siz );
 } else {
  sprintf( tbuf, "%d", val );
  cs_copy_siz( tbuf, buf, siz );
 }
 return( buf );
}

char* write_r( const real val, char* buf, const integer siz )
{
 card tbuf;
 if ( qnan_r( val ) ) {
  cs_copy_siz( nanstring, buf, siz );
 } else if ( val == 0.0 ) {
  cs_copy_siz( "0", buf, siz );
 } else {
  if ( rabs( val ) > efmt_max || rabs( val ) < efmt_min ) {
   sprintf( tbuf, efmt, val );
  } else { 
   sprintf( tbuf, ffmt, val );
  }
  numberclean( tbuf, buf, siz );
 }
 return( buf );
}

char* write_d( const double val, char* buf, const integer siz )
{
 card tbuf;
 if ( qnan_d( val ) ) {
  cs_copy_siz( nanstring, buf, siz );
 } else if ( val == 0.0 ) {
  cs_copy_siz( "0", buf, siz );
 } else {
  if ( fabs( val ) > efmt_max || fabs( val ) < efmt_min ) {
   sprintf( tbuf, efmt, val );
  } else { 
   sprintf( tbuf, ffmt, val );
  }
  numberclean( tbuf, buf, siz );
 }
 return( buf );
}


char write_q( const logical q )
{
 return( q ? 'T' : 'F' );
}

integer read_i1( const char* buf )
{
 integer i;

 i = *buf - '0';
 if ( i < 0 || i > 9 ) i = 0;
 return( i );
}




static void numberclean( char* tbuf, char* buf, integer siz )
{
 integer head;
 integer tail;
 integer expt;
 integer decpt;

 if ( !siz ) return;
 utl_str_rfclean( tbuf );
 head = cs_begs( tbuf );
 tbuf += head-1;
 expt = cs_index_char( tbuf, 'E' );
 if ( expt > 0 ) {
  tail = expt -1 ;
 } else {
  tail = cs_ends( tbuf );
 }
 decpt = cs_index_char( tbuf, '.' );
 if ( decpt > 0 ) {
  while ( tail > decpt && tbuf[tail-1] == '0' ) tail--;
  if ( tail == decpt ) tail--;
 }
 cs_copy_siz( tbuf, buf, tail );
 if ( expt > 0 && !cs_eq( tbuf+expt-1, "E+00" ) ) {
  cs_copy( tbuf+expt-1, buf+tail );
 }
}

static void utl_str_rfclean( char* tbuf )
{
 integer j;
 j = cs_begs( tbuf );
 if ( j > 1 ) {
  if( cs_eq_ss( tbuf, j-1, j+1, " -." ) ) {
   cs_put_ss( "-0.", tbuf, j-1, j+1 );
  } else if ( cs_eq_ss( tbuf, j-1, j, " ." )) {
   cs_put_ss( "0.", tbuf, j-1, j );
  }
 }
}




char* writef_i( const integer val, const char* fmt, char* buf, const integer siz )
{
 card tbuf;
 if( cs_is_blank( fmt ) ) {
  (void)write_i( val, buf, siz );
 } else if ( qnan_i( val ) ) {
  cs_copy_siz( nanstring, buf, siz );
 } else {
  sprintf( tbuf, fmt, val );
  cs_copy_siz( tbuf, buf, siz );
 }
 return( buf );  
}


char* writef_d( const double val, const char* fmt, char* buf, const integer siz )
{
 card tbuf;
 if( cs_is_blank( fmt ) ) {
  (void)write_d( val, buf, siz );
 } else if ( qnan_d( val ) ) {
  cs_copy_siz( nanstring, buf, siz );
 } else {
  sprintf( tbuf, fmt, val );
  utl_str_rfclean( tbuf );
  cs_copy_siz( tbuf, buf, siz );
 }
 return( buf );  
}




char* writef_s( const short val, const char* fmt, char* buf, const  integer siz )
{
 card tbuf;
 if( cs_is_blank( fmt ) ) {
  (void)write_s( val, buf, siz );
 } else if ( qnan_s( val ) ) {
  cs_copy_siz( nanstring, buf, siz );
 } else {
  sprintf( tbuf, fmt, val );
  cs_copy_siz( tbuf, buf, siz );
 }
 return( buf );  
}


char* writef_r( const real val, const char* fmt, char* buf, const integer siz )
{
 card tbuf;
 if( cs_is_blank( fmt ) ) {
  (void)write_r( val, buf, siz );
 } else if ( qnan_r( val ) ) {
  cs_copy_siz( nanstring, buf, siz );
 } else {
  sprintf( tbuf, fmt, val );
  utl_str_rfclean( tbuf );
  cs_copy_siz( tbuf, buf, siz );
 }
 return( buf );  
}






char* write_opt_s( const char* prompt, const short val, char* buf, const integer siz )
{
 t_word tbuf;
 t_word tbuf1;
 t_word tbuf2;
 cs_conc1( prompt, cs_sparen( write_s( val, tbuf, WSIZE ), tbuf1 ), tbuf2  );
 cs_copy_siz( tbuf2, buf, siz );
 return( buf );
}

char* write_opt_i( const char* prompt, const integer val, char* buf, const integer siz )
{
 t_word tbuf;
 t_word tbuf1;
 t_word tbuf2;
 cs_conc1( prompt, cs_sparen( write_i( val, tbuf, WSIZE ), tbuf1 ), tbuf2 );
 cs_copy_siz( tbuf2, buf, siz );
 return( buf );
}

char* write_opt_r( const char* prompt, const real val, char* buf, const integer siz )
{
 t_word tbuf;
 t_word tbuf1;
 t_word tbuf2;
 cs_conc1( prompt, cs_sparen( write_r( val, tbuf, WSIZE ), tbuf1 ), tbuf2  );
 cs_copy_siz( tbuf2, buf, siz );
 return( buf );
}
char* write_opt_d( const char* prompt, const double val, char* buf, const integer siz )
{
 t_word tbuf;
 t_word tbuf1;
 t_word tbuf2;
 cs_conc1( prompt, cs_sparen( write_d( val, tbuf, WSIZE ), tbuf1 ), tbuf2  );
 cs_copy_siz( tbuf2, buf, siz );
 return( buf );
}

char* write_opt_c( const char* prompt, const char* val, char* buf, const integer siz )
{
 t_word tbuf1;
 t_word tbuf2;
 cs_conc1( prompt, cs_sparen( val, tbuf1 ), tbuf2  );
 cs_copy_siz( tbuf2, buf, siz );
 return( buf );
}







char* write_e_r( const real x, const real off, char* buf, const integer siz )
{
 if ( x == 0.0 ) {
  cs_copy( "0", buf );
 } else if ( qnan_r( x ) ) {
  cs_copy( nanstring, buf );
 } else {
  write_e_d( (double)x, (double)off, buf, siz );
 }
 return( buf );
}

char* write_e_d( const double x, const double off, char* buf, const integer siz )
{
 integer ioff;
 double x1;
 double roff;
 integer exponent;
 double mantissa;
 t_word tbuf;
 if ( x == 0.0 ) {
  cs_copy( "0", buf );
 } else if ( qnan_d( x ) ) {
  cs_copy( nanstring, buf );
 } else if ( off != 0.0 ) {
  frac_d( off, &ioff, &roff );
  x1 = x * pow( 10.0, roff );
  exponent = (integer)log10( fabs(x1) );
  if ( fabs(x1) < 1.0 ) exponent--;
  mantissa = x1/ tpow_d( exponent );
  exponent += ioff;
  if ( abs( exponent ) < 100 ) {
   sprintf( tbuf, "%8.5fE%+3.3ld", mantissa, (long)exponent );
  } else {
   sprintf( tbuf, "%7.4fE%+4.4ld", mantissa, (long)exponent );
  }
  cs_copy_siz( tbuf, buf, siz );
 } else {
  write_d( x, buf, siz );
 }
 return( buf );
}



char* write_tp_i( const integer i, char* buf, const integer siz )
{
 t_word tbuf;
 sprintf( tbuf, "{%ld}", (long)i );
 cs_copy_siz( tbuf, buf, siz );
 return( buf );
}


char* write_rj_i( const integer i, char* buf, const integer siz )
{
 t_word fmt;
 if ( qnan_i( i ) ) {
  cs_copy( nanstring, buf );
 } else {
  sprintf( fmt, "%%%ldd", (long)siz );
  sprintf( buf, fmt, i );
 }
 return( buf );
}


integer read_digit( const char c )
{
 integer i;
 i = c - '0';
 if ( i < 0 || i > 9 ) i = 0;
 return( i );
}

integer read_hex_digit( const char c )
{
 integer i;

 i = c - '0';
 if ( i < 0 || i > 9 ) {
  i = c - 'a';
  if ( i < 0 || i > 5 ) i = c - 'A';
  if ( i < 0 || i > 5 ) {
   i = 0;
  } else {
   i += 10;
  }
 }
 return( i );
}


