#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "jcmlib.h"

static integer read_val_sign( char** buf );
static integer read_val_int( char** buf, integer* digits );
static double read_val_frac( char** buf );

/*
 *   Read a value, splitting it into its sign, integer part and fractional part.
 *   Advance the pointer so it points to the end of the value, ready to
 *   read more.
 */

void read_val_d( char** buf, 
                 integer* sign_val,
		 integer* int_val,
		 double* frac_val
		)
{
 char* ptr;
 integer digits;

 *int_val = 0;
 *frac_val = 0.0;
 ptr = *buf;
 if ( !cs_is_blank( ptr ) ) {
  cs_advptr( &ptr );
  if ( nancode( ptr ) ) {
   ptr = NULL;
   *int_val = nan_i();
  } else {
   *sign_val = read_val_sign( &ptr );
   *int_val = read_val_int( &ptr, &digits );
   *frac_val = read_val_frac( &ptr );
  }
 } else {
  ptr = NULL;
 }
 *buf = ptr;
}

/*
 *  Read a value with exponent, returning sign, integer part, and fractional
 *  part.
 */

void read_val_de( char** buf, integer* sign_val, integer* int_val, double* frac_val )
{
 char* ptr;
 integer sign_exp, int_exp;
 double frac_exp;
 double evalue;
 double dvalue;
 ptr = *buf;
 read_val_d( &ptr, sign_val, int_val, frac_val );

 if ( cs_is_blank( ptr ) ) {
  ptr = NULL;
 } else if ( *ptr == 'E' || *ptr == 'D' ) {
  ptr++;
  read_val_d( &ptr, &sign_exp, &int_exp, &frac_exp );
  evalue = sign_exp * ( int_exp + frac_exp );
  if ( evalue != 0.0 ) {
   dvalue = ( *int_val + *frac_val ) * pow( 10.0, evalue );
   frac_d( dvalue, int_val, frac_val );
  }
 }
 *buf = ptr;
}

/*
 *  Read a value with exponent. returning sign, integer part, and fractional part
 *  and exponent.
 *
 */
void read_val_ee( char** buf, integer* sign_val, integer* int_val, double* frac_val, integer* exp_val )
{
 char* ptr;
 integer sign_exp, int_exp;
 double frac_exp;
 double dvalue;
 logical ten;
 double evalue;
 double base;

 ptr = *buf;
 *exp_val = 0;
 read_val_d( &ptr, sign_val, int_val, frac_val );

 if ( cs_is_blank( ptr ) ) {
  ptr = NULL;
 } else if ( cs_ss_char( "EeDd", *ptr ) ) {
  ptr++;
  read_val_d( &ptr, &sign_exp, &int_exp, &frac_exp );
  if ( frac_exp != 0.0 ) {
   if ( sign_exp < 0 ) {
    frac_exp = 1.0 - frac_exp; 
    int_exp++;
   }
   dvalue = ( *int_val + *frac_val ) * pow( 10.0, frac_exp );
   frac_d( dvalue, int_val, frac_val );
  }
  *exp_val = sign_exp * int_exp;
 } else if ( *ptr == 'x' ) {
/* General power notation    n x foo */
  ptr++;
  read_val_d( &ptr, &sign_exp, &int_exp, &frac_exp );
  ten = ( sign_exp == 1 && int_exp == 10 && frac_exp == 0.0 );
  if ( *ptr == '^' ) {    /* buf = n x foo ^ */
   base = sign_exp * ( int_exp + frac_exp );
   ptr++;
   read_val_d( &ptr, &sign_exp, &int_exp, &frac_exp );
   if ( ten ) {
    /* buf = n x 10 ^ foo */
    if ( frac_exp != 0.0 ) {
     if ( sign_exp < 0 ) { frac_exp = 1.0 -frac_exp; int_exp++; }
     dvalue = ( *int_val + *frac_val ) * pow( 10.0, frac_exp );
     frac_d( dvalue, int_val, frac_val );
    }
    *exp_val = sign_exp * int_exp;
   } else {  /* Not ten; buf = n x y ^ foo */
    evalue = sign_exp * ( int_exp + frac_exp );
    dvalue = pow( base, evalue );
    if ( dvalue < 0.0 ) { 
     dvalue *= -1;
     sign_exp = -1;
    } else {
     sign_exp = +1;
    }
    dvalue = ( *int_val + *frac_val ) * dvalue;
    frac_d( dvalue, int_val, frac_val );
    *sign_val *=  sign_exp;
    *exp_val = 0;
   }
  } 
 } else if ( *ptr == '^' ) {    /* buf = n ^ foo */
  ptr++;
  read_val_d( &ptr, &sign_exp, &int_exp, &frac_exp );
  ten = ( *int_val == 10 && *frac_val == 0.0 );
  if ( ten ) {  /* buf = 10 ^ foo */
   *int_val = 1;
   if ( frac_exp != 0.0 ) {
    if ( sign_exp < 0 ) { frac_exp = 1.0 -frac_exp; int_exp++; }
    dvalue = pow( 10.0, frac_exp );
    frac_d( dvalue, int_val, frac_val );
   }
   *exp_val = sign_exp * int_exp;
  } else {  /* Not ten */
   dvalue = *sign_val * ( *int_val + *frac_val );
   evalue = sign_exp  * (  int_exp +  frac_exp );
   dvalue = pow( dvalue, evalue );
   if ( dvalue < 0.0 ) {
    *sign_val = -1; dvalue *= -1;
   } else {
    *sign_val = +1;
   }
   frac_d( dvalue, int_val, frac_val );
   *exp_val = 0.0;
  }

 }
 *buf = ptr;
}

static integer read_val_int( char** buf, integer* digits )
{
 integer val;
 char* ptr;
 logical loop;
 integer n;
 
 ptr = *buf;
 val = 0;
 *digits = 0;
 loop = ( ptr != NULL );
 while( loop && *ptr != '\0' ) {
  n = *ptr - '0';
  loop = ( n >= 0 && n <= 9 );
  if ( loop ) {
   ptr++;
   if ( *digits < 9 ) {
    val = val * 10 + n;
    (*digits)++;
   }
  }
 }
 *buf = ptr;
 return( val );
}

static integer read_val_sign( char** buf )
{
 char* ptr;
 ptr = *buf;
 if ( *ptr == '\0' ) {
  return( 0 );
 } else  if ( *ptr == '+' ) {
  (*buf)++;
  return( +1 );
 } else if ( *ptr == '-' ) {
  (*buf)++;
  return( -1 );
 } else {
  return( +1 ); 
 }
}


static double read_val_frac( char** buf )
{
 double val;
 integer ival;
 integer digits;
 char* ptr;
 val = 0.0;
 ptr = *buf;
 if ( *ptr == '.' ) {
  ptr++;
  ival = read_val_int( &ptr, &digits );
  if ( digits > 0 ) {
   val = (double)ival / tpow_d( digits );
  }
 }
 *buf = ptr;
 return( val );
}
