#include "xgain.h"
/*
 * hri_read_aspect
 *
 * Read the aspect solution from the data file. Apply the boresight offsets. If the
 * aspect solution is given in RA, DEC format, convert to aspect offsets.
 *
 */
Aspect_Solution* xgain_read_aspect( char* anc_file, logical offsets, char** colnames, double* bso, double* nominal_point, double pixel_size )
{
 Aspect_Solution* asp;
 dmDataset* aspect_ds;

/*
 char* rev0_colnames[3] = { "ASPECT_TIME", "ASPECT_XOFF", "ASPECT_YOFF", "ASPECT_ROLL" };
 char* rev0_colnames[3] = { "TIME", "XOFF", "YOFF", "ROFF" };
 char* rdf_colnames[3] = { "TIME", "RA_SC", "DEC_SC", "ROAN_SC" };
 */


 aspect_ds = dmDatasetOpen( anc_file );
 if ( aspect_ds ) {
  dmsg( 1, "Opened Aspect DS" );
 } else {
  tmsg( "Failed to open aspect file" );
  exit( 1 );
 }

 asp = aspect_read( aspect_ds, bso, nominal_point, pixel_size, offsets, colnames );
 if ( asp->n == 0 ) tmsg( "Failed to read aspect" );

 dmDatasetClose( aspect_ds );

 return( asp );
}

/*
 * aspect_read
 *
 * Read aspect solution from ROSAT RDF format aspect file.
 * The RDF file has RA, Dec and Roll. The REV0 file has
 * pixel offsets. These pixel offsets have a different sign from
 * the standardized ones used in this program: in particular, the
 * sign of the y offset is different.
 */

Aspect_Solution* aspect_read(  
  dmDataset* aspect_ds, 	/* Aspect dataset pointer */
  double* bso,                  /* Boresight offsets */
  double* nominal_point,        /* Nominal pointing direction, not used for REV0 */
  double pixel_size,		/* Pixel size, not used for REV0 */
  logical offsets, 		/* True if dataset contains offsets (REV0) or absolute pointing (RDF) */
  char** colnames 		/* Colname for TIME and Three column names for RA DEC ROLL or X Y ROLL */
 )
{
 dmBlock* asp_table;
 Aspect_Solution* asp;

 double point[3];
 dmDescriptor* t_col;
 dmDescriptor* asp_col[3];
 int asp_tables[100];
 long asp_size[100];
 int ntables = 0;
 int tabno;
 card name;
 int block_no;
 long offset;
 double* data;
 double data1[3]; 
 int i;
 card buf;
 long n;
 int nblocks;
 dmsg( 1, "Open Aspect block" );
 n = 0;
 asp_table = dmBlockOpen( aspect_ds, "ASPECT" );
 if ( asp_table ) {
  ntables = 1;
  asp_tables[0] = dmBlockGetNo( asp_table );
  asp_size[0] = dmTableGetNoRows( asp_table ); 
  n = asp_size[0];
  printf( "Found table ASPECT  Aspect records %d\n",(int) n );
 } else {
  nblocks = dmDatasetGetNoBlocks( aspect_ds );
  printf( "ASPECT dataset has %d blocks\n", nblocks );
  for ( block_no = 1; block_no <= nblocks; block_no++ ) {
   
   asp_table = dmDatasetMoveToBlock( aspect_ds, block_no );
   dmBlockGetNameC( asp_table, name, CSIZE );
   if ( cs_eq( name, "AO" ) ) {
    ntables++;
    asp_tables[ntables-1] = block_no;
    asp_size[ntables-1] = dmTableGetNoRows( asp_table );
    n += asp_size[ntables-1];
    printf( "Found table %s Aspect records %d\n", name, (int)asp_size[ntables-1] );
   } else {
    printf( "Found table %s\n", name );
   }
  }
 }
 if ( ntables == 0 ) tmsg( "Failed to open aspect table" );
 dmsg( 1, "Open Aspect table" );
 
 asp = alloc_asp( n );
 asp->n = 0;
 for ( tabno = 0; tabno < ntables; tabno++ ) {
  printf( "Process table %d block %d\n", tabno+1, (int)asp_tables[tabno] );
  asp_table = dmDatasetMoveToBlock( aspect_ds, asp_tables[tabno] );
 if ( !asp_table ) tmsg( "Error moving to aspect table" );
 t_col =  dmTableOpenColumn( asp_table, colnames[0] );
 if ( !t_col ) {
  printf( "Error opening aspect table column %s\n", colnames[0] );
 }
 asp_col[X_OFF] = dmTableOpenColumn( asp_table, colnames[X_OFF+1] );
 asp_col[Y_OFF] = dmTableOpenColumn( asp_table, colnames[Y_OFF+1] );
 asp_col[ROLL]  = dmTableOpenColumn( asp_table, colnames[ROLL+1] );
 if ( ( !asp_col[X_OFF] ) ) {
  printf( "Error opening table columns %s\n", colnames[X_OFF+1] );
 } else {
  printf( "Reading columns %s %s %s\n", colnames[X_OFF+1], colnames[Y_OFF+1], colnames[ROLL+1] );
 }
 sprintf( buf, "Read Aspect table %d\n", tabno+1 ); dmsg( 1, buf );
 for ( i = 0; i < asp_size[tabno]; i++ ) {
  offset = i + asp->n;
  asp->t[offset] = dmGetScalar_d( t_col );
  if ( offsets ) {
  point[0] = dmGetScalar_r( asp_col[X_OFF] );
  point[1] = dmGetScalar_r( asp_col[Y_OFF] );
  point[2] = dmGetScalar_r( asp_col[ROLL] );
  } else {
  point[0] = dmGetScalar_d( asp_col[X_OFF] );
  point[1] = dmGetScalar_d( asp_col[Y_OFF] );
  point[2] = dmGetScalar_d( asp_col[ROLL] );
  }
  dmTableNextRow( asp_table );
  data = asp->data[offset];
  if ( offsets ) { 
   data1[X_OFF] =  point[X_OFF];
   data1[Y_OFF] = -point[Y_OFF];
   data1[ROLL]  =  radout_d( point[ROLL] );
  } else {
   point_to_asp( point, nominal_point, pixel_size, data1 );
  }
   bore_correct( bso, data1 );
   rotate2_d( data1[X_OFF], data1[Y_OFF], -nominal_point[ROLL], &data[RA], &data[DEC] );
   data[ROLL] = data1[ROLL] + nominal_point[ROLL];

 }
 dmsg( 1, "Close Aspect table" );
 dmBlockClose( asp_table );
 asp->n += asp_size[tabno];
 }
 printf( "Aspect records: %d\n", (int)asp->n);
 return( asp );
}


/*
 * Allocate an aspect solution structure with
 * a given number of entries. Space is reserved for n entries
 * of times and corresponding aspect entries. Each aspect entry
 * is X pixel offset, Y pixel offset, roll offset relative to nominal.
 */

Aspect_Solution* alloc_asp( int n )
{
 Aspect_Solution* asp;
 asp = (Aspect_Solution* )malloc( sizeof( Aspect_Solution ) );
 asp->t = (double*)malloc( n*sizeof( double ) );
 asp->data = vec_alloc_d( n );
 asp->nmax = n;
 asp->n = 0;
 return( asp );
}

/*
 * Free an aspect solution structure.
 */
void free_asp( Aspect_Solution* asp )
{
 free( asp->t );
 vec_free_d( asp->data );
 free( asp );
}

/*
 *  Adjust the aspect offsets for the boresight corrections.
 *  The aspect offsets in the Rev0 files need to be adjusted
 *  by these values (probably).
 *  First we rotate the boresight x and y offsets through the
 *  aspect roll to get them in 'nominal' space, then we 
 *  subtract the boresight roll offset from the aspect roll.
 *  (I have no idea why the signs are the way they are, this
 *  was just worked out from the HRI code).
 */

void bore_correct( double* bso, double* asp )
{
 double dx, dy;
 rotate2_d( bso[RA], bso[DEC], asp[ROLL], &dx, &dy );
 asp[X_OFF]   += dx;
 asp[Y_OFF]   -= dy;
 asp[ROLL]    -= bso[ROLL];
}

/*
 *
 *  Calculate the aspect offsets from the actual and nominal pointings.
 *  The roll offset is the difference between actual and nominal roll.
 *  The X and Y sky coordinate offsets in pixels are given
 *  by the differences in pointing direction, approximated as
 *   (RA-RA0)cos(Dec),  (Dec-Dec0)
 *  converted from degrees to pixels.
 *  The X and Y detector coordinate offsets are the sky coordinate offsets
 *  rotated through minus the nominal roll angle.
 *
 */

void point_to_asp( double* point, double* nominal, double pixel_size, double* asp )
{
 double sky_offset[2];

 sky_offset[RA] =  - ( ( point[RA] - nominal[RA] ) * dcosd( nominal[DEC] ) ) / pixel_size;
 sky_offset[DEC] =   ( point[DEC] - nominal[DEC] ) / pixel_size;
 asp[ROLL] = point[ROLL] - nominal[ROLL];
 rotate2_d( sky_offset[RA], sky_offset[DEC], nominal[ROLL], &asp[X_OFF], &asp[Y_OFF] );
}





#if 0
/* 
 * AN ALTERNATIVE APPROACH - ROW BASED IO
 */
Aspect_Solution* rev0_read_aspect( dmDataset* aspect_ds, double* bso  )
{
 dmBlock* asp_table;
 Aspect_Solution* asp;
 double* data;
 int i;
 int n;
 struct { double time; double xoff; double yoff; double roff; } asp_row;
 asp_table = dmBlockOpenWindow( aspect_ds, "AO", "TIME,XOFF,YOFF,ROFF", "d,d,d,d" );
/* Force returned row to be these named columns in this order, after casting to the
  type specified in the type argument. */
 n = dmTableGetNoRows( asp_table );
 asp = alloc_asp( n );
 for ( i = 0; i < n; i++ ) {
/* The int dmTableGetNextRow( void* ) call would place values in the asp_row structure, trusting that
   the user has made the struct compatible with the OpenWindow call. It would return the row number
   as the function value. */
  (void)dmTableGetNextRow( &asp_row );
  asp->t[i] = asp_row.time;
  data = asp->data[i];
  data[0] = asp_row.xoff;
  data[1] = asp_row.yoff;
  data[2] = radout_d( asp_row.roff );
  bore_correct( bso, data );
 }
 dmBlockClose( asp_table );
 asp->n = n;
 return( asp );
}
#endif


