#include "xgain.h"

/*
 * xgain_cal_read
 *
 * Read the HRI gain calibration. Open the MAPS table, and read the gain
 * maps versus time.
 *
 */

static void time_slice( GainCal* gain_data ); /* debug routine */

GainCal* xgain_cal_read( char* maps )
{
 dmBlock* maps_table;
 GainCal* gain_data;
 int n;
 int i;
 card buf;
 double* dptr;
 dmDescriptor* time_col;
 dmDescriptor* GainMap_col;
 int nsteps = 50; /* Large number for array of possible HV changes during mission; actual so far = 4 */

 maps_table = dmDatasetTableOpen( maps );
 if (!maps_table) {
   tmsg( "Failed to open gain calibration dataset" );
   return( NULL );
 }
 n = dmTableGetNoRows( maps_table );
 gain_data = alloc_gain_data( n, nsteps );
 time_col     = dmTableOpenColumn( maps_table, "TIME" );
 GainMap_col =  dmTableOpenColumn( maps_table, "GAIN_MAP" );
 if ( time_col && GainMap_col ) {
  dmsg( 1, "Map  Time" );

  for ( i = 0; i < n; i++ ) {
   gain_data->time[i] = dmGetScalar_d( time_col );
   dptr = (double*)gain_data->maps[i];
   dmGetStaticArray_d( GainMap_col, dptr, GMAPSIZE*GMAPSIZE );
   sprintf( buf, "%6d %p %20.6f %f %f %f", i+1, 
    GainMap_col, gain_data->time[i], dptr[0], dptr[200], dptr[500] );
   dmsg( 1, buf );
   dmTableNextRow( maps_table );
  }
  printf( "Read %d maps from gain calibration table\n", n );
 } else {
  tmsg( "ERROR: TIME and GAIN_MAP columns not found" );
 }
 dmDatasetTableClose( maps_table ); 
 if ( dbg_level(1) )  time_slice( gain_data );
 return( gain_data );
}




static void time_slice( GainCal* gain_data )
{
 int bin;
 card buf;
 double gain;
 double gbar, gvar;
 double* dptr;
 int out;
 double time;
 double gx, gy;
 gx = 16.0;
 gy = 16.0;
 out = fio_open_w( "gain.time" );
 io_out( out );
 msg( "BIN  TIME       GAIN" );
 for ( bin = 0; bin < gain_data->nmap; bin++ ) {
  time = gain_data->time[bin];
  gain = interp_gain( &gain_data->maps[bin], gx, gy );
  dptr = (double*)gain_data->maps[bin];
  ar_stats_d( dptr, GMAPSIZE * GMAPSIZE, &gbar, &gvar );
  sprintf( buf, "%4d %20.3f %10.3f %10.3f %10.3f", bin, time, gain, gbar, gvar );
  msg( buf );
 }
 fio_close( out );
}


/*
 * Allocate a gain data structure
 */

GainCal* alloc_gain_data( int n, int ns )
{
 GainCal* data;
 data = ( GainCal * ) malloc( sizeof( GainCal ) );
 data->time = (double *)malloc( n * sizeof( double ) );
 data->maps = (GainMap *)malloc( n * sizeof( GainMap ) );
 data->nmap = n;
 data->nsteps = 0;
 data->tsteps = (double *)malloc( ns * sizeof( double ) );
 return( data );
}


/*
 *  Free a gain_data structure.
 */
void free_gain_data( GainCal* data )
{
 free( data->maps );
 free( data->time );
 free( data->tsteps );
 free( data );
}


void read_gain_steps_file( GainCal* data, char* steps_file )
{
 /* The Gain steps file is an ASCII file whose first free form column is
    a list of MJD values that we will use as interpolation boundaries.
   Right now it is the list of times of HV change, but it could be any set
   of times at which the gain properties change.
 */
 int fid;
 card buf;
 t_word word;
 char* ptr;
 if ( cs_is_blank( steps_file ) ) return;
 fid = fio_open_r( steps_file );
 if ( fid == 0 ) {
  sprintf( buf, "Failed to open steps file %s", steps_file );
  tmsg( buf );
 } else {
  while( fio_read( fid, buf, CSIZE ) ) {
   ptr = buf;
   cs_get_c( &ptr, word, WSIZE );
   if ( word[0] != '#' && !cs_is_blank( word ) ) {
    data->tsteps[data->nsteps] = read_d( word );
    data->nsteps++;
   }
  }
  fio_close( fid );
  sprintf( buf, "Read %d steps from gain steps file\n", data->nsteps );
  tmsg( buf );
 } 
}
