#include "xgain.h"

/*
 * do_interp
 *
 * Interpolate in gain maps at given time and detector position.
 * Return the interpolated gain.
 *
 * First find the two time bins on either side of the given time. Then
 * for each of these bins, interpolate spatially in the corresponding gain map.
 * Finally interpolate in time between the two results.
 */
double do_interp( GainCal* gain_data, double t, double detx, double dety )
{
 double gx;
 double gy;
 double tol = 50.0; /* Extend calibrated period */
 int time_bin;
 double time_low, time_high;
 double gain_low, gain_high;
 double gain;
 card buf;
 if( t < gain_data->time[0] - tol || t > gain_data->time[gain_data->nmap-1] + tol  ) {
  tmsg( "*** Observation time outside calibrated period ***" );
  sprintf( buf, "Calibrated period is: MJD %f to %f", 
   gain_data->time[0] - tol, gain_data->time[gain_data->nmap-1]+tol );
  tmsg( buf );
  sprintf( buf, "Observation time is: MJD %f\n", t );
  tmsg( buf );
  exit( 1 );
 } 
 gx = detx / 256.0;
 gy = dety / 256.0;
 time_bin = find_entry_and_step_d( t, gain_data->time, gain_data->nmap, gain_data->tsteps, gain_data->nsteps );
 sprintf( buf, "Time = %20.6f %d", t, time_bin );
 dmsg( 2, buf );
 time_low  = gain_data->time[time_bin-1];
 time_high = gain_data->time[time_bin];

 gain_low  = interp_gain( &gain_data->maps[time_bin-1], gx, gy );
 gain_high = interp_gain( &gain_data->maps[time_bin],   gx, gy );
 gain = lint_d( time_low, time_high, gain_low, gain_high, t );
 return( gain );
}



/*
 * find_entry_d (old version)
 *
 * Search through a double precision array, finding the entries which
 * bracket the given target value. Return the right hand entry m, such that
 * value is between array(m-1) and array(m). The returned value of m
 * shall be between 1 and n-1.
 *   Should replace this with ar_cbisect_d( time, times, n ) - 1.
 */
int find_entry_d( double time, double* times, int n )
{
 int i;
 for ( i = 1; i < n; i++ ) {
  if ( times[i] > time ) return( i );
 }
 return( n-1 );
}

/*
 * find_entry_d (new version)
 *
 * Search through a double precision array, finding the entries which
 * bracket the given target value. Return the right hand entry m, such that
 * value is between array(m-1) and array(m). The returned value of m
 * shall be between 1 and n-1.
 * However, make sure that m-1 and m do not cross a step. Search only within
 * time ranges in the same step as t.
 */
int find_entry_and_step_d( double time, double* times, int n, double* steps, int ns )
{
 int i;
 double* subarray;
 int step_no;
 int bin1,bin2;
 double step1, step2;
 if ( ns == 0 ) {
  return( find_entry_d( time, times, n ) );
 }
/* time is between step(step_no-1) and step(step_no) */
 if ( time >= steps[ns-1] ) {
  step_no = ns;
  step1 = steps[ns-1];
  step2 = step1 * 10.0; /* dummy value */
 } else if ( time <= steps[0] ) {
  step_no = 0;
  step1 = 0;
  step2 = steps[0];

 } else {
  step_no = find_entry_d( time, steps, ns );
  step1 = steps[step_no-1];
  step2 = steps[step_no];
 }

/* Now find subarray of times within this step */
 if ( times[0] > step1 ) { 
  bin1 = 0;
 } else {
  bin1 = find_entry_d( step1, times, n );
 }
 if ( times[n-1] < step2 ) {
  bin2 = n-1;
 } else {
  bin2 = max_i( find_entry_d( step2, times, n ) - 1, 1 );
 }
 if ( bin1 == bin2 ) { 
  bin1--; 
  tmsg( "Warning: interpolating across step boundary" ); 
 }
/* Now find bin of that subarray which time is in */
 subarray = times + bin1;
 i = find_entry_d( time, subarray, bin2-bin1+1 ) + bin1;
#if 0
 printf( "CAL1  = %20.2f 0\n", times[0] );
 printf( "STEP1 = %20.2f Step %d\n", step1, step_no-1 );
 printf( "BIN1  = %20.2f %d\n", times[bin1], bin1 );
 printf( "TIME1 = %20.2f %d\n", times[bin1+i-1], bin1+i-1 );
 printf( "TIME  = %20.2f\n", time );
 printf( "TIME2 = %20.2f %d\n", times[bin1+i], bin1+i );
 printf( "BIN2  = %20.2f %d\n", times[bin2], bin2 );
 printf( "STEP2 = %20.2f Step %d\n", step2, step_no );
 printf( "CAL2  = %20.2f %d\n", times[n-1], n-1 );
#endif
 return( i );
}

/*
 * interp_gain
 *
 * Binlinearly interpolate in a GMAPSIZE x GMAPSIZE gain map.
 *
 * Note that the 2D array was read in in Fortran order so that
 * the first axis is Y and the second is X.
 */
double interp_gain( GainMap* map, double gx, double gy )
{
 int ix;
 card buf;
 int iy;
 double t, u;
 double y1, y2, y3, y4;
 double gain;

 ix = (int)gx;
 iy = (int)gy;
 t = gx - ix;
 u = gy - iy;
/*
 y1 = (*map)[iy-1][ix-1];
 y2 = (*map)[iy-1][ix];
 y3 = (*map)[iy][ix];
 y4 = (*map)[iy][ix-1];
*/
 y1 = (*map)[ix-1][iy-1];
 y2 = (*map)[ix][iy-1];
 y3 = (*map)[ix][iy];
 y4 = (*map)[ix-1][iy];
 gain = ( 1 - t ) * ( 1 - u ) * y1 + t * ( 1 - u ) * y2 + t * u * y3 + ( 1 - t ) * u * y4;
 sprintf(buf , "GAIN: %4d %4d %8.4f %8.4f %8.4f %8.4f", ix, iy, y1, y2, y3, y4 );
 dmsg( 2, buf );
 return( gain );
}


