#include "xgain.h"
/*
 * xgain_calc
 * 
 * Loop for each row in the aspect data.
 * First calculate the amount of good time in the aspect row.
 * In practice the aspect rows are usually 1.0 second long and 
 * the good time is mostly either 1.0s or 0.0s.
 * Then for each pixel in the source, calculate the gain using the aspect solution. 
 * Add the time for this pixel to the gain histogram. Also accumulate totals of other useful info.
 */
void xgain_calc( Source* source, Aspect_Solution* aspect, Observation* obs, GainCal* gain_data, t_ghist* ghist )
{
 GTI* gti;                 /* Good time intervals */
 int row;
 double t1;
 double t2;
 double wt;
 t_buf bbuf;
 int spatial_pixel;
 double* asp;
 double MJD;
 double detx, dety;
 double detx1, dety1;
 double xcen, ycen;
 double gain;

 xcen = obs->cen[0];
 ycen = obs->cen[1];
/* Convert to raw coord convention */

 gti = obs->gti; 
 cs_copy( " ", bbuf );
 tmsg( " " );
 for ( row = 0; row < aspect->n - 1; row++ ) {
  t1 = aspect->t[row];
  t2 = aspect->t[row+1];
  wt = in_gti( t1, t2, gti );
  if (  wt > 0.0 ) {  
   asp = aspect->data[row];
   MJD = sc_time_to_mjd( t1, obs->tzero );
   for ( spatial_pixel = 0; spatial_pixel < source->n; spatial_pixel++ ) {
    
    unaspect( source->xpos[spatial_pixel], source->ypos[spatial_pixel], xcen, ycen, asp, &detx, &dety );
    gain = do_interp( gain_data, MJD, detx, dety );
    detx1 = detx;
    dety1 = dety;
    if( dbg_level( 2 ) ) {
    sprintf( bbuf, "%d %12.4f %12.4f %12.4f %12.4f %12.4f %10.4f %10.4f %8.4f %f %f\n", 
     row, detx1, dety1,
     detx, dety, asp[0], asp[1], asp[2], gain, t1, MJD );
    dmsg( 2, bbuf );
    }
    xgain_accumulate_histogram( gain, wt/source->n, t1, t2, detx, dety, asp, ghist, source->n );
   }
  }
 }
 
}

/*
 * sc_time_to_mjd
 *
 * Convert spacecraft time to MJD. 
 * Converts seconds to days and adds to the zero point in days.
 */
double sc_time_to_mjd( 
 double t,                /* Spacecraft time in seconds */
 double mjd_zero          /* Zero point: MJD of time = 0 */
)
{
 double mjd;
 const double DL_SEC = 86400.0;  /* Seconds in day */
 mjd = t / DL_SEC + mjd_zero;
 return( mjd );
}

/*
 * Accumulate results into gain histogram
 */
void xgain_accumulate_histogram( double gain, double wt, double t1, double t2, double detx, double dety, double* xasp, t_ghist* ghist, int npos )
{
 int bin;
 int i;
 const int ngti = 100;

 bin = ( gain - ghist->g0 ) / ghist->dg;
 if ( bin >= 0 && bin < ghist->ng ) {
  ghist->data[bin] += wt;
 } else {
  return;
 }
 if ( ghist->gt1[bin] < 1.0 ) {    /* Really == 0.0 but check for rounding errors */
  ghist->gt1[bin] = t1;
 } else if ( t1 < ghist->gt1[bin] ) {
  ghist->gt1[bin] = t1;
 }
 if ( t2 > ghist->gt2[bin] ) ghist->gt2[bin] = t2;
 if ( npos == 1 ) {
/* Single pixel source case: record time filter for each bin.
  The time filter can be used to refilter the event file to get only
  source photons with a given gain. 
   We don't bother doing this for the multiple pixel source case because it doesn't make sense there.
   (the passes through this routine are no longer in time order so
   we lose track of the correct intervals, and a single time no longer
   corresponds to a single gain value.)
   
   We only allocate the GTI for a bin if it is going to be used.
   We allocate 100 entries maximum for it.
 */
  if ( ghist->gti[bin] == NULL ) {
   ghist->gti[bin] = alloc_gti( ngti );
  }
  accum_time_filter( t1, t2, ghist->gti[bin] );
 }

/* Accumulate mean and variance of aspect parameters */
 ghist->mean.x[bin] += wt * detx;
 ghist->mean.y[bin] += wt * dety;
 ghist->var.x[bin]  += wt * detx * detx;
 ghist->var.y[bin]  += wt * dety * dety;
 for ( i = 0; i < 3; i++ ) {
  ghist->mean.asp[i][bin] += wt * xasp[i];
  ghist->var.asp[i][bin]  += wt * xasp[i] * xasp[i];
 }
 
}

/*
 * Add a time interval to a list. If it is contiguous with the previous
 * time interval (to within a tolerance of 0.5 seconds)
 * just extend the size of that one, otherwise start a new one.
 * 
 */

void accum_time_filter( double t1, double t2, GTI* gti )
{
 int n;
 const double tolerance = 0.5;
 n = gti->n;
  if ( n == 0 ) {
   /* Begin first time interval */

   n = 1;
   gti->start[0] = t1;
   gti->stop[0]  = t2;
  } else if ( t1 > gti->stop[n-1] + tolerance ) {
    /* Begin a new time interval */
    if ( n < gti->nmax ) n++;
    gti->start[n-1] = t1;
    gti->stop[n-1]  = t2;
  } else if ( t1 < gti->stop[n-1] - tolerance ) {    
    tmsg( "Warning: times not consecutive" );
  } else {
    /* Add this time interval to the current one */
    gti->stop[n-1] = t2;
  }
  gti->n = n;
}


/*
 * unaspect
 *
 * Convert from sky pixel coords to detector coords.
 * Specialized to ROSAT HRI.
 */

void unaspect( double x, double y, double xcen, double ycen, double* asp, double* detx, double* dety )
{
 double r_x, r_y;
/* Convert to raw coord convention */
/* Offset from center */
 r_x = x - xcen - asp[RA];
 r_y = y - ycen - asp[DEC];
/* Apply roll */
 rotate2_d( r_x, r_y, asp[ROLL], detx, dety );
/* Offset from lower left */
 *detx += xcen;
 *dety += ycen;
}


/*
 * Routine to list the time intervals for each gain bin,
 * for the case of a point source.
 */
void  xgain_list_times( t_ghist* ghist )
{
 int i;
 double dt;
 double gain;
#define PMAX 511
 char buf[PMAX];
 char* ptr;
 int j;
 int k;
 t_word tmp1;
 t_word tmp2;
 GTI* gti;


 msg( "Gain   Time filter" );
 for ( i = 0; i < ghist->ng; i++ ) {
  dt = ghist->data[i];
  if ( dt > 0.0 ) {
   gain = ghist->g0 + ( i + 0.5 ) * ghist->dg;
   cs_null( buf, PMAX );
   ptr = buf;
   ptr += sprintf( ptr, "%6.2f  ", gain );
   gti = ghist->gti[i];
   if ( gti == NULL ) {
    sprintf( ptr, "()" );
   } else if ( gti->n > 10 ) {
    sprintf( ptr, "(Many time intervals)" );
   } else {
    *ptr++ = '(';
    for ( j = 0; j < gti->n; j++ ) {
     if ( j >= 1 ) *ptr++ = ',';
     writef_d( gti->start[j], "%12.2f", tmp1, WSIZE );
     writef_d( gti->stop[j],  "%12.2f", tmp2, WSIZE );
     k = cs_begs( tmp1 )-1;
     cs_copy( tmp1 + k, ptr );
     ptr += cs_ends( ptr );
     *ptr++ = ':';
     k = cs_begs( tmp2 )-1;
     cs_copy( tmp2 + k, ptr );
     ptr += cs_ends( ptr );
     if ( ptr - buf > PMAX - 20 ) { tmsg( "Buffer overflow" ); msg( buf); return; }
    } /* end for j */
    *ptr++ = ')';    
    *ptr = '\0';
   } /* end if ntimes */
   msg( buf );
  } /* end dt > 0 */
 } /* end for i */
}

/*
 * List the results of the program to the message stream.
 * Give the main result: the normalized weight in each gain bin.
 * For each bin, also include the mean and standard deviation of the 
 * source detector position and the aspect solution, for informational purposes.
 * Finally, print the overall mean gain for the source.
 */

void xgain_list( t_ghist* ghist )
{
 double gsum;
 int asp_element;
 double time_in_gain_bin;
 double wt; /* Weight */
 double sigma_detx;
 double sigma_dety;
 double detx, dety;
 double asp[3];
 double sigma_asp[3];
 double gain;
 int gain_bin;
 t_buf buf;
 double total_time;
 double fraction_in_gain_bin;
 msg( " " );
 msg( "Gain histogram: " );
 msg( "Gain    Weight     Time(s)    Start       Stop     DETX               DETY                AX                AY                  ROLL" );
 msg( " " );

 total_time = ghist->total_time;
 gsum = 0.0;
 for ( gain_bin = 0; gain_bin < ghist->ng; gain_bin++ )  {
  time_in_gain_bin = ghist->data[gain_bin];
  if ( time_in_gain_bin  > 0.0 ) {
   fraction_in_gain_bin = time_in_gain_bin / total_time;
   wt = 1.0 / time_in_gain_bin;
   gain = ghist->g0 + ( gain_bin + 0.5 ) * ghist->dg;
   detx = ghist->mean.x[gain_bin] * wt;
   dety = ghist->mean.y[gain_bin] * wt;
   sigma_detx = psqrt(ghist->var.x[gain_bin] * wt - detx * detx );
   sigma_dety = psqrt(ghist->var.y[gain_bin] * wt - dety * dety );
   for ( asp_element = 0; asp_element < 3; asp_element++ ) {
    asp[asp_element] = ghist->mean.asp[asp_element][gain_bin] * wt;
    sigma_asp[asp_element] = psqrt( ghist->var.asp[asp_element][gain_bin] * wt -  asp[asp_element] * asp[asp_element] );
   }
   sprintf( buf, "%6.2f%10.4f%10.4f%12.2f%12.2f%7.1f +/-%7.1f %7.1f +/-%7.1f %7.1f +/-%7.1f %7.1f +/-%7.1f %7.1f +/-%7.1f",
    gain, fraction_in_gain_bin, time_in_gain_bin, ghist->gt1[gain_bin], ghist->gt2[gain_bin], detx, sigma_detx, dety, sigma_dety,
    asp[X_OFF], sigma_asp[X_OFF], asp[Y_OFF], sigma_asp[Y_OFF], asp[ROLL], sigma_asp[ROLL] );
   msg( buf );
   gsum += gain * fraction_in_gain_bin;
  }
 } 
 msg( " " );
 sprintf( buf, "Mean Gain: %10.2f", gsum );
 msg( buf );
 msg( " " );

}




/*
 * Set up the gain histogram.
 * Specify the min and max of the histogram and the step size.
 * Specify the maximum number of time intervals.
 *
 */


t_ghist* ghset( double* g )
{
 double gmax;
 double g0, dg;
 t_ghist* ghist;
 int ng;
 dg = g[2];
 g0 = g[0] - 0.5* dg;
 gmax=g[1] + 0.5* dg;
 ng = ( gmax - g0 ) / dg + 1;
 ghist = alloc_ghist( ng );
 ghist->dg = dg;
 ghist->g0 = g0;
 ghist->ng = ng;
 return( ghist );
}



/*
 * Allocate gain history
 *
 */

t_ghist* alloc_ghist( int ng )
{
 t_ghist* g;
 GainHist_Accum* mean;
 GainHist_Accum* var;

 g =(t_ghist *)malloc( sizeof( t_ghist ) );
 g->ng = ng;
 g->data =  (double *)calloc( ng, sizeof( double ) );

 g->gt1 = (double *)malloc( ng * sizeof( double ) );
 g->gt2 = (double *)malloc( ng * sizeof( double ) );

 mean = &g->mean;
 var  = &g->var;

 mean->x = (double *)calloc( ng,  sizeof( double ) );
 mean->y = (double *)calloc( ng,  sizeof( double ) );
 mean->asp[X_OFF] =(double *)calloc( ng * 3,  sizeof( double ) );
 mean->asp[Y_OFF]  = mean->asp[X_OFF] + ng;
 mean->asp[ROLL]   = mean->asp[Y_OFF] + ng;

 var->x  =(double *)calloc( ng,  sizeof( double ) );
 var->y  =(double *)calloc( ng,  sizeof( double ) );
 var->asp[X_OFF]=(double *)calloc( ng * 3,  sizeof( double ) );
 var->asp[Y_OFF] = var->asp[X_OFF] + ng;
 var->asp[ROLL]  = var->asp[Y_OFF] + ng;


 g->gti = ( GTI ** ) calloc( ng, sizeof( GTI* ) );

 return( g );
}

/*
 * Free gain histogram
 */

void free_ghist( t_ghist* g )
{
 int i;

 for ( i = 0; i < g->ng; i++ ) {
  free_gti( g->gti[i] );
 }
 free( g->gti );
 free( g->gt1 );
 free( g->gt2 );
 free( g->mean.x );
 free( g->mean.y );
 free( g->mean.asp[0] );
 free( g->var.x );
 free( g->var.y );
 free( g->var.asp[0] );
 free( g->data );
 free( g );
}


/* 
 * psqrt 
 *
 * Positive argument square root
 * Forces argument to be at least zero
 *
 */
double psqrt( double x )
{
 return( x > 0.0 ? sqrt( x ) : 0.0 );
}



void xgain_write( char* filename, t_ghist* ghist )
{
 dmBlock* block;
 t_buf blockname;

 double gsum;
 double time_in_gain_bin;
 double gain;
 int gain_bin;
 t_buf buf;
 double total_time;
 double fraction_in_gain_bin;

 dmDescriptor* gain_col;
 dmDescriptor* wt_col;
 dmDescriptor* t_col;
 sprintf( blockname, "%s[%s]", filename, "GAIN_HIST" );
 block = dmDatasetTableCreate( blockname );
 if ( !block ) { 
  tmsg( "Failed to create output file" );
  return;
 }
 gain_col = dmColumnCreate( block, "GAIN", dmDOUBLE, 0, " ", "Gain" );
 wt_col = dmColumnCreate( block, "WEIGHT", dmDOUBLE, 0, " ", "Gain histogram" );
 t_col  = dmColumnCreate( block, "EXPOSURE", dmDOUBLE, 0, "s", "Exposure time per bin" );
 total_time = ghist->total_time;
 gsum = 0.0;
 for ( gain_bin = 0; gain_bin < ghist->ng; gain_bin++ )  {
  time_in_gain_bin = ghist->data[gain_bin];
  fraction_in_gain_bin = time_in_gain_bin / total_time;
  gain = ghist->g0 + ( gain_bin + 0.5 ) * ghist->dg;
  dmTablePutRow( block, NULL );
  dmSetScalar_d( gain_col, gain );
  dmSetScalar_d( wt_col, fraction_in_gain_bin );
  dmSetScalar_d( t_col,  time_in_gain_bin );
  gsum += gain * fraction_in_gain_bin;
 } 
 msg( " " );
 sprintf( buf, "Mean Gain: %10.2f", gsum );
 dmKeyWrite_d( block, "AVG_GAIN", gsum, " ", "Mean gain" );
 msg( buf );
 msg( " " );
 dmDatasetTableClose( block );
}
