#include "xgain.h"


/*
 * Routines dealing with GTIs (Good Time Intervals). 
 * GTIs are sets of (start,stop) time pairs and represent times when the
 * data is valid. The values are measured
 * in seconds from an arbitrary start time, usually the start of the mission.
 * The start time is stored in the observation structure.
 */

/*
 * dataset_get_gti
 *
 * Read the good time intervals from the STDGTI table. Calculate the total good time and
 * store it in the GTI structure.
 * Close the GTI table but leave the dataset open.
 */
GTI* dataset_get_gti( dmDataset* event_ds )
{
 dmBlock* gti_table;
 GTI* gti;
 card buf;
 dmsg( 1, "Open STDGTI" );
 gti_table = dmBlockOpen( event_ds, "STDGTI" );
 if ( !gti_table ) gti_table = dmBlockOpen( event_ds, "GTI" );
 if ( gti_table ) {
 dmsg( 1, "Read GTI" );
 gti = read_gti( gti_table );
 dmBlockClose( gti_table );
 gti->total = sum_gti( gti );
 sprintf( buf, "GTI records: %6d Total time %f", gti->n, gti->total );
 tmsg( buf );
 } else {
  tmsg( "Failed to open GTI" );
  exit( 1 );
 }
 return( gti );
}


/*
 * read_gti
 *
 * Read a GTI table and return a GTI structure.
 * The GTI table is assumed to contain start and stop times in the
 * first two columns of the table. We make no assumptions about the column names.
 *
 */

GTI* read_gti( dmBlock* table )
{
 GTI* gti;
 dmDescriptor* start;
 dmDescriptor* stop;
 int n;
/* The first two columns whatever they are called */
 dmsg( 1, "Open columns" );
 start = dmTableOpenColumnNo( table, 1 );
 stop  = dmTableOpenColumnNo( table, 2 );
 n = dmTableGetNoRows( table );
 dmsg( 1, "Allocate GTI" );
 gti = alloc_gti( n );
 dmsg( 1, "Get n rows" );
 dmGetScalars_d( start, gti->start, 1, n );
 dmGetScalars_d( stop, gti->stop, 1, n );
 gti->n = n;
 dmsg(1, "Got GTI" );
 return( gti );
}


/*
 * Sum the exposure time in a GTI. No dead time correction is made.
 * Return the sum.
 */

double sum_gti( GTI* gti )
{
 double sum;
 int i;
 sum = 0;
 for ( i = 0; i < gti->n; i++ ) {
  sum += gti->stop[i] - gti->start[i];
 }
 return( sum );
}


/*
 * in_gti
 *
 * Return the amount of good time in the interval [t1,t2]
 * by adding up the overlap with each good time interval.
 *
 */
double in_gti( double t1, double t2, GTI* gti )
{
 int i;
 logical q;		/* Is the overlap nonzero? */
 double u1,u2;		/* Overlap interval for [t1,t2] and one GTI interval */
 double sum = 0.0;	/* Accumulated good time */

 for ( i = 0; i < gti->n; i++ ) {
  q = overlap_d( t1, t2, gti->start[i], gti->stop[i], &u1, &u2 );
  if ( q ) sum += u2 - u1;
 }
 return( sum );
}



/*
 * Alloc GTI structure  with given max number of intervals.
 *
 */

GTI* alloc_gti( int n )
{
 GTI* gti;
 gti = (GTI* )malloc( sizeof( GTI ) );
 gti->start = (double*)malloc( n*sizeof( double ) );
 gti->stop  = (double*)malloc( n*sizeof( double ) );
 gti->nmax = n;
 gti->n = 0;
 return( gti );
}

/*
 * Free GTI structure
 */

void free_gti( GTI* gti )
{
 if ( gti ) {
  free( gti->start );
  free( gti->stop  );
  free( gti );
 }
}



