#include "xgain.h"

Observation* xgain_read_event_file( char* event_file, char* extname )
{
 Observation* observation;
 dmDataset* event_ds;


 observation = alloc_obs();

 dmsg1( 1, "Open DS ", event_file );
 event_ds  = dmDatasetOpen( event_file );
 dmsg( 1, "Open DS OK, now get GTI" );
 observation->gti = dataset_get_gti( event_ds );

 xgain_get_pointing_info( event_ds, extname, observation );
 dmsg( 1, "Close DS" );
 dmDatasetClose( event_ds );

 if ( dbg_level(1) ) obs_print( observation );

 return( observation );
}

/*
 * xgain_get_pointing_info
 *
 * Get the nominal pointing direction and the spacecraft clock zero point from
 * the events table. Support various alternate possibilities for the keyword names.
 *
 * Assumes that the pointing direction is given in the
 * TCRVLn keywords for the columns named "X" and "Y".
 * Assumes that the pixel size is given in the TCDLTn keywords for those columns.
 * Assumes that the nominal roll angle is given in the ROLL_NOM keyword.
 * (for back compatibility, also can be in the XS-DANG keyword with opposite sign.)
 * Assumes that the zero time is given in either the MJDREF keyword
 * or in a pair of MJDREFI, MJDREFF keywords. (XS-MJDRD, XS-MJDRF also
 * supported for ancient back compatbility).
 */
void xgain_get_pointing_info( dmDataset* ds, char* tablename, Observation* observation )
{
 dmBlock* event_table;
 dmDescriptor* xwcs;
 dmDescriptor* ywcs;
 long itzero;
 double nroll;
 double angle;
 double tzero;
 Transform wcs_data;
 Transform* wcs = &wcs_data;


/* Find the coordinate systems associated with columns X and Y
   in the event table. 
 */
 dmsg( 1, "Open events" );
 event_table = dmBlockOpen( ds, tablename );
 if ( !event_table ) {
  printf( "Failed to open table %s\n", tablename );
 }
 dmsg( 1, "Get coords" );
 xwcs = dmDescriptorGetCoord( dmTableOpenColumn( event_table, "X" ) );
 ywcs = dmDescriptorGetCoord( dmTableOpenColumn( event_table, "Y" ) );
 dmCoordGetTransform_d( xwcs, &wcs->crpix[0], &wcs->crval[0], &wcs->cdelt[0], 1 );

 dmCoordGetTransform_d( ywcs, &wcs->crpix[1], &wcs->crval[1], &wcs->cdelt[1], 1 );
/* Search for MJDREF keyword or equivalent */
 dmsg( 1, "Get time keywords" );
 
 if( dmKeyRead_l( event_table, "MJDREFI", &itzero ) ) {
  dmKeyRead_d( event_table, "MJDREFF", &tzero );
  tzero += itzero;
 } else {
  if( !dmKeyRead_d( event_table, "MJDREF", &tzero ) ) {
   if ( dmKeyRead_l( event_table, "XS-MJDRD", &itzero ) ) {
    dmKeyRead_d( event_table, "XS-MJDRF", &tzero );
    tzero += itzero;
   } else {
    tzero = 0;
    tmsg( "No MJDREF keyword found" );
   }
  }
 }
/* Search for nominal roll */
 dmsg( 1, "Got MJDREF" );
 if ( dmKeyRead_d( event_table, "XS-DANG", &angle ) ) {
  nroll = -angle;
 } else if( !dmKeyRead_d( event_table, "ROLL_NOM", &nroll ) ) {
   tmsg( "No ROLL NOM in header" );
 }
 dmsg( 1, "Got ROLL" );

 observation->tzero = tzero;
 observation->cen[0] = wcs->crpix[0];
 observation->cen[1] = wcs->crpix[1];
 observation->nominal_point[RA]   = wcs->crval[0]; 
 observation->nominal_point[DEC]  = wcs->crval[1];
 observation->nominal_point[ROLL] = nroll;
 observation->pixel_size = wcs->cdelt[1]; 
 printf( "Detector center: %f %f\n", observation->cen[0], observation->cen[1] );
 printf( "Nominal pointing: %f %f %f\n", observation->nominal_point[RA],
                                         observation->nominal_point[DEC], observation->nominal_point[ROLL] );
 printf( "Pixel size: %f   Zero time: %f\n", observation->pixel_size, observation->tzero );
}



Observation* alloc_obs( void )
{
 return( (Observation*)malloc( sizeof( Observation ) ) );
}

void free_obs( Observation* obs )
{
 free_gti( obs->gti );
 free( obs );
}

/* 
 * Dump the GTIs and pointing direction
 */
void obs_print( Observation* obs )
{
 int i;
 GTI* gti;
 card buf;

 gti = obs->gti;
 msg( " " );
 msg( "Good time intervals:" );
 for ( i = 1; i <= gti->n; i++ ) {
  sprintf( buf, "%6d %20.6f %20.6f", i, gti->start[i-1], gti->stop[i-1] );
  msg( buf );
 }
 msg( "Pointing direction:" );
 sprintf( buf, "RA: %12.6f deg   Dec: %12.6f deg  Roll: %12.6f deg",
     obs->nominal_point[RA], 
     obs->nominal_point[DEC], 
     obs->nominal_point[ROLL] );
 msg( buf );
 sprintf( buf, "Pixel size:  %12.6f arcsec", obs->pixel_size * 3600.0 );
 msg( buf );
}
