/*
 * HRI GAIN CODE - DM VERSION
 */
#include "xgain.h"

int main( int nargs, char** args )
{
 ParameterList* parlist;            /* Program parameter structure */
 t_filename event_file;      /* Name of event FITS file */
 t_filename anc_file;        /* Name of aspect FITS file */
 t_filename pos_file;        /* Name of ASCII source spatial description file */
 t_filename maps;            /* HRI gain calibration file */
 t_filename outfile;
 t_filename steps_file;      /* Times of gain steps in MJD */
 t_buf aspcols;			/* Names for aspect columns */
 t_word mode;                /* REV0 or RDF format */
 char** colnames;
 card evt_tabname;
 logical offsets;
 GainCal* gain_data;   /* HRI gain calibration */
 Source* source;            /* Source spatial description */
 t_ghist* ghist;             /* Gain histogram structure */
 Aspect_Solution* aspect;           /* Aspect solution */
 Observation* observation;     /* Observation info */
 double bso[3];               /* Boresight offsets */
 double bso0[3] = { 48.0, 40.0, 0.0 };  /* Default boresight offsets */

 double gdef[3]= { 3.0, 7.0, 0.05};   /* Default gain histogram parameters */
 double g[3];                 /* Gain histogram parameters */
 int d; 
 t_buf buf;
 char* ptr;
 int i;

 jcmlib_init();
 dmInit();

/*
 * Initialize the parameter list from a parameter file or
 * from command line arguments.
 * Command line parkernel=sao uses SAO parameter interface
 */

 parlist = par_init( "jcm", "xgain.par", args, nargs );
/* 
 * Undocumented "debug" parameter.
 * The default is d=0. Try d=1 or more for debugging info.
 */

 d = par_get_i( parlist, "d", 0 );
 dbg_set( d );

/*
 * Look for the event file, the aspect file, and the gain maps 
 * calibration file.
 */

 par_get( parlist, "events", "bas.fits", event_file, FSIZE );
 par_get( parlist, "aspect", "anc.fits", anc_file, FSIZE );
 par_get( parlist, "maps", "gain.fits[RHRI_GAIN]", maps, FSIZE );
 par_get( parlist, "steps", " ", steps_file, FSIZE );
 par_get( parlist, "outfile", "xgain.fits", outfile, FSIZE );
 par_get( parlist, "aspcols", "TIME,RA_SC,DEC_SC,ROAN_SC", aspcols, BSIZE );
 colnames = alloc_array_c( CSIZE, 4 );

/* Read the source pixel position file. Failing that, read
 * a single pair of sky coordinates xpos, ypos. Failing that,
 * choose default values (chosen to be near the center of the 
 * ROSAT HRI field).
 */

 par_get( parlist, "pos", " ", pos_file, FSIZE );
 if ( cs_is_blank( pos_file ) ) {
  source = alloc_source( 1 );
  dmsg( 1, "No position file: using xpos and ypos parameters" );
  source->x0 = par_get_d( parlist, "xpos", 4096.0 );
  source->y0 = par_get_d( parlist, "ypos", 4097.0 );
 } else {
  source = source_read( pos_file, NPMAX );
  if ( !source ) exit(1);
 }

/*
 * Read the mode, which by default is "RDF".
 */
 par_get( parlist, "mode", " ", mode, WSIZE );
 i = cs_ends( anc_file );
 while ( anc_file[i] != '.' && i > 0 ) i--;
 if ( cs_is_blank( mode ) ) {
  if ( cs_eq( anc_file+i, ".ao" ) ) {
    cs_copy( "REV0", mode );
  } else {
    cs_copy( "RDF", mode );
  }
 }

 if ( cs_eq( mode, "RDF" ) ) {
  offsets = FALSE;
  cs_copy( "STDEVT", evt_tabname );
 } else {
  offsets = TRUE;
  cs_copy( "EVENTS", evt_tabname );
 }
/*
 * Read the boresight offsets, if non-default.
 * Hopefully the defaults are OK for the ROSAT HRI.
 */

 par_get_vec_d( parlist, "bso", bso0, bso );

/*
 * Declare the range of the gain histogram.
 */

 par_get_vec_d( parlist, "ghist", gdef, g );

 par_list( parlist );

/*
 * Start reading the data
 */
 ptr = aspcols;
 for ( i = 0; i < 4; i++ ) {
  if( ptr ) cs_get_c( &ptr, colnames[i], CSIZE );
 }

 dashline( 45 );
 tmsg( "SOURCE GAIN HISTOGRAM" );
 sprintf( buf, "Archival data type: %s", mode );
 tmsg( buf );
 dashline( 45 );
 tmsg( " " );
 sprintf( buf, "Reading event file %s", event_file );
 tmsg( buf );

 observation = xgain_read_event_file( event_file, evt_tabname );
 if (!observation) exit(1);
 sprintf( buf, "Reading aspect file %s", anc_file );
 tmsg( buf );

 aspect = xgain_read_aspect( anc_file, offsets, colnames, bso, observation->nominal_point,
      observation->pixel_size );
 if ( !aspect ) exit(1);
 sprintf( buf, "Reading gain map calibrations %s", maps );
 tmsg( buf );

 gain_data = xgain_cal_read( maps );
 if ( !gain_data ) exit(1);
 dmsg( 1, "Read gain maps" );

 if ( !cs_is_blank( steps_file ) ) {
  sprintf( buf, "Reading gain steps file %s", steps_file );
  tmsg( buf );
  read_gain_steps_file( gain_data, steps_file );
 }
/* Initialize the gain histogram structure and do the processing */

 ghist = ghset( g );

 ghist->total_time = observation->gti->total;
 dmsg( 1, "Run binning code" );
 xgain_calc( source, aspect, observation, gain_data, ghist );
 xgain_write( outfile, ghist );
/* Print the results */
 dmsg( 1, "List results" );
 xgain_list( ghist );
 if ( source->n <= 1 ) {
  dmsg( 1, "List times" );
  xgain_list_times( ghist );
 } 
 dmsg( 1, "Free structures" );
/* Free memory */
 free_array_c( colnames );
 free_source( source );
 free_obs( observation );
 free_asp( aspect );
 free_gain_data( gain_data );
 free_ghist( ghist );
 return(0);
}



