#ifndef XGAIN_H
#define XGAIN_H

/* The following things are currently hardcoded; in a multi-mission
 * version they would be parameters or read from the data:
 *
 *  Gain map size:  GMAPSIZE x GMAPSIZE with GMAPSIZE = 32 
 *  Gain map WCS relative to sky map (hard coded in gain.c)
 *  Aspect file table name and column names: 
 *        ASPECT     RA_SC DEC_SC ROAN_SC     (for mode RDF)
 *        AO         XOFF YOFF ROFF           (for mode REV0)
 *  GTI table name:  STDGTI
 *  GTI start/stop column positions: 1 and 2
 *  Event table name: STDEVT or EVENTS    ( mode RDF, mode REV0)
 *  Event table RA and Dec pixel column names: X and Y
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
/*
 * The JCMLIB library is my own utility library, used for
 * string handling, simple file I/O, etc.
 * The ASCDM (ASC Data Model) library provides a high level interface
 * to CFITSIO. The version initially distributed with XGAIN 
 * is my own prototype; a more robust version with much more functionality
 * will be available soon from the AXAF Science Center.
 * The PARS library is a simple parameter file interface, incompatible
 * with the SAO/XANADU/IRAF style interface - sorry!
 */

#include "jcmlib.h"
#include "dm.h"
#ifndef JCMDM_H
#include "compat.h"
#endif

#include "jcmpar.h"

/* There are two kinds of aspect data used: 
 * absolute pointing triples  (RA, DEC, ROLL)   units (deg, deg, deg)
 * relative pointing triples  (X,Y,ROLL) units (pixel,pixel,deg)
 * The latter are parallel to detector axes not sky axes.
 * We use symbolic constants to index into the triples.
 */
#define RA 0
#define DEC 1
#define ROLL 2
#define X_OFF 0
#define Y_OFF 1

/*
 * Structures for the ROSAT HRI Gain Map program
 *
 */

/*
 * A gain map is a 32 x 32 map in blocked detector coordinates.
 */
#define GMAPSIZE 32
typedef double GainMap[GMAPSIZE][GMAPSIZE];

/*
 * The gain data structure is a set of gain maps versus time.
 * These are read in from a FITS BINTABLE.
 * The time values are in MJD.
 */
typedef struct {
 int nmap;		/* Number of maps */
 int nmax;
 double* time;		/* Start time of map being applicable */
 GainMap* maps;	/* Map for corresponding time */
 int nsteps;       /* Number of interpolation boundaries due to HV changes */
 double* tsteps;   /* MJD times of HV changes */
} GainCal;


/*
 * The source is modelled as a list of sky pixel values.
 * These may be entered in an ASCII file.
 * The default mode is to use a single-pixel 'point source'
 * whose sky coords can be entered as command line parameters XPOS, YPOS.
 */

#define NPMAX 1000
typedef struct {
 int n;
 int nmax;
 double x0;
 double y0;
 double* xpos;
 double* ypos;
} Source;

/*
 * The good time intervals record the relevant data periods in seconds.
 */

typedef struct {
 int n;
 int nmax;
 double* start;
 double* stop;
 double total;
} GTI;

/* 
 * The relevant observational details: the good time intervals, the
 * time zero point, the nominal pointing direction and the pixel size.
 */
typedef struct {
 GTI* gti;
 double tzero;
 double nominal_point[3];
 double cen[2];
 double pixel_size;
} Observation;

/*
 * The aspect offsets versus time
 */

typedef struct {
 int n;
 int nmax;
 double* t;
 double** data;
} Aspect_Solution;

/*
 *  The coordinate system, read from the event table
 */

typedef struct {
 double crpix[2];
 double crval[2];
 double cdelt[2];
} Transform;

/* 
 * The gain histogram is the output structure for this program.
 *
 */

 typedef struct { 
  double* x;
  double* y;
  double* asp[3];
 } GainHist_Accum;


#define NGMAX 1000
typedef struct {
 double dg;
 double g0;
 int ng;
 double total_time;
 double* data;
 double* gt1;    /* Overall time for each gain histogram bin */
 double* gt2;
 GTI** gti;       /* GTI for each gain bin */
 GainHist_Accum mean;
 GainHist_Accum var;
} t_ghist;


/* Prototypes */

void free_gain_data( GainCal* data );
GainCal* alloc_gain_data( int n, int ns );
GainCal* xgain_cal_read( char* maps );
void read_gain_steps_file( GainCal* data, char* steps_file );

Aspect_Solution* xgain_read_aspect( char* anc_file, logical offsets, char** colnames, double* bso, double* nominal_point, double pixel_size );
Aspect_Solution* aspect_read(  dmDataset* aspect_ds, double* bso, double* nominal_point, double pix, logical offsets, char** colnames );
Aspect_Solution* alloc_asp( int n );
void free_asp( Aspect_Solution* asp );
void bore_correct( double* bso, double* asp );
void point_to_asp( double* point, double* nominal, double pixel_size, double* asp );

void free_obs( Observation* obs );
Observation* alloc_obs( void );
Observation* xgain_read_event_file( char* event_file, char* mode );
void xgain_get_pointing_info( dmDataset* ds, char* tablename, Observation* observation );
void obs_print( Observation* obs );

double sum_gti( GTI* gti );
GTI* read_gti( dmBlock* table );
GTI* alloc_gti( int n );
GTI* dataset_get_gti( dmDataset* event_ds );
void free_gti( GTI* gti );
double in_gti( double t1, double t2, GTI* gti );

Source* alloc_source( int n );
void free_source( Source* source );
Source* source_read( char* file, int nmax );

void xgain_calc( Source* source, Aspect_Solution* aspect, Observation* obs, GainCal* gain_data, t_ghist* ghist );

void xgain_list( t_ghist* );
void  xgain_list_times( t_ghist* ghist );
void accum_time_filter( double t1, double t2, GTI* gti );

void unaspect( double x, double y, double xcen, double ycen, double* asp, double* detx, double* dety );

void xgain_accumulate_histogram( double gain, double wt, double t1, double t2, double detx, double dety, double* xasp, t_ghist* ghist, int npos );
void xgain_write( char* filename, t_ghist* g );

void free_ghist( t_ghist* g );
t_ghist* alloc_ghist( int ng );
t_ghist* ghset( double* g );

double sc_time_to_mjd( double t, double mjd_zero );
int find_entry_d( double time, double* times, int n );
int find_entry_and_step_d( double time, double* times, int n, double* steps, int ns );
double do_interp( GainCal* gain_data, double t, double detx, double dety );
double interp_gain( GainMap* map, double gx, double gy );

double psqrt( double x );

#endif
